import json
import urllib2

from django.core.exceptions import ValidationError
from django.test.utils import override_settings
from gargoyle.testutils import switches

from api.v20 import registration
from identityprovider.fields import FIELD_REQUIRED, INVALID_EMAIL
from identityprovider.forms import INVALID_NAME
from identityprovider.login import PasswordPolicyError
from identityprovider.models import Account
from identityprovider.models.const import (
    AccountCreationRationale,
    AccountStatus,
)
from identityprovider.tests import DEFAULT_USER_PASSWORD
from identityprovider.tests.utils import SSOBaseTestCase


class RegistrationTestCase(SSOBaseTestCase):

    class fake_lp_response():
        def __init__(self, code, message, exception=False):
            self.code = code
            self.message = message
            self.exception = exception

        def read(self):
            if not self.exception:
                return json.dumps(self.message)
            return self.message

        def readline(self):
            return self.message

    def setUp(self):
        super(RegistrationTestCase, self).setUp()
        self.email = self.factory.make_email_address()
        self.data = [
            self.sso_root_url, self.email, 'SecretPassword1', 'displayname']

        self.mock_send_impersonation_email = self.patch(
            'api.v20.registration.send_impersonation_email')
        self.mock_send_new_user_email = self.patch(
            'api.v20.registration.send_new_user_email')

    def assert_correct_account_content(
            self, account, username=None, password=None):
        self.assertEqual(Account.objects.all().count(), 1)
        account = Account.objects.get()
        self.assertEqual(account.displayname, 'displayname')
        self.assertEqual(account.person_name, username)
        if username is None:
            username = account.openid_identifier
        self.assertEqual(account.username, username)
        self.assertEqual(account.preferredemail.email, self.email)
        self.assertFalse(account.is_verified)

        if password is not None:
            account.user.check_password(password)
        self.mock_send_new_user_email.assert_called_once_with(
            root_url=self.sso_root_url, account=account, email=self.email,
            oid_token=None)
        self.assertFalse(self.mock_send_impersonation_email.called)

    def test_register_already_registered_and_inactive(self):
        account = self.factory.make_account(
            email=self.email, status=AccountStatus.DEACTIVATED)
        assert not account.is_active

        with self.assertRaises(registration.EmailAlreadyRegistered):
            registration.register(
                self.sso_root_url, email=self.email, password="",
                displayname="")

        self.assertFalse(self.mock_send_impersonation_email.called)
        self.assertFalse(self.mock_send_new_user_email.called)

    def test_register_already_registered_and_active(self):
        account = self.factory.make_account(email=self.email)
        assert account.is_active

        with self.assertRaises(registration.EmailAlreadyRegistered):
            registration.register(
                self.sso_root_url, email=self.email, password="",
                displayname="")

        self.mock_send_impersonation_email.assert_called_once_with(
            self.sso_root_url, self.email)
        self.assertFalse(self.mock_send_new_user_email.called)

    def test_register_success(self):
        assert not Account.objects.exists()
        account = registration.register(*self.data)

        self.assert_correct_account_content(
            account, password='SecretPassword1')

    def test_register_with_username_when_feature_flag_not_enabled(self):
        assert not Account.objects.exists()
        self.data.append('thisisausername')
        account = registration.register(*self.data)

        # At the registration stage, if username is not required,
        # the username of the account will be its openid_identifier
        # and the `person_name` of the account will be `None`
        self.assert_correct_account_content(
            account, password='SecretPassword1')

    @switches(USERNAME_UI=True)
    @override_settings(LP_API_URL='test.com')
    def test_register_with_username_when_feature_flag_enabled(self):
        """Tests cases where `username` supplied to APIs with USERNAME_UI set.

        This tests cases of API calls with `username` supplied with
        USERNAME_UI flag is set. This requires to set the lp_person name
        in the placeholder launchpad account.
        """
        assert not Account.objects.exists()
        username = 'thisisausername'
        self.data.append(username)

        fake_lp_response = self.fake_lp_response(200, username)
        self.patch(
            'webservices.launchpad.http_request_with_timeout',
            return_value=fake_lp_response)

        account = registration.register(*self.data)

        self.assert_correct_account_content(
            account, username=username, password='SecretPassword1')

    @switches(USERNAME_UI=True)
    @override_settings(LP_API_URL='test.com')
    def test_register_with_existing_username_when_feature_flag_enabled(self):
        username = 'thisisausername'
        self.data.append(username)

        mock_http_call = self.patch(
            'webservices.launchpad.http_request_with_timeout')
        fake_resp = self.fake_lp_response(
            401,
            'name: .* is already in use by another person or team.',
            exception=True)
        mock_http_call.side_effect = urllib2.HTTPError(
            None, fake_resp.code, None, None, fake_resp)

        # Make sure that there are no accounts at the start.
        assert not Account.objects.exists()
        with self.assertRaises(ValidationError) as ctx:
            registration.register(*self.data)

        # An error from lp should not let the username creation.
        assert not Account.objects.exists()
        self.assertEqual(
            ctx.exception.message_dict,
            {'username': ['thisisausername is already in use.']})

    @switches(USERNAME_UI=True)
    @override_settings(LP_API_URL='test.com')
    def test_register_no_placeholder_username_when_feature_flag_enabled(self):
        username = 'thisisausername'
        self.data.append(username)

        mock_http_call = self.patch(
            'webservices.launchpad.http_request_with_timeout')
        error_message = 'An account for that OpenID identifier already exists.'
        fake_resp = self.fake_lp_response(401, error_message, exception=True)
        mock_http_call.side_effect = urllib2.HTTPError(
            None, fake_resp.code, None, None, fake_resp)

        # Make sure that there are no accounts at the start.
        assert not Account.objects.exists()
        with self.assertRaises(Exception) as ctx:
            registration.register(*self.data)

        # An error from lp should not let the username creation.
        assert not Account.objects.exists()
        e = ctx.exception
        self.assertEqual(
            str(e),
            ('Dry run of setting username(%s) with dummy '
             'openid_identifier(%s) in Launchpad failed with (%s).') % (
                username, 'dummyID~', error_message))

    @switches(USERNAME_UI=True)
    @override_settings(LP_API_URL='test.com')
    def test_register_with_random_http_error_when_feature_flag_enabled(self):
        username = 'thisisausername'
        self.data.append(username)

        mock_http_call = self.patch(
            'webservices.launchpad.http_request_with_timeout')
        error_message = 'Just another random http error.'
        fake_resp = self.fake_lp_response(401, error_message, exception=True)
        mock_http_call.side_effect = urllib2.HTTPError(
            None,  fake_resp.code, error_message, None, fake_resp)

        # Make sure that there are no accounts at the start.
        assert Account.objects.all().count() == 0
        with self.assertRaises(Exception) as ctx:
            registration.register(*self.data)

        # An error from lp should not let the username creation.
        assert Account.objects.all().count() == 0
        e = ctx.exception
        self.assertEqual(
            str(e),
            ('Dry run of setting username(%s) with dummy '
             'openid_identifier(%s) in Launchpad failed with (%s).') % (
                username, 'dummyID~', error_message))

    @switches(USERNAME_UI=False)
    def test_register_no_placeholder_username_when_feature_flag_disabled(self):
        assert not Account.objects.exists()
        username = 'thisisausername'
        self.data.append(username)
        mock_http_call = self.patch(
            'webservices.launchpad.http_request_with_timeout')
        fake_resp = self.fake_lp_response(
            401,
            'An account for that OpenID identifier already exists.',
            exception=True)
        mock_http_call.side_effect = urllib2.HTTPError(
            None, fake_resp.code, None, None, fake_resp)

        account = registration.register(*self.data)

        self.assert_correct_account_content(
            account, password='SecretPassword1')

    @switches(USERNAME_UI=False)
    def test_register_existing_username_when_feature_flag_disabled(self):
        assert not Account.objects.exists()
        username = 'thisisausername'
        self.data.append(username)
        mock_http_call = self.patch(
            'webservices.launchpad.http_request_with_timeout')
        fake_resp = self.fake_lp_response(
            401,
            'name: .* is already in use by another person or team.',
            exception=True)
        mock_http_call.side_effect = urllib2.HTTPError(
            None, fake_resp.code, None, None, fake_resp)

        account = registration.register(*self.data)

        self.assert_correct_account_content(
            account, password='SecretPassword1')

    @switches(USERNAME_UI=True)
    def test_register_without_username_when_feature_flag_enabled(self):
        """Test when `username` is not supplied in API calls with USERNAME_UI set.

        This tests cases of API calls without `username` supplied even
        with USERNAME_UI flag is set, making sure the backword compatibility
        is not broken. There is no need in this case to set the lp_person name
        in the placeholder launchpad account.
        """
        assert not Account.objects.exists()
        account = registration.register(*self.data)

        self.assert_correct_account_content(
            account, password='SecretPassword1')

    @switches(USERNAME_UI=True)
    def test_reject_invalid_username_with_blank_password(self):
        username = 'InvalidUsername'
        with self.assertRaises(ValidationError) as ctx:
            registration.register(
                self.sso_root_url, email=self.email, password=None,
                displayname='displayname', username=username)
        e = ctx.exception
        self.assertEqual(e.message_dict['username'], [INVALID_NAME])
        # Make sure no account is created.
        assert not Account.objects.exists()

    @switches(USERNAME_UI=True)
    @override_settings(LP_API_URL='test.com')
    def test_can_create_account_without_pw_with_feature_flag_enabled(self):
        # This test is to show that password is not strictly needed and this
        # can be achieved by using `registration.register` with
        # `require_password=False`
        username = 'validusername'
        fake_lp_response = self.fake_lp_response(200, username)
        self.patch(
            'webservices.launchpad.http_request_with_timeout',
            return_value=fake_lp_response)

        account = registration.register(
            self.sso_root_url, email=self.email, password=None,
            displayname='displayname', username=username,
            require_password=False)

        self.assert_correct_account_content(account, username=username)

    @switches(USERNAME_UI=False)
    def test_can_create_account_without_pw_with_feature_flag_disabled(self):
        account = registration.register(
            self.sso_root_url, email=self.email, password=None,
            displayname='displayname', username='validusername',
            require_password=False)

        self.assert_correct_account_content(account)

    @switches(USERNAME_UI=True)
    @override_settings(LP_API_URL='test.com')
    def test_can_create_account_with_empty_pw_with_feature_flag_enabled(self):
        # This test is to show that empty password cases are treated the same
        # way as `password=None` cases.
        username = 'validusername'
        fake_lp_response = self.fake_lp_response(200, username)
        self.patch(
            'webservices.launchpad.http_request_with_timeout',
            return_value=fake_lp_response)

        account = registration.register(
            self.sso_root_url, email=self.email, password='',
            displayname='displayname', username=username,
            require_password=False)

        self.assert_correct_account_content(account, username=username)

    @switches(USERNAME_UI=False)
    def test_can_create_account_with_empty_pw_with_feature_flag_disabled(self):
        account = registration.register(
            self.sso_root_url, email=self.email, password='',
            displayname='displayname', require_password=False)

        self.assert_correct_account_content(account)

    def test_password_required_by_default(self):
        with self.assertRaises(ValidationError) as ctx:
            registration.register(
                self.sso_root_url, email=self.email, password=None,
                displayname='displayname', username='validusername')

        e = ctx.exception
        self.assertEqual(e.message_dict['password'], [FIELD_REQUIRED])

    @switches(USERNAME_UI=True)
    def test_form_validation_with_valid_pw_with_feature_flag_enabled(self):
        with self.assertRaises(ValidationError) as ctx:
            registration.register(
                self.sso_root_url, email='bad.email', password='TheSecret1',
                displayname=None, username='InvalidUsername')
        e = ctx.exception
        self.assertEqual(e.message_dict['displayname'], [FIELD_REQUIRED])
        self.assertEqual(e.message_dict['email'], [INVALID_EMAIL])
        self.assertEqual(e.message_dict['username'], [INVALID_NAME])
        # Make sure no account is created.
        assert not Account.objects.exists()

    @switches(USERNAME_UI=False)
    def test_form_validation_with_valid_pw_with_feature_flag_disabled(self):
        with self.assertRaises(ValidationError) as ctx:
            registration.register(
                self.sso_root_url, email='bad.email', password='TheSecret1',
                displayname=None, username='InvalidUsername')

        e = ctx.exception
        self.assertEqual(e.message_dict['displayname'], [FIELD_REQUIRED])
        self.assertEqual(e.message_dict['email'], [INVALID_EMAIL])
        # When `username` is not mandatory, we don't flag on invalid usernames,
        # we're simply not going to use it.
        self.assertNotIn('username', e.message_dict)
        # Make sure no account is created.
        assert not Account.objects.exists()

    @switches(USERNAME_UI=True)
    def test_form_validation_with_no_pw_and_with_feature_flag_enabled(self):
        with self.assertRaises(ValidationError) as ctx:
            registration.register(
                self.sso_root_url, email='bad.email', password=None,
                displayname=None, username='InvalidUsername')

        e = ctx.exception
        self.assertEqual(e.message_dict['displayname'], [FIELD_REQUIRED])
        self.assertEqual(e.message_dict['email'], [INVALID_EMAIL])
        self.assertEqual(e.message_dict['username'], [INVALID_NAME])
        self.assertEqual(e.message_dict['password'], [FIELD_REQUIRED])
        # Make sure no account is created.
        assert not Account.objects.exists()

    @switches(USERNAME_UI=False)
    def test_form_validation_with_no_pw_and_feature_flag_disabled(self):
        with self.assertRaises(ValidationError) as ctx:
            registration.register(
                self.sso_root_url, email='bad.email', password=None,
                displayname=None, username='InvalidUsername',
                require_password=False)

        e = ctx.exception
        self.assertEqual(e.message_dict['displayname'], [FIELD_REQUIRED])
        self.assertEqual(e.message_dict['email'], [INVALID_EMAIL])
        # When `username` is not mandatory, we don't err on invalid usernames,
        # we're simply not going to use it.
        self.assertNotIn('username', e.message_dict)
        # Empty `password` is not an issue when `require_password` is `False`.
        self.assertNotIn('password', e.message_dict)
        # Make sure no account is created.
        assert not Account.objects.exists()

    @switches(USERNAME_UI=False)
    def test_ignore_invalid_username_when_feature_flag_disabled(self):
        """Test when invalid `username` is provided with feature flag disabled.

        Since we do not set `username` for cases USERNAME_UI flag is not set,
        we can ignore them even if they're of invalid format.
        """
        assert not Account.objects.exists()
        self.data.append('InvalidUsername')
        account = registration.register(*self.data)

        # Make sure, the `username` in the json_body is `None`.
        self.assert_correct_account_content(account)

    def test_invalid_email_and_no_pw_with_default_require_password(self):
        with self.assertRaises(ValidationError) as ctx:
            registration.register(
                self.sso_root_url, email="", password=None,
                displayname='displayname', require_password=True)

        e = ctx.exception
        self.assertEqual(e.message_dict['email'], [FIELD_REQUIRED])
        self.assertEqual(e.message_dict['password'], [FIELD_REQUIRED])

    def test_invalid_email_and_password_with_no_require_password(self):
        with self.assertRaises(ValidationError) as ctx:
            registration.register(
                self.sso_root_url, email="", password=123,
                displayname='displayname')

        e = ctx.exception
        self.assertEqual(e.message_dict['email'], [FIELD_REQUIRED])
        self.assertEqual(
            e.message_dict['password'],
            ['Password must be at least 8 characters long'])

    def test_register_success_with_oid_token(self):
        assert Account.objects.all().count() == 0

        registration.register(
            self.sso_root_url, self.email, 'SecretPassword1', 'displayname',
            oid_token="fake_oid_token")

        self.assertEqual(Account.objects.all().count(), 1)
        account = Account.objects.get()
        self.assertEqual(account.displayname, 'displayname')
        self.assertEqual(account.preferredemail.email, self.email)
        self.assertFalse(account.is_verified)

        self.assertTrue(
            self.client.login(username=self.email, password='SecretPassword1'),
            'Password was not properly set on new account.')
        self.mock_send_new_user_email.assert_called_once_with(
            root_url=self.sso_root_url, oid_token="fake_oid_token",
            account=account, email=self.email)
        self.assertFalse(self.mock_send_impersonation_email.called)

    def test_invalid_email_and_pw(self):
        with self.assertRaises(ValidationError) as ctx:
            registration.register(self.sso_root_url, "", '', 'displayname')

        e = ctx.exception
        self.assertEqual(e.message_dict['email'], [FIELD_REQUIRED])
        self.assertEqual(e.message_dict['password'], [FIELD_REQUIRED])

    def test_leaked_credentials(self):
        self.factory.make_leaked_credential(email='foo@example.com',
                                            password=DEFAULT_USER_PASSWORD)
        with self.assertRaises(PasswordPolicyError) as ctx:
            registration.register(self.sso_root_url, 'foo@example.com',
                                  DEFAULT_USER_PASSWORD, 'displayname')

        self.assertIn('there was a security breach', str(ctx.exception))
        self.assertIn('Ubuntu One', str(ctx.exception))

    def test_email_not_validated(self):
        account = registration.register(
            self.sso_root_url, self.email, 'SecretPassword1', 'displayname',
            email_validated=False)
        self.assertFalse(account.is_verified)

    def test_email_validated(self):
        account = registration.register(
            self.sso_root_url, self.email, 'SecretPassword1', 'displayname',
            email_validated=True)
        self.assertTrue(account.is_verified)

    def test_creation_rationale_unset_source(self):
        account = registration.register(
            self.sso_root_url, self.email, 'SecretPassword1', 'displayname')
        self.assertEqual(
            account.creation_rationale, AccountCreationRationale.API_CREATED)
        self.assertIsNone(account.creation_source)

    def test_creation_rationale_set_source(self):
        account = registration.register(
            self.sso_root_url, self.email, 'SecretPassword1', 'displayname',
            creation_source='foo bar baz')
        self.assertEqual(
            account.creation_rationale, AccountCreationRationale.API_CREATED)
        self.assertEqual(account.creation_source, 'foo bar baz')
