import json
from datetime import timedelta
from functools import partial
from operator import attrgetter

from django.urls import reverse

from api.v20.utils import (
    EnsureJSONResponseOnAPIErrorMiddleware,
    get_account_data,
    get_accounts_data,
    get_email_data,
    get_minimal_account_data,
    get_minimal_accounts_data,
    get_token_data,
)
from identityprovider.models.const import EmailStatus
from identityprovider.tests.utils import SSOBaseTestCase


class EnsureJSONResponseOnAPIErrorMiddlewareTestCase(SSOBaseTestCase):

    def test_process_nothing_but_api_v20_urls(self):
        exception = RuntimeError('foo')
        instance = EnsureJSONResponseOnAPIErrorMiddleware()

        for url in ('/', '/api/', '/api/1.0/', '/api/1.1/'):
            request = self.factory.make_request(url)
            response = instance.process_exception(request, exception)
            self.assertIsNone(response)

    def test_only_intercep_api_v20_urls(self):
        request = self.factory.make_request('/api/v2/')
        exception = RuntimeError('foo')
        response = EnsureJSONResponseOnAPIErrorMiddleware().process_exception(
            request, exception)
        self.assertIsNotNone(response)
        self.assertEqual(
            response['Content-Type'], 'application/json; charset=utf-8')
        expected_data = {
            'code': 'SERVER_ERROR',
            'message': 'RuntimeError: foo',
            'extra': {}
        }
        self.assertEqual(json.loads(response.content), expected_data)


class PluralTestCase(SSOBaseTestCase):

    def assert_equivalent_plural(self, plural_func, singular_func, objects,
                                 get_key):
        """Assert that a plural function matches its singular sibling.

        `plural_func` applied to all objects at once produces the same
        results as `singular_func` applied to each object in turn.
        """
        plural_results = plural_func(objects)
        singular_results = {
            get_key(obj): singular_func(obj) for obj in objects
        }
        self.assertEqual(plural_results, singular_results)


class GetAccountsDataTestCase(PluralTestCase):

    def make_accounts(self, count):
        return [
            self.factory.make_account(email_validated=(i % 2) == 0)
            for i in range(count)
        ]

    def test_get_minimal_accounts_data(self):
        accounts = self.make_accounts(4)
        self.factory.make_person(account=accounts[0])

        self.assert_equivalent_plural(
            get_minimal_accounts_data, get_minimal_account_data,
            accounts[:3], attrgetter('openid_identifier'))

    def test_get_accounts_data(self):
        accounts = self.make_accounts(4)
        self.factory.make_person(account=accounts[0])
        for _ in range(15):
            self.factory.make_email_for_account(account=accounts[1])
        for _ in range(15):
            self.factory.make_oauth_token(account=accounts[2])

        self.assert_equivalent_plural(
            get_accounts_data, partial(get_account_data, include_tokens=False),
            accounts[:3], attrgetter('openid_identifier'))
        self.assert_equivalent_plural(
            partial(get_accounts_data, limit=5),
            partial(get_account_data, limit=5, include_tokens=False),
            accounts[:3], attrgetter('openid_identifier'))
        self.assert_equivalent_plural(
            partial(get_accounts_data, expand=True),
            partial(get_account_data, expand=True, include_tokens=False),
            accounts[:3], attrgetter('openid_identifier'))
        self.assert_equivalent_plural(
            partial(get_accounts_data, limit=5, expand=True),
            partial(
                get_account_data, limit=5, expand=True, include_tokens=False),
            accounts[:3], attrgetter('openid_identifier'))


class GetAccountDataTestCase(SSOBaseTestCase):

    openid = '1234567'
    email = 'utils@example.com'
    account_verified = False

    def setUp(self):
        super(GetAccountDataTestCase, self).setUp()
        self.account = self.factory.make_account(
            email=self.email, openid_identifier=self.openid,
            email_validated=self.account_verified)

    def test_get_minimal_account_data(self):
        result = get_minimal_account_data(self.account)
        expected = dict(
            href=reverse('api-account', args=(self.openid,)),
            openid=self.openid,
            username=self.account.person_name,
            verified=self.account_verified,
        )

        self.assertEqual(result, expected)
        self.assertIsNone(result['username'])

    def test_get_minimal_account_data_with_lp_name(self):
        self.factory.make_person(account=self.account)
        result = get_minimal_account_data(self.account)
        expected = dict(
            href=reverse('api-account', args=(self.openid,)),
            openid=self.openid,
            username=self.account.person_name,
            verified=self.account_verified,
        )

        self.assertIsNotNone(result['username'])
        self.assertEqual(result, expected)

    def test_get_account_data_no_tokens(self):
        result = get_account_data(self.account)
        expected = dict(
            href=reverse('api-account', args=(self.openid,)),
            openid=self.openid,
            verified=self.account_verified,
            displayname=self.account.displayname,
            username=self.account.person_name,
            status='Active',
            email=self.email,
            emails=[dict(href=reverse('api-email', args=(self.email,)),
                         verified=self.account_verified)],
            tokens=[],
        )

        self.assertIsNone(result['username'])
        self.assertEqual(result, expected)

    def test_get_account_data_with_tokens(self):
        token1 = self.account.create_oauth_token('a-token')
        token2 = self.account.create_oauth_token('another')
        token2.date_updated = token1.date_updated + timedelta(seconds=1)
        token2.save()

        result = get_account_data(self.account)

        expected = dict(
            href=reverse('api-account', args=(self.openid,)),
            openid=self.openid,
            verified=self.account_verified,
            displayname=self.account.displayname,
            username=self.account.person_name,
            status='Active',
            email=self.email,
            emails=[dict(href=reverse('api-email', args=(self.email,)),
                         verified=self.account_verified)],
            tokens=[
                dict(href=reverse('api-token', args=(token2.pk,)),
                     token_name='another'),
                dict(href=reverse('api-token', args=(token1.pk,)),
                     token_name='a-token'),
            ],
        )

        self.assertIsNone(result['username'])
        self.assertEqual(result, expected)

    def test_get_account_data_with_tokens_and_lp_name(self):
        token1 = self.account.create_oauth_token('a-token')
        token2 = self.account.create_oauth_token('another')
        token2.date_updated = token1.date_updated + timedelta(seconds=1)
        token2.save()

        self.factory.make_person(account=self.account)
        result = get_account_data(self.account)

        expected = dict(
            href=reverse('api-account', args=(self.openid,)),
            openid=self.openid,
            verified=self.account_verified,
            displayname=self.account.displayname,
            username=self.account.person_name,
            status='Active',
            email=self.email,
            emails=[dict(href=reverse('api-email', args=(self.email,)),
                         verified=self.account_verified)],
            tokens=[
                dict(href=reverse('api-token', args=(token2.pk,)),
                     token_name='another'),
                dict(href=reverse('api-token', args=(token1.pk,)),
                     token_name='a-token'),
            ],
        )

        self.assertIsNotNone(result['username'])
        self.assertEqual(result, expected)

    def test_get_account_data_no_tokens_but_with_lp_name(self):
        self.factory.make_person(account=self.account)
        result = get_account_data(self.account)

        expected = dict(
            href=reverse('api-account', args=(self.openid,)),
            openid=self.openid,
            verified=self.account_verified,
            displayname=self.account.displayname,
            username=self.account.person_name,
            status='Active',
            email=self.email,
            emails=[dict(href=reverse('api-email', args=(self.email,)),
                         verified=self.account_verified)],
            tokens=[],
        )

        self.assertIsNotNone(result['username'])
        self.assertEqual(result, expected)

    def test_get_account_data_default_limit_emails(self):
        for i in range(15):
            self.factory.make_email_for_account(self.account)

        result = get_account_data(self.account)

        self.assertEqual(len(result['emails']), 10)

    def test_get_account_data_override_limit_emails(self):
        for i in range(15):
            self.factory.make_email_for_account(self.account)

        result = get_account_data(self.account, limit=15)

        self.assertEqual(len(result['emails']), 15)

    def test_get_account_data_default_limit_tokens(self):
        for i in range(15):
            self.account.create_oauth_token(str(i))

        result = get_account_data(self.account)

        self.assertEqual(len(result['tokens']), 10)

    def test_get_account_data_override_limit_tokens(self):
        for i in range(15):
            self.account.create_oauth_token(str(i))

        result = get_account_data(self.account, limit=15)

        self.assertEqual(len(result['tokens']), 15)

    def test_expanded(self):
        tokens = []
        emails = [self.account.preferredemail]
        for i in range(8):
            tokens.append(self.account.create_oauth_token(str(i)))
            emails.append(self.factory.make_email_for_account(
                self.account, status=EmailStatus.NEW))

        result = get_account_data(self.account, expand=True)
        expected = dict(
            href=reverse('api-account', args=(self.openid,)),
            openid=self.openid,
            verified=self.account_verified,
            displayname=self.account.displayname,
            username=self.account.person_name,
            status='Active',
            email=self.email,
            emails=[get_email_data(e) for e in reversed(emails)],
            tokens=[get_token_data(t) for t in reversed(tokens)],
        )

        self.assertEqual(result, expected)
        self.assertIsNone(result['username'])

    def test_expanded_for_account_with_lp_name(self):
        tokens = []
        emails = [self.account.preferredemail]
        for i in range(8):
            tokens.append(self.account.create_oauth_token(str(i)))
            emails.append(self.factory.make_email_for_account(
                self.account, status=EmailStatus.NEW))
        self.factory.make_person(account=self.account)

        result = get_account_data(self.account, expand=True)
        expected = dict(
            href=reverse('api-account', args=(self.openid,)),
            openid=self.openid,
            verified=self.account_verified,
            displayname=self.account.displayname,
            username=self.account.person_name,
            status='Active',
            email=self.email,
            emails=[get_email_data(e) for e in reversed(emails)],
            tokens=[get_token_data(t) for t in reversed(tokens)],
        )

        self.assertEqual(result, expected)
        self.assertIsNotNone(result['username'])


class GetAccountDataVerifiedTestCase(GetAccountDataTestCase):

    account_verified = True


class GetEmailDataTestCase(SSOBaseTestCase):

    def setUp(self):
        super(GetEmailDataTestCase, self).setUp()
        self.account = self.factory.make_account()

    def assert_email_data(self, result, email, verified=True):
        expected = dict(
            email=email.email,
            verified=verified,
            date_created=email.date_created,
            username=email.account.person_name,
            href=reverse('api-email', args=(email.email,)),
        )
        self.assertEqual(result, expected)

    def test_data_returned_preferred_email(self):
        email = self.account.preferredemail

        result = get_email_data(email)

        self.assert_email_data(result, email)

    def test_data_returned_validated_email(self):
        email = self.factory.make_email_for_account(
            self.account, status=EmailStatus.VALIDATED)
        assert email != self.account.preferredemail

        result = get_email_data(email)

        self.assert_email_data(result, email)

    def test_data_returned_new_email(self):
        email = self.factory.make_email_for_account(
            self.account, status=EmailStatus.NEW)

        result = get_email_data(email)

        self.assert_email_data(result, email, verified=False)

    def test_not_expanded(self):
        email = self.account.preferredemail

        result = get_email_data(email, expand=False)

        expected = dict(
            verified=True,
            href=reverse('api-email', args=(email.email,)),
        )
        self.assertEqual(result, expected)


class GetTokenDataTestCase(SSOBaseTestCase):

    def setUp(self):
        super(GetTokenDataTestCase, self).setUp()
        self.account = self.factory.make_account()

    def test_data_returned(self):
        token = self.account.create_oauth_token('test')

        result = get_token_data(token)

        expected = dict(
            consumer_key=token.consumer.key,
            consumer_secret=token.consumer.secret,
            token_key=token.key,
            token_secret=token.secret,
            token_name=token.name,
            date_created=token.date_created,
            date_updated=token.date_updated,
            href=reverse('api-token', args=(token.pk,)),
        )
        self.assertEqual(result, expected)

    def test_not_expanded(self):
        token = self.account.create_oauth_token('test')

        result = get_token_data(token, expand=False)

        expected = dict(
            token_name=token.name,
            href=reverse('api-token', args=(token.pk,)),
        )
        self.assertEqual(result, expected)
