# Copyright 2010-2018 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import json
from operator import attrgetter

from django.http import HttpResponse
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _

from identityprovider.models.account import Account
from identityprovider.models.const import (
    AccountStatus,
    TokenScope,
)


class EnsureJSONResponseOnAPIErrorMiddleware(object):

    def process_exception(self, request, exception):
        response = None
        if request.META['PATH_INFO'].startswith('/api/v2/'):
            # response from inside the API are always expected to have a JSON
            # content
            msg = '%s: %s' % (exception.__class__.__name__, unicode(exception))
            response = api_error(status=500, code='SERVER_ERROR', message=msg,
                                 error_code=ErrorCode.INTERNAL_ERROR)

        return response


def api_error(status, code, message, error_code, extra=None):
    if extra is None:
        extra = {}

    extra_data = {}
    for k, v in extra.iteritems():
        if isinstance(v, list):
            v = [unicode(i) for i in v]
        elif isinstance(v, basestring):
            v = unicode(v)
        extra_data[k] = v

    data = {
        'code': code,
        'message': unicode(message),
        'extra': extra_data,
    }

    # check if we're going to use the new style error format
    new_style = extra_data.pop('new_style', False)
    if new_style:
        error_list_data = {
            'code': error_code,
            'message': data['message'],
        }
        if data['extra']:
            error_list_data['extra'] = data['extra']
        data['error_list'] = [error_list_data]

    response = HttpResponse(content=json.dumps(data), status=status,
                            content_type='application/json; charset=utf-8')
    response['Vary'] = 'Accept'
    return response


class ErrorCode:
    ACCOUNT_NOT_READY = 'account-not-ready'
    BAD_REQUEST = 'bad-request'
    INTERNAL_ERROR = 'internal-server-error'
    INVALID_DATA = 'invalid-data'
    INVALID_CREDENTIALS = 'invalid-credentials'
    INVALID_REQUEST = 'invalid-request'
    PERMISSION_REQUIRED = 'permission-required'
    REQUIRED = 'field-required'
    RESOURCE_NOT_FOUND = 'resource-not-found'
    RESOURCE_NOT_READY = 'resource-not-ready'
    TWOFACTOR_REQUIRED = 'twofactor-required'

    def __init__(self):
        self._codes = {getattr(self, code) for code in dir(self)
                       if code[0] != "_" and code != 'is_valid'}

    def is_valid(self, code):
        return code in self._codes


ERRORS = dict(
    INVALID_DATA=(400, _("Invalid request data"), ErrorCode.INVALID_DATA),

    CAPTCHA_REQUIRED=(
        401, _('A captcha challenge is required to complete the request.'),
        ErrorCode.REQUIRED),

    INVALID_CREDENTIALS=(
        401, _('Provided email/password is not correct.'),
        ErrorCode.INVALID_CREDENTIALS),

    TWOFACTOR_REQUIRED=(
        401, _("2-factor authentication required."),
        ErrorCode.TWOFACTOR_REQUIRED),

    ACCOUNT_SUSPENDED=(
        403, _('Your account has been suspended. Please contact login '
               'support to re-enable it'),
        ErrorCode.ACCOUNT_NOT_READY),

    ACCOUNT_DEACTIVATED=(
        403, _('Your account has been deactivated. To reactivate it, '
               'please reset your password'),
        ErrorCode.ACCOUNT_NOT_READY),

    EMAIL_INVALIDATED=(
        403, _("This email address has been invalidated. Please contact "
               "login support."),
        ErrorCode.ACCOUNT_NOT_READY),

    CAN_NOT_RESET_PASSWORD=(
        403, _('Can not reset password. Please contact login support'),
        ErrorCode.INVALID_REQUEST),

    CAPTCHA_FAILURE=(
        403, _('Failed response to captcha challenge.'),
        ErrorCode.RESOURCE_NOT_FOUND),

    TOO_MANY_TOKENS=(
        403, _('Too many non-consumed tokens exist. Further '
               'token creation is not allowed until existing '
               'tokens are consumed.'),
        ErrorCode.RESOURCE_NOT_READY),

    TWOFACTOR_FAILURE=(
        403, _("The provided 2-factor key is not recognised."),
        ErrorCode.INVALID_CREDENTIALS),

    PASSWORD_POLICY_ERROR=(
        403, _("Your password doesn't meet security constraints."),
        ErrorCode.INVALID_CREDENTIALS),

    FORBIDDEN=(
        403, _("Insufficient permissions."),
        ErrorCode.PERMISSION_REQUIRED),

    RESOURCE_NOT_FOUND=(
        404, _("The resource requested was not found."),
        ErrorCode.RESOURCE_NOT_FOUND),

    ALREADY_REGISTERED=(
        409, _("The email address is already registered"),
        ErrorCode.RESOURCE_NOT_READY),

    TOO_MANY_REQUESTS=(
        429, _('Too many requests. Please try again later.'),
        ErrorCode.BAD_REQUEST),

    CAPTCHA_ERROR=(
        502, _('Unable to get a valid response from reCaptcha service'),
        ErrorCode.INTERNAL_ERROR),

    SERVER_ERROR=(
        500, _('internal server error'),
        ErrorCode.INTERNAL_ERROR)
)


class Errors(object):
    def wrap(self, code):
        status, message, error_code = ERRORS[code]

        def wrapper(**extra):
            return api_error(status, code, message, error_code, extra)
        wrapper.__name__ = code
        return wrapper

    def __getattr__(self, attr):
        if attr in ERRORS:
            return self.wrap(attr)
        else:
            raise AttributeError


errors = Errors()


def _get_account_status_text(status):
    return unicode(AccountStatus._verbose[status])


def get_minimal_accounts_data(accounts):
    """Return the minimal non-private data for a sequence of accounts."""
    Account.objects.prefetch_person(accounts)
    # Given that we've prefetched related EmailAddress rows already, this is
    # equivalent to account.is_verified for each account: an account is
    # verified if it has at least one PREFERRED or VALIDATED email address.
    is_verified = {
        account: (
            account.preferredemail is not None and
            account.preferredemail.is_verified)
        for account in accounts
    }
    return {
        account.openid_identifier: {
            'href': reverse('api-account', args=(account.openid_identifier,)),
            'openid': account.openid_identifier,
            'verified': is_verified[account],
            'username': account.person_name,
        }
        for account in accounts
    }


def get_minimal_account_data(account):
    """Return the minimal non-private data for the account."""
    return get_minimal_accounts_data([account])[account.openid_identifier]


def get_accounts_data(accounts, limit=10, expand=False, include_tokens=False):
    """Get the relevant data from a sequence of accounts to be serialized.

    Emails and tokens will be limited to the latest 10 results for each
    account, where latest for email means creation date, and for tokens
    means updated date.

    """
    data = get_minimal_accounts_data(accounts)

    for account in accounts:
        # Don't use helpers such as Account.oauth_tokens or directly slice
        # QuerySets here, since that would invalidate previous
        # prefetch_related calls.  Iterating over the whole
        # EmailAddress/Token sets shouldn't be too bad in this case.
        emails = sorted(
            account.emailaddress_set.all(),
            key=attrgetter('date_created'), reverse=True)[:limit]
        data[account.openid_identifier].update({
            'email': (
                None if account.preferredemail is None
                else account.preferredemail.email),
            'displayname': account.displayname,
            'status': _get_account_status_text(account.status),
            'emails': [get_email_data(e, expand=expand) for e in emails],
        })
        if include_tokens:
            tokens = sorted(
                (token for token in account.token_set.all()
                 if token.scope == TokenScope.VERSION_2),
                key=attrgetter('date_updated'), reverse=True)[:limit]
            data[account.openid_identifier]['tokens'] = [
                get_token_data(t, expand=expand) for t in tokens]

    return data


# When fetching a single account, include_tokens defaults to True for
# historical reasons, although the tokens should now be mostly unused.
def get_account_data(account, limit=10, expand=False, include_tokens=True):
    """Get the relevant data from an account to be serialized.

    Emails and tokens will be limited to the latest 10 results, where latest
    for email means creation date, and for tokens means updated date.

    """
    return get_accounts_data(
        [account], limit=limit, expand=expand,
        include_tokens=include_tokens)[account.openid_identifier]


def get_email_data(email, expand=True):
    result = dict(
        verified=email.is_verified,
        href=reverse('api-email', args=(email.email,)),
    )
    if expand:
        result.update(dict(
            email=email.email,
            date_created=email.date_created,
            username=email.account.person_name,
        ))
    return result


def get_token_data(token, expand=True):
    result = dict(
        token_name=token.name,
        href=reverse('api-token', args=(token.pk,)),
    )
    if expand:
        result.update(dict(
            date_created=token.date_created,
            date_updated=token.date_updated,
            consumer_key=token.consumer.key,
            consumer_secret=token.consumer.secret,
            token_key=token.key,
            token_secret=token.secret,
        ))
    return result


def assert_response_error_format(resp):
    """Check that the error format is ok."""
    if resp.status_code < 400:
        # not checked
        return

    if resp.status_code >= 500 and resp['content-type'] == 'text/plain':
        # simple crashing
        return

    # we have now a 4xx response or a 5xx with interesting body, verify it!
    assert resp['content-type'] == 'application/json; charset=utf-8', (
        "bad CT: " + repr(resp['content-type']))
    content = json.loads(resp.content)

    error_list = content['error_list']
    assert error_list, "Empty error list"

    ec = ErrorCode()
    for item in error_list:
        code = item['code']
        assert ec.is_valid(code), "Invalid code: %r" % (code,)

        message = item['message']
        assert isinstance(message, basestring), "message must be a string"

        if 'extra' in item:
            assert isinstance(item['extra'], dict), (
                "extra info must be a dict")


def install_error_format_checker(testcase):
    # put the ApiResource import here to avoid the following error
    # AttributeError: 'module' object has no attribute 'test_handlers'
    from api.v20.auth import ApiResource
    """Install an error format checker in the TestCase instance."""
    orig_method = ApiResource.__call__

    def mitm(*a, **k):
        """Send resource/handler response for proper format assertion."""
        resp = orig_method(*a, **k)
        try:
            assert_response_error_format(resp)
        except Exception as err:
            testcase.fail("[Response Format Error] %s(%s)" % (
                err.__class__.__name__, err))
        return resp

    testcase.patch(ApiResource, '__call__', new=mitm)
