import logging

from email.header import Header
from email.utils import formataddr
from urlparse import urljoin

from django.conf import settings
from django.core.mail import send_mail
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils.timezone import now
from django.utils.translation import ugettext_lazy as _

from identityprovider.models import AuthToken, EmailAddress
from identityprovider.models.const import AuthTokenType, EmailStatus
from identityprovider.utils import use_root_url_scheme


logger = logging.getLogger(__name__)


EMAIL_TEMPLATES = {
    'forgot_password': {
        'subject': _('Forgotten Password'),
        'template': 'email/forgottenpassword.txt',
    },
    'password_leak': {
        'subject': _('You need to change your password'),
        'template': {
            'adobe': 'email/password_leak/adobe.txt',
        },
    }
}


def format_address(address, name=None):
    """Format a name and address to be used as an email header.

    >>> format_address('foo@bar.com', 'Name')
    'Name <foo@bar.com>'
    >>> format_address('foo@bar.com', '')
    'foo@bar.com'
    >>> format_address(u'foo@bar.com')
    'foo@bar.com'

    It handles unicode and characters that need quoting as well.

    Really long names doesn't get folded, since we're not constructing
    an e-mail header here.

    """
    if name is None:
        name = u''
    elif isinstance(name, str):
        name = name.decode('utf-8')

    if isinstance(address, str):
        address = address.decode('utf-8')

    name = str(Header(name, 'utf8'))
    # Using Header to encode the name has the side-effect that long
    # names are folded, so let's unfold it again.
    name = ''.join(name.splitlines())
    return formataddr((name, address))


def send_templated_email(subject, template, context, email, from_address=None):
    msg = render_to_string(template, context=context)

    if from_address is None:
        from_address = format_address(settings.NOREPLY_FROM_ADDRESS)
    if not isinstance(email, (tuple, list)):
        email = [email]
    return send_mail(
        subject,
        msg,
        from_address,
        [format_address(e) for e in email]
    )


def send_fancy_email(subject, template, context, email, from_address=None):
    if from_address is None:
        from_address = format_address(
            settings.NOREPLY_FROM_ADDRESS,
            settings.NOREPLY_FROM_NAME,
        )
    subject = u"%s: %s" % (settings.NOREPLY_FROM_NAME, subject)
    return send_templated_email(subject, template, context, email,
                                from_address)


def _should_add_invalidation_link(email):
    emails = EmailAddress.objects.filter(email__iexact=email)
    result = (emails.count() == 0 or
              emails.filter(status=EmailStatus.NEW).count() > 0)
    return result


def _context_for_email_request(root_url, account, email, token_type,
                               redirection_url, requester_email=None,
                               oid_token=None,
                               **kwargs):
    """
    Build and return a context (dictionary) for rendering an e-mail template.

    :returns: The context dictionary, one AuthToken representing the token
        to confirm the action indicated in the email, and optionally, one
        AuthToken representing an action invalidation token. The latter will be
        None if no invalidation token was required.
    """
    token = AuthToken.objects.create(
        requester=account, requester_email=requester_email,
        email=email, token_type=token_type,
        redirection_url=redirection_url, **kwargs)

    name = getattr(account, 'displayname', kwargs.get('displayname'))
    token_path = token.get_absolute_url()
    if oid_token:
        token_path = oid_token + token_path
    root_url = use_root_url_scheme(root_url)
    token_url = urljoin(root_url, token_path)
    context = {
        'requester': name,
        'requester_email': token.requester_email,
        'toaddress': token.email,
        'token': token.raw_token,
        'token_url': token_url,
    }

    # only send invalidation links for those email addresses that either:
    # * are unknown to the system, or
    # * are in the system but are not validated
    invalidate_email_token = None
    if _should_add_invalidation_link(email):
        invalidate_email_token = AuthToken.objects.create(
            requester=account, requester_email=requester_email,
            email=email, token_type=AuthTokenType.INVALIDATEEMAIL,
            redirection_url=redirection_url)
        root_url = use_root_url_scheme(root_url)
        token_url = urljoin(
            root_url, invalidate_email_token.get_absolute_url())
        context['invalidate_email_url'] = token_url

    return context, token, invalidate_email_token


def send_impersonation_email(root_url, email):
    """Send an email to user warning of attempted registration"""
    root_url = use_root_url_scheme(root_url)
    url = urljoin(root_url, reverse('forgot_password'))
    context = {
        'forgotten_password_url': url, 'to_email': email,
    }
    send_fancy_email(
        _("Warning"), 'email/impersonate-warning.txt', context, email)


def send_new_user_email(root_url, account, email, redirection_url=None,
                        platform='all', oid_token=None):
    """
    Send an email to a new user.

    :returns: two elements: the first is an AuthToken and the second
        can be either AuthToken or None. These are the objects representing
        the tokens which will be sent in the e-mail. Note the return
        values are mainly used for testing and can be safely ignored if
        all you care about is sending out the e-mail.
    """

    context, token, invalidate_email_token = _context_for_email_request(
        root_url, account, email, AuthTokenType.VALIDATEEMAIL, redirection_url,
        oid_token=oid_token)

    if platform != 'desktop':
        template = 'email/welcome.txt'
    else:
        template = 'email/desktop-newuser.txt'

    send_fancy_email(
        _('Finish your registration'), template, context, email)
    return token, invalidate_email_token


def get_email_template(reason):
    if reason.startswith('password_leak'):
        reason, sep, source = reason.partition('|')

    data = EMAIL_TEMPLATES.get(reason, None)
    if data is None:
        return

    subject = data['subject']
    template = data['template']
    if reason == 'password_leak':
        template = template.get(source, 'email/password_leak/default.txt')

    return {'subject': subject, 'template': template}


def send_password_reset_email(root_url, account, email, redirection_url=None,
                              reason='forgot_password'):
    """
    Send an email with password reset instructions.

    :returns: two elements: the first is an AuthToken and the second
        can be either AuthToken or None. These are the objects representing
        the tokens which will be sent in the e-mail. Note the return
        values are mainly used for testing and can be safely ignored if
        all you care about is sending out the e-mail.
    """
    data = get_email_template(reason)
    if not data:
        return

    subject = data['subject']
    template = data['template']

    context, token, invalidate_email_token = _context_for_email_request(
        root_url, account, email, AuthTokenType.PASSWORDRECOVERY,
        redirection_url, requester_email=email)
    send_fancy_email(subject, template, context, email)
    return token, invalidate_email_token


def send_password_change_email(account):
    email = account.preferredemail
    if email is None:
        logger.warning(
            'Unable to find a suitable email address to send '
            'password change notification for account %s', account)
        return
    context = {
        'display_name': account.displayname
    }
    send_fancy_email(
        _("Password change notification"),
        'email/password-changed.txt',
        context=context, email=email.email)


def send_printed_codes_renewal_email(root_url, account, devices):
    email = account.preferredemail
    if email is None:
        logger.warning(
            'Unable to find a suitable email address to send OTP reminder '
            'for account %s', account)
        return
    email = email.email
    device_urls = []
    root_url = use_root_url_scheme(root_url)
    for device in devices:
        device_urls.append(
            urljoin(root_url,
                    reverse('device-print', kwargs=dict(device_id=device.id))))
    context = dict(display_name=account.displayname, device_urls=device_urls)
    send_fancy_email(
        _("Printed list of backup codes is nearly used up"),
        'email/printed-codes-nearly-exhausted-warning.txt',
        context=context, email=email)


def send_validation_email_request(root_url, account, email,
                                  redirection_url=None, oid_token=None):
    """
    Send an email asking to validate the email account.

    :returns: two elements: the first is an AuthToken and the second
        can be either AuthToken or None. These are the objects representing the
        tokens which will be sent in the e-mail. Note the return values are
        mainly used for testing and can be safely ignored if all you care about
        is sending out the e-mail.
    """
    if account.preferredemail is None:
        preferredemail_email = None
    else:
        preferredemail_email = account.preferredemail.email

    context, token, invalidate_email_token = _context_for_email_request(
        root_url, account, email, AuthTokenType.VALIDATEEMAIL, redirection_url,
        requester_email=preferredemail_email, oid_token=oid_token)
    send_fancy_email(
        _("Validate your email address"), 'email/validate-email.txt',
        context, email)
    return token, invalidate_email_token


def send_invalidation_email_notice(root_url, account, invalidated_email):
    preferredemail = account.preferredemail
    assert preferredemail is not None
    email = preferredemail.email
    context = dict(
        display_name=account.displayname, invalidated_email=invalidated_email,
        to_email=email,
    )
    if account.unverified_emails().count() > 0:
        root_url = use_root_url_scheme(root_url)
        url = urljoin(root_url, reverse('account-emails'))
        context['verify_emails_link'] = url
    send_fancy_email(
        _("The email address {email} was removed from your account").format(
            email=invalidated_email),
        'email/email-invalidated.txt', context=context, email=email)


def send_notification_to_invalidated_email_address(root_url, account,
                                                   invalidated_email):
    context = dict(
        display_name=account.displayname, invalidated_email=invalidated_email,
        to_email=invalidated_email,
    )
    send_fancy_email(
        _("The email address {email} has been invalidated").format(
            email=invalidated_email
        ),
        'email/invalidate_email_confirmation.txt',
        context=context,
        email=invalidated_email
    )


def send_preferred_changed_notification(email, new_preferred):
    send_fancy_email(
        _('E-mail change notification'), 'email/preferred-changed.txt',
        {'new_preferred': new_preferred}, email)


def send_invitation_after_password_reset(root_url, email):
    root_url = use_root_url_scheme(root_url)
    url = urljoin(root_url, reverse('new_account'))
    send_fancy_email(
        _("Password reset request"), 'email/invitation.txt',
        {'email': email, 'signup': url}, email,
    )


def send_action_required_warning(root_url, account, days_of_warning, action):
    """
    Send an email to warn the user that an action is required.

    :returns: two elements: the first is an AuthToken and the second
        can be either AuthToken or None. These are the objects representing
        the tokens which will be sent in the e-mail. Note the return
        values are mainly used for testing and can be safely ignored if
        all you care about is sending out the e-mail.
    """
    assert action in ('suspend', 'delete')

    preferredemail = account.preferredemail
    assert preferredemail is not None
    email = preferredemail.email
    context, token, invalidate_email_token = _context_for_email_request(
        root_url, account, email, AuthTokenType.VALIDATEEMAIL,
        redirection_url=None)
    root_url = use_root_url_scheme(root_url)
    context.update(dict(
        emails_url=urljoin(root_url, reverse('account-emails')),
        created=account.date_created, action=action,
        days_of_warning=days_of_warning,
    ))
    if action == 'suspend':
        subject = _('Account to be suspended - action required')
    else:
        subject = _('Account to be deleted - action required')
    send_fancy_email(
        subject, 'email/account-action-required.txt', context, email,
    )
    return token, invalidate_email_token


def send_action_applied_notice(email, display_name, days_of_warning, action):
    assert action in ('suspend', 'delete')

    if action == 'suspend':
        subject = _('Account suspended')
        template = 'email/account-suspend-applied.txt'
    else:
        subject = _('Account deleted')
        template = 'email/account-delete-applied.txt'
    context = dict(
        action=action, display_name=display_name,
        days_of_warning=days_of_warning,
    )
    send_fancy_email(subject, template, context, email)


def send_twofactor_failure_warning(account, headers):
    # scrub out HTTP_COOKIE and CSRF_COOKIE
    headers.pop('HTTP_COOKIE', None)
    headers.pop('CSRF_COOKIE', None)
    if headers:
        headers = '\n'.join('%r: %r' % i for i in sorted(headers.iteritems()))
    else:
        headers = '<no headers>'

    warnings = [
        now().isoformat(),
        'Multiple 2FA failures (%s) for %s (lp_person=%s, openid=%s).' % (
            account.twofactor_attempts, account.displayname,
            account.person_name, account.openid_identifier),
        'Headers from last request are:\n%s' % headers,
    ]
    subject = 'Multiple 2FA failures (%s) for %s' % (
        account.twofactor_attempts, account.displayname)
    msg = '\n'.join(warnings)

    from_email = settings.NOREPLY_FROM_ADDRESS
    to = [format_address(e)
          for e in settings.TWOFACTOR_FAILURE_NOTIFICATION_EMAILS]
    return send_mail(subject, msg, from_email, to)


def send_account_suspended_warning(account, reason):
    warnings = [
        now().isoformat(),
        'Account suspended for %s (lp_person=%s, openid=%s).' % (
            account.displayname, account.person_name,
            account.openid_identifier),
        'Reason: ' + reason,
    ]
    subject = 'Account suspended for %s' % account.displayname
    msg = '\n'.join(warnings)

    from_email = settings.NOREPLY_FROM_ADDRESS
    to = [format_address(e)
          for e in settings.TWOFACTOR_FAILURE_NOTIFICATION_EMAILS]
    return send_mail(subject, msg, from_email, to)
