# Copyright 2010-2018 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from __future__ import print_function

from django.conf import settings
from django.core.management.base import BaseCommand, CommandError
from django.template import loader

from identityprovider.readonly import get_server_atts, update_server


class Command(BaseCommand):
    help = ('Manage readonly mode.\n\n'
            'options --set and --clear are all mutually exclusive.\n'
            'You can only choose one at a time.')

    def add_arguments(self, parser):
        parser.add_argument(
            '--list', action='store_true', dest='list_servers', default=False,
            help='List available application servers.')
        parser.add_argument(
            '--all', action='store_true', dest='all_servers',
            default=False, help='Select all servers.')
        parser.add_argument(
            '--set', action='store_const', const='set',
            dest='action', help='Set server to read-only.')
        parser.add_argument(
            '--clear', action='store_const', const='clear',
            dest='action', help='Set server to read-write.')
        parser.add_argument('servers', metavar='server', nargs='*')

    def handle(self, *args, **options):
        servers = options.get('servers')
        list_servers = options.get('list_servers')
        all_servers = options.get('all_servers')
        action = options.get('action')

        # determine servers to act upon
        if all_servers:
            servers = [app['SERVER_ID'] for app in settings.APP_SERVERS]
            servers.sort()
        else:
            msgs = (
                'Enter at least one server, or specify the --all option.',
                'Use --list to get a list of configured servers.',
            )
            if not servers and not list_servers:
                raise CommandError('\n  '.join(msgs))

        # determine action to perform
        if action is not None:
            for server in servers:
                update_server(action, server)
        elif not list_servers:
            msg = 'Enter one of --set or --clear.'
            raise CommandError(msg)

        # list action is special as it can be combined with the other actions
        if list_servers:
            self.show_servers()

    def show_servers(self):
        """Provides a report about readonly status of all app servers."""
        atts = get_server_atts(settings.APP_SERVERS)
        print(loader.render_to_string('admin/readonly.txt', atts))
