from __future__ import absolute_import

import re

from django.conf import settings
from django.template.loader import render_to_string
try:
    from django.utils.encoding import force_text
except ImportError:
    from django.utils.encoding import force_unicode as force_text
from honeypot.middleware import (
    _HTML_TYPES,
    _POST_FORM_RE,
    HoneypotMiddleware as upstream_honeypot_middleware
)

HONEYPOT_EXEMPT_URLS = re.compile(
    '^/api/|^/[A-Za-z_]+/\+openid|^/\+openid|^/\+saml|^/saml')


class HoneypotMiddleware(upstream_honeypot_middleware):
    """Middleware that verifies a valid honeypot on all POSTs.

    Excluded are AJAX's requests and API requests.

    """

    def process_response(self, request, response):
        """Move the honeypot field from the top of the form to the bottom"""
        response = super(HoneypotMiddleware, self).process_response(
            request, response)
        try:
            content_type = response['Content-Type'].split(';')[0]
        except (KeyError, AttributeError):
            content_type = None
        if content_type in _HTML_TYPES:
            value = getattr(settings, 'HONEYPOT_VALUE', '')
            if callable(value):
                value = value()
            honey_text = render_to_string(
                'honeypot/honeypot_field.html',
                {'fieldname': settings.HONEYPOT_FIELD_NAME,
                 'value': value}).encode('utf-8')
            content = force_text(response.content).encode('utf-8')
            for form_match in _POST_FORM_RE.finditer(content):
                honey_offset = content.find(honey_text, form_match.end())
                end_form = content.find('</form>', form_match.end())
                response.content = content = str(
                    # Everything up to the old honeypot
                    content[:form_match.end()] +
                    # everything after the old honeypot, but before
                    # the end of the form
                    content[honey_offset + len(honey_text):end_form] +
                    # the honeypot
                    honey_text +
                    # rest of the page
                    content[end_form:]
                )
        return response

    def process_view(self, request, *args, **kwargs):
        response = None
        if not HONEYPOT_EXEMPT_URLS.match(request.path):
            # Only validate honeypot fields for requests that are not going
            # to our API endpoints.
            response = super(HoneypotMiddleware, self).process_view(
                request, *args, **kwargs)

        return response
