# Copyright 2011-2017 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from django.core.exceptions import ValidationError
from django.core.validators import validate_ipv46_address


class SetRemoteAddrFromForwardedFor(object):
    """Set REMOTE_ADDR based on HTTP_X_FORWARDED_FOR, if the latter is present.

    This is useful if you're sitting behind a reverse proxy that causes each
    request's REMOTE_ADDR to be set to 127.0.0.1. Note that this does NOT
    validate HTTP_X_FORWARDED_FOR. If you're not behind a reverse proxy that
    sets HTTP_X_FORWARDED_FOR automatically, do not use this middleware.

    Anybody can spoof the value of HTTP_X_FORWARDED_FOR, and because this sets
    REMOTE_ADDR based on HTTP_X_FORWARDED_FOR, that means anybody can "fake"
    their IP address. Only use this when you can absolutely trust the value of
    HTTP_X_FORWARDED_FOR.

    """

    def validate_ip(self, ip):
        """"clean up" and validate an ip address."""
        ip = ip.strip()
        try:
            validate_ipv46_address(ip)
            return ip
        except ValidationError:
            return None

    def process_request(self, request):
        real_ip = request.META.get('HTTP_X_FORWARDED_FOR', None)
        if real_ip is not None:
            # HTTP_X_FORWARDED_FOR can be a comma-separated list of IPs. The
            # client's IP will be the first valid one.
            for clean_ip in (
                    self.validate_ip(ip) for ip in real_ip.split(",")):
                if clean_ip is not None:
                    request.META['REMOTE_ADDR'] = clean_ip
                    break
