import urllib
import urlparse

from itertools import count

from django.contrib.auth.hashers import make_password
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.contrib.messages.storage.fallback import FallbackStorage
from django.contrib.sessions.middleware import SessionMiddleware
from django.db import IntegrityError
from django.test import RequestFactory
from django.utils.timezone import now
from oath import hotp, totp
from oauthlib import oauth1
from openid.message import IDENTIFIER_SELECT

from timeline import Timeline

from identityprovider.const import (
    PERSON_VISIBILITY_PRIVATE_MEMBERSHIP,
    PERSON_VISIBILITY_PUBLIC,
)
from identityprovider.models import (
    APIUser,
    Account,
    AuthToken,
    AuthenticationDevice,
    EmailAddress,
    LPOpenIdIdentifier,
    LeakedCredential,
    OpenIDRPConfig,
    Person,
    TeamParticipation,
)
from identityprovider.models.const import (
    AccountCreationRationale,
    AuthTokenType,
    EmailStatus,
    TokenScope,
)
from identityprovider.tests import (
    DEFAULT_API_PASSWORD,
    DEFAULT_USER_PASSWORD,
)


class SSOObjectFactory(object):
    """A factory for creating model objects for tests."""

    request_factory = RequestFactory()

    def __init__(self):
        super(SSOObjectFactory, self).__init__()
        self.counter = count()

    def get_unique_string(self, prefix=None):
        if prefix is None:
            prefix = "generic-string"
        return prefix + str(self.get_unique_integer())

    def get_unique_integer(self):
        """Return a unique integer for the test run."""
        return self.counter.next()

    def make_apiuser(self, username=None, password=DEFAULT_API_PASSWORD):
        if username is None:
            username = self.get_unique_string(prefix='apiuser-')
        user, created = APIUser.objects.get_or_create(username=username)
        user.set_password(password)
        user.save()
        return user

    def make_account(
            self, displayname=None, email=None, password=DEFAULT_USER_PASSWORD,
            creation_rationale=None, salt=None, password_encrypted=False,
            email_validated=True, openid_identifier=None, status=None,
            date_created=None, teams=None, twofactor_required=False,
            permissions=None, is_superuser=False, is_staff=True,
            warn_about_backup_device=True, **kwargs):
        if displayname is None:
            displayname = self.get_unique_string(prefix='Test Account ')
        if email is None:
            email = self.make_email_address()
        account = Account.objects.create_account(
            displayname, email, password=password,
            creation_rationale=creation_rationale,
            salt=salt,
            password_encrypted=password_encrypted,
            email_validated=email_validated,
            openid_identifier=openid_identifier,
            twofactor_required=twofactor_required,
            warn_about_backup_device=warn_about_backup_device,
            **kwargs)
        if status is not None:
            account.status = status
        if date_created is not None:
            account.date_created = date_created

        account.save()

        if teams:
            for t in teams:
                team = Person.objects.filter(name=t)
                if len(team) > 0:
                    team = team[0]
                else:
                    team = self.make_team(t)
                self.add_account_to_team(account, team)

        if permissions:
            for p in permissions:
                self.add_permission_to_account(account, p)

        account.user.is_superuser = is_superuser
        account.user.is_staff = is_staff
        account.user.save()

        return account

    def make_account_token(self, account, email=None):
        if email is None:
            email = account.preferredemail
        token = self.make_authtoken(
            requester=account, requester_email=email, email=email,
            token_type=AuthTokenType.VALIDATEEMAIL)
        return token

    def make_person(self, name=None, displayname=None, account=None, **kwargs):
        i = self.get_unique_integer()
        if name is None:
            name = 'person-%s' % i
        if displayname is None:
            displayname = 'Person %s' % i

        person = None
        msg = 'Getting an available lp_account id (%s used, got error "%s").'
        while person is None:
            try:
                # lp_account binds Person and LPOpenIdIdentifier
                person = Person.objects.create(
                    name=name, displayname=displayname, lp_account=i, **kwargs)
            except IntegrityError as e:
                print(msg % (i, e))
                i += 1

        # LPOpenIdIdentifier.identifier should equal account.openid_identifier
        if account is not None:
            LPOpenIdIdentifier.objects.create(
                identifier=account.openid_identifier,
                lp_account=person.lp_account)
            try:
                del account.person
            except AttributeError:
                pass

        return person

    def make_team(self, name=None, private=False, owner=None):
        if private:
            visibility = PERSON_VISIBILITY_PRIVATE_MEMBERSHIP
        else:
            visibility = PERSON_VISIBILITY_PUBLIC
        if owner is None:
            owner = self.make_person()
        if name is None:
            name = self.get_unique_string(prefix='team-')
        displayname = 'Team ' + name
        team = self.make_person(
            name, displayname=displayname, visibility=visibility,
            teamowner=owner.id)
        return team

    def add_account_to_team(self, account, team):
        if account.person_in_team(team.name):
            return

        if account.person is None:
            person = self.make_person(account=account)
            assert account.person == person

        TeamParticipation.objects.create(person=account.person, team=team)

    def add_permission_to_account(self, account, permission_name):
        content_type = ContentType.objects.get_for_model(Account)
        permission, _ = Permission.objects.get_or_create(
            codename=permission_name, content_type=content_type)
        account.user.user_permissions.add(permission)

    def make_email_address(self, prefix='email-', domain='example.com'):
        return "%s@%s" % (self.get_unique_string(prefix=prefix), domain)

    def make_email_for_account(self, account, email=None, status=None):
        if email is None:
            email = self.make_email_address()

        if status is None:
            status = EmailStatus.VALIDATED
        elif status is EmailStatus.PREFERRED:
            # make sure there is no other preferred email address
            # or reset it if there is
            account_emails = EmailAddress.objects.filter(account=account,
                                                         status=status)
            account_emails.update(status=EmailStatus.VALIDATED)

        email, _ = EmailAddress.objects.get_or_create(
            email=email, defaults={'account': account, 'status': status})
        if status == EmailStatus.PREFERRED:
            account.preferredemail = email
        return email

    def make_device(
            self, account=None, name='MyDevice', key=None, counter=0,
            device_type=None, oath_mode=None):
        if account is None:
            account = self.make_account()
        if key is None:
            key = ''
        device = AuthenticationDevice.objects.create(
            account=account, name=name, key=key, counter=counter,
            device_type='{}:{}'.format(
                device_type if device_type else '',
                oath_mode if oath_mode else ''))
        return device

    def make_next_otp(self, device):
        if device.is_totp():
            oath_token = totp(device.key)
        else:
            oath_token = hotp(device.key, device.counter)
        return oath_token

    def make_oauth_token(
            self, account=None, token_name=None, scope=TokenScope.VERSION_2):
        """Create a new set of OAuth token and consumer creds."""
        if account is None:
            account = self.make_account()
        if token_name is None:
            token_name = self.get_unique_string(prefix='token-name')
        return account.create_oauth_token(token_name, scope=scope)

    def make_authtoken(self, token_type=None, email=None, redirection_url=None,
                       displayname=None, password=None, requester=None,
                       requester_email=None, date_created=None):
        if token_type is None:
            token_type = AuthTokenType.VALIDATEEMAIL
        if date_created is None:
            date_created = now()
        token = AuthToken.objects.create(
            token_type=token_type, email=email,
            redirection_url=redirection_url, displayname=displayname,
            password=password, requester=requester,
            requester_email=requester_email, date_created=date_created)
        return token

    def make_leaked_credential(self, email, password, source='source',
                               date_leaked=None):
        if date_leaked is None:
            date_leaked = now().date()
        leak, _ = LeakedCredential.objects.get_or_create(
            email=email, source=source, password=make_password(password),
            when_leaked=date_leaked)
        return leak

    def make_openid_rp_config(
            self, trust_root=None, displayname=None, creation_rationale=None,
            twofactor_required=False):
        if trust_root is None:
            trust_root = 'http://example-%s.com' % self.get_unique_string()
        if displayname is None:
            displayname = "Example OpenIDRPConfig instance at %s" % trust_root
        if creation_rationale is None:
            creation_rationale = AccountCreationRationale.UNKNOWN

        return OpenIDRPConfig.objects.create(
            trust_root=trust_root, displayname=displayname,
            creation_rationale=creation_rationale,
            twofactor_required=twofactor_required)

    def make_oauth_authorization_from_token(
            self, url, token, http_method='GET', parameters=None,
            realm='Some client', plaintext=False, base_url='http://testserver',
            as_query_string=False, consumer=None, nonce=None):

        if not isinstance(token, dict):
            if consumer is None:
                consumer = token.consumer
            token = {
                'token_key': token.key,
                'token_secret': token.secret,
                'consumer_key': consumer.key,
                'consumer_secret': consumer.secret,
            }

        if not url.startswith('http'):
            url = urlparse.urljoin(base_url, url)

        if parameters is not None:

            # A parameter may contain non-ascii data
            parameters = {
                k: v.encode('utf-8') for k, v in parameters.items()}
            url += '?' + urllib.urlencode(parameters)

        # Need to mangle with the URL a bit here, because:
        # reverse() will escape chars like @, and to build the proper
        # signature we need to quote only the query string, not the paths
        parts = urlparse.urlparse(url)
        url = urlparse.urlunparse((
            parts.scheme, parts.netloc, urllib.unquote(parts.path),
            parts.params, parts.query, parts.fragment))

        if plaintext:
            signature_method = oauth1.SIGNATURE_PLAINTEXT
        else:
            signature_method = oauth1.SIGNATURE_HMAC

        if as_query_string:
            signature_type = oauth1.SIGNATURE_TYPE_QUERY
        else:
            signature_type = oauth1.SIGNATURE_TYPE_AUTH_HEADER

        client = oauth1.Client(
            client_key=token['consumer_key'],
            client_secret=token['consumer_secret'],
            resource_owner_key=token['token_key'],
            resource_owner_secret=token['token_secret'],
            signature_method=signature_method,
            signature_type=signature_type,
            realm=realm, nonce=nonce,
        )
        uri, header, body = client.sign(url, http_method)
        return uri, header, body

    def make_authorization_header_from_token(self, url, token, **kwargs):
        kwargs['as_query_string'] = False
        uri, header, body = self.make_oauth_authorization_from_token(
            url, token, **kwargs)
        return {'HTTP_AUTHORIZATION': header['Authorization']}

    def make_request(
            self, path='/', method='GET', user=None, session=False,
            messages=False, timeline=False,
            preseed_cookie_check=False, **kwargs):
        request = getattr(self.request_factory, method.lower())(path, **kwargs)
        if user is not None:
            request.user = user
        if session:
            # add session to request
            middleware = SessionMiddleware()
            middleware.process_request(request)
            request.session.save()
        if messages:
            request._messages = FallbackStorage(request)
        if timeline:
            request.META['timeline.timeline'] = Timeline()
        if preseed_cookie_check:
            # Inject the test cookie to avoid being redirected to /+cookie.
            # This avoids needing disable_cookie_check=True.
            # See the preseed_cookie_check method for more information.
            request.COOKIES['C'] = '1'
        return request

    def make_authenticated_request(self, token, path='/', method='GET'):
        auth_header = self.make_authorization_header_from_token(
            url=path, token=token, http_method=method)
        return self.make_request(path, method, **auth_header)

    def make_openid_request_params(
            self, trust_root='http://localhost/',
            return_to='http://localhost/'):
        request = {
            'openid.mode': 'checkid_setup',
            'openid.trust_root': trust_root,
            'openid.return_to': return_to,
            'openid.identity': IDENTIFIER_SELECT,
        }
        return request
