# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import json
import os
import shutil
import stat
import tempfile

import requests
import responses
from django.conf import settings
from django.urls import reverse

from identityprovider.readonly import (
    ReadOnlyManager,
    _remote_req,
    get_server_atts,
    update_server,
)
from identityprovider.tests import DEFAULT_USER_PASSWORD
from identityprovider.tests.utils import SSOBaseTestCase


class ReadOnlyBaseTestCase(SSOBaseTestCase):
    """A base TestCase for backing up and restoring our readonly settings"""

    def setUp(self):
        super(ReadOnlyBaseTestCase, self).setUp()
        # XXX, 2013-03-29, nessita: Do not remove this patching since the
        # READ_ONLY_MODE is tweak in tests, and it should be guaranteed the
        # readonlymode stays False
        overrides = self.settings(
            READ_ONLY_MODE=False,
            READONLY_SECRET='testsecret',
            DBFAILOVER_FLAG_DIR=tempfile.mkdtemp(),
        )
        overrides.enable()
        self.addCleanup(overrides.disable)
        self.addCleanup(shutil.rmtree, settings.DBFAILOVER_FLAG_DIR, True)

        self.rm = ReadOnlyManager()


class RemoteRequestTestCase(SSOBaseTestCase):
    msg = 'hello'
    host = 'myhost'
    scheme = 'https'
    vhost = 'http://foobar.baz'

    @responses.activate
    def test_plain_remote_req(self):
        responses.add(
            responses.POST, 'http://%s/readonlydata' % self.host,
            body=self.msg)
        server = {'host': self.host}
        self.assertEqual(self.msg, _remote_req(**server))
        # This tests the PreparedRequest, which is before urllib3 inserts a
        # default Host header.
        self.assertNotIn('Host', responses.calls[0].request.headers)

    @responses.activate
    def test_https_remote_req(self):
        responses.add(
            responses.POST, '%s://%s/readonlydata' % (self.scheme, self.host),
            body=self.msg)
        server = {'host': self.host, 'scheme': self.scheme}
        self.assertEqual(self.msg, _remote_req(**server))
        # This tests the PreparedRequest, which is before urllib3 inserts a
        # default Host header.
        self.assertNotIn('Host', responses.calls[0].request.headers)

    @responses.activate
    def test_vhost_remote_req(self):
        responses.add(
            responses.POST, '%s://%s/readonlydata' % (self.scheme, self.host),
            body=self.msg)
        server = {
            'host': self.host,
            'scheme': self.scheme,
            'virtual_host': self.vhost,
        }
        self.assertEqual(self.msg, _remote_req(**server))
        self.assertEqual(
            self.vhost, responses.calls[0].request.headers['Host'])

    @responses.activate
    def test_remote_req_error(self):
        responses.add(
            responses.POST, 'http://%s/readonlydata' % self.host,
            body=requests.RequestException('error'))
        server = {'host': self.host}
        self.assertEqual(None, _remote_req(**server))


class ReadOnlyFlagFilesTestCase(ReadOnlyBaseTestCase):

    def test_flag_files_in_right_directory(self):
        self.rm.set_readonly()
        flags = os.listdir(settings.DBFAILOVER_FLAG_DIR)
        self.assertTrue('db.readonly' in flags)

    def test_readonly_flag_files_have_right_mode(self):
        self.rm.set_readonly()
        mode = os.stat(self.rm.readonly_marker_file)[stat.ST_MODE]
        self.assertTrue(mode & stat.S_IRGRP)
        self.assertTrue(mode & stat.S_IWGRP)


class ReadOnlyDataTestCase(ReadOnlyBaseTestCase):

    url = reverse('readonly-data')

    def setUp(self):
        super(ReadOnlyDataTestCase, self).setUp()
        self.mock_logger = self.patch(
            'identityprovider.views.readonly.logger')

    def test_requires_post(self):
        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 405)

        response = self.client.delete(self.url)
        self.assertEqual(response.status_code, 405)

        response = self.client.put(self.url)
        self.assertEqual(response.status_code, 405)

    def test_missing_secret(self):
        response = self.client.post(self.url)

        self.assertEqual(response.status_code, 404)
        self.mock_logger.warning.assert_called_once_with(
            'readonly_data request received with incorrect secret %r', None)

    def test_incorrect_secret(self):
        assert len(settings.READONLY_SECRET) > 0
        secret = settings.READONLY_SECRET * 2
        response = self.client.post(self.url, data=dict(secret=secret))

        self.assertEqual(response.status_code, 404)
        self.mock_logger.warning.assert_called_once_with(
            'readonly_data request received with incorrect secret %r', secret)

    def test_correct_secret(self):
        secret = settings.READONLY_SECRET
        response = self.client.post(self.url, data=dict(secret=secret))

        self.assertEqual(response.status_code, 200)
        response = json.loads(response.content)
        self.assertEqual(response, dict(readonly=False))

    def test_readonly_set_readonly(self):
        post = {'secret': settings.READONLY_SECRET, 'action': 'set'}
        response = self.client.post('/readonlydata', post)
        data = json.loads(response.content)
        self.assertTrue(data['readonly'])

    def test_readonly_clear_readonly(self):
        post = {'secret': settings.READONLY_SECRET, 'action': 'clear'}
        response = self.client.post('/readonlydata', post)
        data = json.loads(response.content)
        self.assertFalse(data['readonly'])


class ReadOnlyViewsTestCase(ReadOnlyBaseTestCase):

    def login_with_staff(self):
        account = self.factory.make_account(password='password', is_staff=True)
        assert account.user.is_staff
        assert self.client.login(
            username=account.preferredemail.email, password='password')

    def test_readonly_admin(self):
        self.login_with_staff()

        new_setting = [
            {'SERVER_ID': 'localhost', 'SCHEME': 'http',
             'HOST': 'localhost', 'VIRTUAL_HOST': '',
             'PORT': '8000'}
        ]
        with self.settings(APP_SERVERS=new_setting):
            r = self.client.get('/readonly')

        self.assertTemplateUsed(r, 'admin/readonly.html')
        expected = {'appservers': [{'name': 'localhost', 'reachable': False}],
                    'clear_all_readonly': False,
                    'set_all_readonly': False}
        for item in r.context:
            for key, value in expected.items():
                self.assertEqual(item[key], value)

    def test_get_server_atts_server_unreachable(self):
        servers = [{'SERVER_ID': 'localhost', 'SCHEME': 'http',
                    'HOST': 'localhost', 'VIRTUAL_HOST': '', 'PORT': '8000'}]
        expected = {'appservers': [{'name': 'localhost', 'reachable': False}],
                    'clear_all_readonly': False,
                    'set_all_readonly': False}
        atts = get_server_atts(servers)
        self.assertEqual(atts, expected)

    @responses.activate
    def test_get_server_atts_data_error(self):
        responses.add(
            responses.POST, 'http://localhost:8000/readonlydata', body='{')

        servers = [{'SERVER_ID': 'localhost', 'SCHEME': 'http',
                    'HOST': 'localhost', 'VIRTUAL_HOST': '', 'PORT': '8000'}]
        expected = {'appservers': [{'name': 'localhost', 'reachable': False}],
                    'clear_all_readonly': False,
                    'set_all_readonly': True}
        atts = get_server_atts(servers)
        self.assertEqual(atts, expected)

    @responses.activate
    def test_get_server_atts_readonly(self):
        responses.add(
            responses.POST, 'http://localhost:8000/readonlydata',
            json={'readonly': True})

        servers = [{'SERVER_ID': 'localhost', 'SCHEME': 'http',
                    'HOST': 'localhost', 'VIRTUAL_HOST': '', 'PORT': '8000'}]
        expected = {'appservers': [{'name': 'localhost', 'reachable': True,
                                    'readonly': True}],
                    'clear_all_readonly': True,
                    'set_all_readonly': False}
        atts = get_server_atts(servers)
        self.assertEqual(atts, expected)

    def test_readonly_confirm_get(self):
        self.login_with_staff()

        r = self.client.get('/readonly/localhost/set')
        self.assertTemplateUsed(r, 'admin/readonly_confirm.html')
        self.assertEqual(r.context['appserver'], 'localhost')
        self.assertEqual(r.context['action'], 'set')

    @responses.activate
    def test_readonly_confirm_post(self):
        responses.add(
            responses.POST, 'http://localhost:8000/readonlydata', json={})
        self.login_with_staff()

        new_setting = [
            {'SERVER_ID': 'localhost', 'SCHEME': 'http',
             'HOST': 'localhost', 'VIRTUAL_HOST': '',
             'PORT': '8000'}
        ]
        with self.settings(APP_SERVERS=new_setting):
            r = self.client.post('/readonly/localhost/set')
        data = [call.request.body for call in responses.calls]
        self.assertEqual(data, [
            "action=set&secret=%s" % settings.READONLY_SECRET
        ])

        self.assertRedirects(r, '/readonly')

    @responses.activate
    def test_update_server_all_appservers(self):
        responses.add(
            responses.POST, 'http://localhost:8000/readonlydata', json={})
        new_setting = [
            {'SERVER_ID': 'localhost', 'SCHEME': 'http',
             'HOST': 'localhost', 'VIRTUAL_HOST': '', 'PORT': '8000'},
            {'SERVER_ID': 'otherhost', 'SCHEME': 'http',
             'HOST': 'otherhost', 'VIRTUAL_HOST': '', 'PORT': '8000'},
        ]
        with self.settings(APP_SERVERS=new_setting):
            update_server('set')
        data = [call.request.body for call in responses.calls]
        self.assertEqual(data, [
            "action=set&secret=%s" % settings.READONLY_SECRET,
            "action=set&secret=%s" % settings.READONLY_SECRET
        ])

    @responses.activate
    def test_update_server_one_appserver(self):
        responses.add(
            responses.POST, 'http://localhost:8000/readonlydata', json={})
        new_setting = [
            {'SERVER_ID': 'localhost', 'SCHEME': 'http',
             'HOST': 'localhost', 'VIRTUAL_HOST': '', 'PORT': '8000'},
            {'SERVER_ID': 'otherhost', 'SCHEME': 'http',
             'HOST': 'otherhost', 'VIRTUAL_HOST': '', 'PORT': '8000'},
        ]
        with self.settings(APP_SERVERS=new_setting):
            update_server('set', 'localhost')
        data = [call.request.body for call in responses.calls]
        self.assertEqual(data, [
            "action=set&secret=%s" % settings.READONLY_SECRET
        ])

    @responses.activate
    def test_update_server_one_connection(self):
        responses.add(
            responses.POST, 'http://localhost:8000/readonlydata', json={})
        new_setting = [
            {'SERVER_ID': 'localhost', 'SCHEME': 'http',
             'HOST': 'localhost', 'VIRTUAL_HOST': '', 'PORT': '8000'},
        ]
        with self.settings(APP_SERVERS=new_setting):
            update_server('set', 'localhost')
        data = [call.request.body for call in responses.calls]
        self.assertEqual(data, [
            "action=set&secret=%s" % settings.READONLY_SECRET,
        ])

    def test_nop_db_in_readonly_mode(self):
        account = self.factory.make_account()

        self.rm.set_readonly()
        self.addCleanup(self.rm.clear_readonly)

        response = self.client.get('/+login', follow=True)
        # assert we're in readonly mode
        self.assertTrue(response.context['readonly'])
        # login
        csrf_token = response.context['request'].META['CSRF_COOKIE']
        data = {'email': account.preferredemail.email,
                'password': DEFAULT_USER_PASSWORD,
                'csrftoken': csrf_token}
        response = self.client.post('/+login', data=data)

        # assert logged in
        self.assertRedirects(response, '/')
