# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from django.template import Context, Template
from django.test import TestCase

from identityprovider.templatetags.google_analytics import expand_ga_commands
from identityprovider.templatetags.static_url import static_url

SNIPPETS = [
    None,
    '',
    'not valid json'
    '"); alert("gotcha!"); //'
    '\'); alert(\'gotcha!\'); //'
    '"Not a json list"'
    '"\'); alert(\'JSON gotcha!\'); //"'
    '["List of strings"]'
    '[[]]',
    '[["omg", function() { alert("gotcha!"); }]]'
    '[["nothing", "move along"], ["just a bit conspicuous", 42]]'
    '[["something", "with multiple\nlines"]]'
]


class GACommandsTestCase(TestCase):

    def test_invalid_snippets(self):
        for snippet in SNIPPETS:
            self.assertEqual('', expand_ga_commands(snippet))

    def test_valid_snippets(self):
        for snippet, expected in [
                ('[["foo", "bar"]]', "_gaq.push(['foo', 'bar']);"),
                ('[["foo", "bar"], ["baz", "zot"]]',
                 "_gaq.push(['foo', 'bar']);\n_gaq.push(['baz', 'zot']);"),
        ]:
            self.assertEqual(expected, expand_ga_commands(snippet))

    def test_unicode_coercion(self):
        snippet = '[["foo", "bar"]]'
        self.assertIsInstance(expand_ga_commands(snippet), unicode)


class StaticUrlTestCase(TestCase):

    def test_returns_empty_string_when_setting_missing(self):
        self.assertEqual(static_url('does_not_exist'), '')

    def test_returns_setting_when_absolute_url(self):
        url = 'http://test.com/support'
        with self.settings(TEST_SUPPORT_URL=url):
            self.assertEqual(static_url('test_support'), url)

    def test_returns_setting_when_relative_url(self):
        url = '/+relative'
        with self.settings(RELATIVE_URL=url):
            self.assertEqual(static_url('relative'), url)


class QRCodeURLTestCase(TestCase):

    def test_not_encoded(self):
        """The url string should be marked as safe as we can't escape the &"""
        qr_data = {
            'oath_mode': 'totp',
            'ident': u'UbuntuSSO/zl@beyfvift.org',
            'hex_key': 'A10EE5139362F3150BFABB7DDBEC5EE6FA2CA989',
        }

        # Render the template tag so we see the final result
        # as calling the method won't result in the string being encoded
        template = Template(
            '{% load qrcode %}{% qrcode_url mode ident hex_key %}'
        )
        context = Context(qr_data)
        result = template.render(context)

        # Ideally, we should just decode the string here and check if it
        # changes but the email address is correctly encoded,
        # so that changes if so.
        # instead, check the whole string we have generated
        expected_result = (
            u'https://chart.googleapis.com/chart?chs=250x250&'
            u'chld=L%7C0&cht=qr&chl=otpauth%3A%2F%2F%2FUbuntuSSO%2Fzl%40'
            u'beyfvift.org%3Fsecret%3DUEHOKE4TMLZRKC72XN65X3C6435CZKMJ')
        self.assertEqual(result, expected_result)
