import json
import M2Crypto

from django.core.exceptions import ValidationError
from django.db import models


def validate_json(value, field='field'):
    """Validate value is a proper JSON document."""
    try:
        value = json.loads(value)
    except ValueError as err:
        raise ValidationError(str(err))

    if not isinstance(value, dict):
        raise ValidationError(
            "'{}' must be a valid JSON object (dict).".format(field))


def validate_attributes(value):
    """Validate attributes field value."""
    validate_json(value, field='attributes')


def validate_links(value):
    """Validate links field value."""
    validate_json(value, field='links')


def validate_certificate(value):
    """Validate base64-encoded with delimiters X.509 certificate value."""
    try:
        M2Crypto.X509.load_cert_string(str(value))
    except (M2Crypto.X509.X509Error, ValueError):
        raise ValidationError(
            "'certificate' must be a PEM-formatted X.509 certificate.")


class SAMLConfig(models.Model):
    acs_url = models.TextField(unique=True)
    acs_domain = models.TextField(blank=True)
    issuer = models.TextField(
        null=True, blank=True,
        help_text=("Use the issuer to identify the SP when the AuthnRequest "
                   "doesn't contain an ACS url."))
    allowed_teams = models.TextField(
        blank=True,
        default='',
        help_text=("Comma-separated list of Launchpad team names that "
                   "should be allowed to authenticate to this RP. If "
                   "left blank, anyone can access regardless of "
                   "team membership."))
    audience = models.TextField(blank=True)
    displayname = models.TextField(blank=True)
    certificate = models.TextField(
        blank=True,
        default='',
        validators=[validate_certificate],
        help_text=("PEM-format Certificate to use for signing SAML "
                   "assertions for this RP. Include the -----BEGIN/END "
                   "CERTIFICATE----- lines."))
    enabled = models.BooleanField(default=False)
    prefer_canonical_email = models.BooleanField(default=False)
    email_pattern = models.TextField(
        default='{username}@{domain}',
        help_text=(
            'A pattern to apply to the user email address being returned '
            'to the SP. You can use {username} and {domain} for the '
            'distinct parts of the email address.'))
    attributes = models.TextField(
        blank=True,
        validators=[validate_attributes],
        help_text='This field should contain a valid json document.')
    links = models.TextField(
        blank=True,
        validators=[validate_links],
        help_text='This field should contain a valid json document.')

    def __getitem__(self, key):
        return self.__dict__[key]

    def __unicode__(self):
        return self.displayname

    def __repr__(self):
        return (
            "<SAMLConfig(displayname='{}', acs_url='{}', acs_domain='{}', "
            "audience='{}', prefer_canonical_email={}, enabled={}, "
            "email_pattern='{}', attributes='{}', links='{}')>"
            .format(self.displayname, self.acs_url, self.acs_domain,
                    self.audience, self.prefer_canonical_email, self.enabled,
                    self.email_pattern, self.attributes, self.links)
        )

    def save(self, *args, **kwargs):
        self.full_clean()
        return super(SAMLConfig, self).save(*args, **kwargs)

    def get_allowed_teams_list(self):
        """Returns a list of teams allowed to authenticate to this SP.

           The returned list can be empty, which means anybody can
           authenticate to the SP.
        """
        return [team.strip() for team in
                self.allowed_teams.split(",")
                if team.strip() != ""]
