# Copyright 2012-2016 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).
import base64
import json
import urlparse

from BeautifulSoup import BeautifulStoneSoup
from django.conf import settings
from saml2idp import codex, google_apps, salesforce, xml_render
from saml2idp.base import Processor, get_time_string
from saml2idp.exceptions import CannotHandleAssertion

from identityprovider.models import OpenIDRPConfig
from ubuntu_sso_saml.utils import (
    certificate_from_sp_config,
    get_config_from_processor,
)


__all__ = [
    'ExpensifyProcessor',
    'GoogleAppsProcessor',
    'SalesForceProcessor',
    'SalesForceAttributeProcessor',
]


class CanonicalOverrides(object):

    def _is_canonical_email(self, email):
        return email.lower().endswith('@canonical.com')

    def _is_long_form_email(self, email):
        user, _ = email.split('@')
        return '.' in user and '+' not in user

    def _get_canonical_email(self, account):
        preferred = account.preferredemail.email
        if (self._is_canonical_email(preferred) and
                self._is_long_form_email(preferred)):
            return preferred

        preferred = None
        # user has @canonical email address?
        emails = account.emailaddress_set.filter(
            email__iendswith='@canonical.com')
        for email in emails:
            # only consider emails formatted like first.last@canonical.com
            if self._is_long_form_email(email.email):
                preferred = email.email
                break
        if preferred is None and account.preferredemail in emails:
            # support preferred email if it's a @canonical.com email
            preferred = account.preferredemail.email
        return preferred

    def _determine_subject(self):
        account = self._django_request.user
        preferred = account.preferredemail.email

        # check if RP prefers @canonical emails
        trust_root = self._request_params.get('ACS_URL')
        if trust_root is not None:
            rpconfig = OpenIDRPConfig.objects.for_url(trust_root)
            if rpconfig is not None and rpconfig.prefer_canonical_email:
                canonical_email = self._get_canonical_email(account)
                if canonical_email is not None:
                    preferred = canonical_email

        self._subject = preferred


class GoogleAppsProcessor(CanonicalOverrides, google_apps.Processor):
    """
    Google Apps SAML 2.0 Processor, specific to Canonical Identity Provider.
    """


class SalesForceProcessor(CanonicalOverrides, salesforce.Processor):
    """
    SalesForce SAML 2.0 Processor, specific to Canonical Identity Provider.
    """


class SalesForceAttributeProcessor(CanonicalOverrides, salesforce.Processor):
    """
    SalesForce SAML 2.0 Processor, specific to Canonical Identity Provider.
    Adds attributes for SalesForce.
    """

    def _validate_request(self):
        """Validate the _saml_request.

        Throw an Exception if the validation does not succeed.
        """
        # skip salesforce.Processor as that requires the acs_url to be on the
        # salesforce.com domain
        Processor._validate_request(self)

        acs_domain = self._sp_config.get('acs_domain', 'salesforce.com')
        acs_name = self._sp_config.get('acs_name', 'SalesForce')

        if not ".{}".format(acs_domain) in self._request_params['ACS_URL']:
            raise CannotHandleAssertion(
                "AssertionConsumerService is not a {} URL.".format(acs_name))

    def _determine_audience(self):
        self._audience = self._sp_config.get(
            'saml_url',
            'https://saml.salesforce.com')

    def _get_canonical_email(self, account):
        """
        For the SalesForce portal, use email+portal@canonical.com.
        """
        email = super(SalesForceAttributeProcessor,
                      self)._get_canonical_email(account)
        if email is not None:
            username, domain = email.split('@')
            pattern = self._sp_config.get('email_pattern',
                                          '{username}+portal@{domain}')
            email = pattern.format(username=username, domain=domain)
            return email

    def _format_assertion(self):
        self._assertion_params['ATTRIBUTES'] = self._sp_config.get(
            'attributes', {})
        super(SalesForceAttributeProcessor, self)._format_assertion()


class ExpensifyProcessor(CanonicalOverrides, google_apps.Processor):
    """
    Expensify SAML 2.0 Processor, specific to Canonical Identity Provider.
    """

    def _validate_request(self):
        # override validation as the google_apps.Processor constrains requests
        # to the google.com domain only
        Processor._validate_request(self)
        if '.expensify.com/' not in self._request_params['ACS_URL']:
            raise CannotHandleAssertion(
                'AssertionConsumerService is not an Expensify URL.')


class CornerstoneProcessor(CanonicalOverrides, google_apps.Processor):
    """
    Cornerstone SAML 2.0 Processor, specific to Canonical Identity Provider.
    """
    def _validate_request(self):
        # override validation as the salesforce.Processor constrains requests
        # to the salesforce.com domain only
        Processor._validate_request(self)
        if '.csod.com/' not in self._request_params['ACS_URL']:
            raise CannotHandleAssertion(
                'AssertionConsumerService is not a Cornerstone URL.')

    def _decode_request(self):
        """
        Decodes non compressed request data.
        """
        Processor._decode_request(self)


class RocketchatProcessor(CanonicalOverrides, google_apps.Processor):
    """
    Rocketchat SAML 2.0 Processor, specific to Canonical Identity Provider.
    """
    ACS_URLS = ['https://rocket.ubuntu.com', 'https://rocket.canonical.com']

    def _validate_request(self):
        # override validation
        Processor._validate_request(self)
        if not any((acs_url in self._request_params['ACS_URL']
                   for acs_url in self.ACS_URLS)):
            raise CannotHandleAssertion(
                'AssertionConsumerService is not a Rocket.Chat URL.')

    def _format_assertion(self):
        # inject email attribute
        attributes = {
            'email': self._subject,
        }
        # if available, inject username attribute
        person_name = self._django_request.user.person_name
        if person_name:
            attributes['username'] = person_name
        self._assertion_params['ATTRIBUTES'] = attributes
        super(RocketchatProcessor, self)._format_assertion()


class BlueJeansProcessor(GoogleAppsProcessor):
    """
    BlueJeans SAML 2.0 Processor, specific to Canonical Identity Provider.
    """
    def _validate_request(self):
        # override validation as the GoogleAppsProcessor constrains requests
        # to the google.com domain
        Processor._validate_request(self)
        if 'https://bluejeans.com/' not in self._request_params['ACS_URL']:
            raise CannotHandleAssertion(
                'AssertionConsumerService is not a BlueJeans URL.')

    def _format_assertion(self):
        # inject email attribute
        attributes = {
            'email': self._subject,
        }
        # if available, inject username attribute
        person_name = self._django_request.user.person_name
        if person_name:
            attributes['username'] = person_name
        self._assertion_params['ATTRIBUTES'] = attributes
        super(BlueJeansProcessor, self)._format_assertion()


class LoopioProcessor(CanonicalOverrides, google_apps.Processor):
    """
    Loopio SAML 2.0 Processor, specific to Canonical Identity Provider.
    """
    def _validate_request(self):
        # override validation as the salesforce.Processor constrains requests
        # to the salesforce.com domain only
        Processor._validate_request(self)
        if '.loopio.com/' not in self._request_params['ACS_URL']:
            raise CannotHandleAssertion(
                'AssertionConsumerService is not a Loopio URL.')


class InfoProcessor(Processor):
    """
    A simple processor with a main utility method to just obtain the
    parameters from a SAML request. Not expected to be used for real
    SAML processing.
    """
    def get_request_params(self, request):
        """Extract and return the parameters from the request."""
        decode_methods = [base64.b64decode, codex.decode_base64_and_inflate]

        self._reset(request)
        self._extract_saml_request()

        # The request could simply be b64ed or b64-zipped (or maybe others?).
        # Try them all until one of them yields parseable SAML.
        saml_parsed = False
        for method in decode_methods:
            try:
                self._decode_request_with(method)
                self._parse_request()
                saml_parsed = True
                break
            except Exception as e:
                msg = 'Exception while decoding/parsing request: %s'
                self._logger.debug(msg, e)
        if not saml_parsed:
            raise CannotHandleAssertion(
                "Unable to parse SAML assertion from request")
        return self._request_params

    def _decode_request_with(self, method):
        """Decode the saml request with the given method."""
        self._request_xml = method(self._saml_request)


class CanonicalProcessor(CanonicalOverrides, Processor):
    """A generic processor supporting SAML 2.0 auth flows.

    This processor will enforce Canonical specific overrides and
    read SP configuration from the database using the SAMLConfig
    model for the entry associated with the request.
    """

    def _get_canonical_email(self, account):
        email = super(CanonicalProcessor, self)._get_canonical_email(account)
        if email is not None:
            sp_config = get_config_from_processor(self)
            username, domain = email.split('@')
            email = sp_config.email_pattern.format(
                username=username, domain=domain)
        return email

    # saml2idp.base.Processor interface

    def _build_assertion(self):
        super(CanonicalProcessor, self)._build_assertion()
        self._assertion_params['NOT_BEFORE'] = get_time_string(
            -settings.SAML_NOT_BEFORE_MINUTES * 60)
        self._assertion_params['NOT_ON_OR_AFTER'] = get_time_string(
            settings.SAML_NOT_AFTER_MINUTES * 60)
        self._assertion_params['ACS_URL'] = self._determine_acs_url()

    def _get_django_response_params(self):
        """
        Returns a dictionary of parameters for the response template.
        """
        tv = super(CanonicalProcessor, self)._get_django_response_params()
        tv['acs_url'] = self._response_params['ACS_URL']
        return tv

    def _build_response(self):
        super(CanonicalProcessor, self)._build_response()
        self._response_params['ACS_URL'] = self._determine_acs_url()

    def _determine_acs_url(self):
        sp_config = get_config_from_processor(self)
        if sp_config is not None:
            acs_url = sp_config.acs_url
        else:
            acs_url = self._request_params['ACS_URL']
        return acs_url

    def _decode_request(self):
        try:
            self._request_xml = self._decode_request_compressed()
        except Exception:
            self._request_xml = self._decode_request_uncompressed()

    def _parse_request(self):
        """
        Parses various parameters from _request_xml into _request_params.
        """
        # Minimal test to verify that it's not binarily encoded still:
        if not self._request_xml.strip().startswith('<'):
            raise Exception(
                'RequestXML is not valid XML; it may need to be decoded '
                'or decompressed.')
        soup = BeautifulStoneSoup(self._request_xml)
        request = soup.findAll()[0]
        params = {}
        # attempt to get ACS_URL
        acs_url = request.get('assertionconsumerserviceurl')
        issuer = request.findChild('saml:issuer')
        if issuer is not None:
            issuer = issuer.text
        params['ACS_URL'] = acs_url
        params['REQUEST_ISSUER'] = issuer
        params['REQUEST_ID'] = request['id']
        params['DESTINATION'] = request.get('destination', '')
        params['PROVIDER_NAME'] = request.get('providername', '')
        self._request_params = params

    def _determine_audience(self):
        self._audience = super(CanonicalProcessor, self)._determine_audience()

        sp_config = get_config_from_processor(self)
        if sp_config is not None:
            self._audience = sp_config.audience or self._audience

    def _determine_subject(self):
        account = self._django_request.user
        preferred = account.preferredemail.email

        # check if SP prefers @canonical emails
        sp_config = get_config_from_processor(self)
        if sp_config is not None:
            if sp_config.prefer_canonical_email:
                canonical_email = self._get_canonical_email(account)
                if canonical_email is not None:
                    preferred = canonical_email

        self._subject = preferred

    def _format_assertion(self):
        sp_config = get_config_from_processor(self)
        if sp_config is not None:
            if sp_config.attributes:
                attrs = json.loads(sp_config.attributes)
            else:
                attrs = {}
            certificate = certificate_from_sp_config(sp_config)
            self._assertion_params['ATTRIBUTES'] = self._eval_attributes(attrs)
            if sp_config.audience:
                self._format_assertion_restricted(certificate=certificate)
            else:
                self._format_assertion_unrestricted(certificate=certificate)

    def _validate_sp_config(self):
        acs_url = self._request_params.get('ACS_URL')
        issuer = self._request_params.get('REQUEST_ISSUER')
        sp_config = get_config_from_processor(self)
        if sp_config is not None:
            return sp_config

        msg = "Could not find saml config for "
        if acs_url:
            msg += "ACS URL '{}'.".format(acs_url)
        else:
            msg += "Issuer '{}'.".format(issuer)
        raise CannotHandleAssertion(msg)

    def _validate_request(self):
        sp_config = self._validate_sp_config()
        # validate request domain
        acs_url = sp_config.acs_url
        acs_domain = sp_config.acs_domain
        parts = urlparse.urlparse(acs_url)
        is_secure = parts.scheme == 'https'
        is_valid_domain = parts.netloc.endswith(acs_domain)
        if not (is_secure and is_valid_domain):
            msg = "AssertionConsumerService is not a {} URL."
            raise CannotHandleAssertion(msg.format(acs_domain))

        # success, so keep a reference to the config
        self._sp_config = sp_config

    def _validate_user(self):
        sp_config = self._validate_sp_config()
        # Check for team allowance/restriction.
        user = self._django_request.user
        allow_team_list = sp_config.get_allowed_teams_list()
        if allow_team_list and not user.person_in_any_team(allow_team_list):
            msg = "User not allowed to access ACS URL {}."
            raise CannotHandleAssertion(msg.format(sp_config.acs_url))

    # internal implementation

    def _decode_request_uncompressed(self):
        return base64.b64decode(self._saml_request)

    def _decode_request_compressed(self):
        return codex.decode_base64_and_inflate(self._saml_request)

    def _format_assertion_unrestricted(self, certificate=None):
        self._assertion_xml = xml_render.get_assertion_googleapps_xml(
            self._assertion_params, signed=True, certificate=certificate)

    def _format_assertion_restricted(self, certificate=None):
        self._assertion_xml = xml_render.get_assertion_salesforce_xml(
            self._assertion_params, signed=True, certificate=certificate)

    def _format_response(self):
        sp_config = get_config_from_processor(self)
        # Helper properly handles sp_config being None.
        certificate = certificate_from_sp_config(sp_config)
        super(CanonicalProcessor, self)._format_response(
            certificate=certificate)

    def _eval_attributes(self, attributes):
        for key, value in attributes.items():
            if value == '{{subject}}':
                attributes[key] = self._subject
            elif value == '{{person_name}}':
                person_name = self._django_request.user.person_name
                if person_name:
                    attributes[key] = person_name
            elif value == '{{openid}}':
                request = self._django_request
                attributes[key] = request.build_absolute_uri(
                    request.user.get_absolute_url())
        return attributes
