import json

from django.core.exceptions import ValidationError
from django.test import TestCase

from ubuntu_sso_saml.models import SAMLConfig
from ubuntu_sso_saml.tests.helpers import find_cert_path


class SAMLConfigTestCase(TestCase):

    def test_required_fields(self):
        with self.assertRaises(ValidationError) as err:
            SAMLConfig.objects.create()
        self.assertEqual(err.exception.error_dict.keys(), ['acs_url'])

    def test_defaults(self):
        conf = SAMLConfig.objects.create(acs_url='http://localhost')
        self.assertEqual(conf.acs_url, 'http://localhost')
        self.assertEqual(conf.acs_domain, '')
        self.assertIsNone(conf.issuer)
        self.assertEqual(conf.audience, '')
        self.assertEqual(conf.displayname, '')
        self.assertFalse(conf.prefer_canonical_email)
        self.assertFalse(conf.enabled)
        self.assertEqual(conf.email_pattern, '{username}@{domain}')
        self.assertEqual(conf.attributes, '')
        self.assertEqual(conf.links, '')

    def test_validate_attributes_not_json(self):
        for attributes in [
            'invalid',
            '[("foo", "bar")]',
        ]:
            with self.assertRaises(ValidationError) as err:
                SAMLConfig.objects.create(
                    acs_url='http://localhost', attributes=attributes)
            self.assertEqual(err.exception.message_dict['attributes'],
                             ["No JSON object could be decoded"])

    def test_validate_attributes_json(self):
        for attributes in [
            '[]',
            '["foo", "bar"]',
            '[["foo", "bar"]]',
        ]:
            with self.assertRaises(ValidationError) as err:
                SAMLConfig.objects.create(
                    acs_url='http://localhost', attributes=attributes)
            self.assertEqual(
                err.exception.message_dict['attributes'],
                ["'attributes' must be a valid JSON object (dict)."])

    def test_validate_links_not_json(self):
        for links in [
            'invalid',
            '[("foo", "bar")]',
        ]:
            with self.assertRaises(ValidationError) as err:
                SAMLConfig.objects.create(
                    acs_url='http://localhost', links=links)
            self.assertEqual(err.exception.message_dict['links'],
                             ["No JSON object could be decoded"])

    def test_validate_links_json_not_dict(self):
        for links in [
            '[]',
            '["foo", "bar"]',
            '[["foo", "bar"]]',
        ]:
            with self.assertRaises(ValidationError) as err:
                SAMLConfig.objects.create(
                    acs_url='http://localhost', links=links)
            self.assertEqual(
                err.exception.message_dict['links'],
                ["'links' must be a valid JSON object (dict)."])

    def test_validate_cert_invalid(self):
        with self.assertRaises(ValidationError) as err:
            SAMLConfig.objects.create(
                acs_url='http://localhost', certificate="something bogus")
        self.assertEqual(
            err.exception.message_dict['certificate'],
            ["'certificate' must be a PEM-formatted X.509 certificate."])

    def test_validate_cert_valid(self):
        cert_filename = find_cert_path('certificate.pem')
        with open(cert_filename) as cert_file:
            cert_data = cert_file.read()
        try:
            SAMLConfig.objects.create(
                acs_url='http://localhost', certificate=cert_data)
        except ValidationError:
            self.fail("Unexpectedly raised ValidationError")

    def test_getitem(self):
        kwargs = {
            'acs_url': 'http://localhost',
            'issuer': 'http://localhost/metadata',
            'displayname': 'Foo',
            'audience': 'localhost',
            'prefer_canonical_email': True,
            'enabled': True,
            'email_pattern': '{username}@{domain}',
            'attributes': '{"subject": "{{email}}"}',
            'links': '{"foo": "http://example.com/foo"}',
        }
        conf = SAMLConfig(**kwargs)
        for key in kwargs:
            self.assertEqual(conf[key], kwargs[key])

    def test_unicode(self):
        conf = SAMLConfig(acs_url='http://localhost', displayname='Foo',
                          audience='localhost', prefer_canonical_email=True,
                          enabled=True)
        self.assertEqual(unicode(conf), 'Foo')

    def test_repr(self):
        links = {
            'prefix/(?P<target>\\w+)': 'http://example.com/%(target)s',
        }
        expected_links = json.dumps(links)
        conf = SAMLConfig(
            acs_url='http://localhost', displayname='Foo',
            audience='localhost', prefer_canonical_email=True,
            enabled=True, email_pattern='{username}+test@{domain}',
            attributes='{"foo": "bar"}', acs_domain='localhost',
            links=expected_links)

        expected = (
            "<SAMLConfig(displayname='Foo', acs_url='http://localhost', "
            "acs_domain='localhost', audience='localhost', "
            "prefer_canonical_email=True, enabled=True, "
            "email_pattern='{username}+test@{domain}', "
            "attributes='{\"foo\": \"bar\"}', "
            "links='%s')>" % expected_links)
        self.assertEqual(repr(conf), expected)

    def test_allowed_teams_single(self):
        conf = SAMLConfig(acs_url='http://localhost', displayname='Foo',
                          audience='localhost', allowed_teams='foo')
        self.assertEqual(conf.get_allowed_teams_list(), ['foo'])

    def test_allowed_teams_empty(self):
        conf = SAMLConfig(acs_url='http://localhost', displayname='Foo',
                          audience='localhost', allowed_teams='')
        self.assertEqual(conf.get_allowed_teams_list(), [])

    def test_allowed_teams_empty_substrings(self):
        conf = SAMLConfig(acs_url='http://localhost', displayname='Foo',
                          audience='localhost', allowed_teams=',,,,')
        self.assertEqual(conf.get_allowed_teams_list(), [])

    def test_allowed_teams_empty_with_spaces(self):

        conf = SAMLConfig(acs_url='http://localhost', displayname='Foo',
                          audience='localhost', allowed_teams='  ')
        self.assertEqual(conf.get_allowed_teams_list(), [])

    def test_allowed_teams_many_test_stripping(self):
        conf = SAMLConfig(acs_url='http://localhost', displayname='Foo',
                          audience='localhost',
                          allowed_teams=' one, two  , three ')
        self.assertEqual(conf.get_allowed_teams_list(),
                         ['one', 'two', 'three'])

    def test_allowed_teams_many_weird_commas(self):
        conf = SAMLConfig(acs_url='http://localhost', displayname='Foo',
                          audience='localhost',
                          allowed_teams=', two  ,')
        self.assertEqual(conf.get_allowed_teams_list(),
                         ['two'])
