"""Tests for SAML2 Processors.

A lot of this is borrowed directly from the saml2idp tests.
"""
import base64
import copy
import json
import time
import urllib
from StringIO import StringIO
from xml.dom import minidom

from django.test.client import Client
from django.test.utils import override_settings
from django.urls import reverse
from gargoyle.models import GLOBAL, Switch
from gargoyle.testutils import switches
from lxml import etree
from mock import patch
from pyquery import PyQuery
from saml2idp import codex, saml2idp_metadata
from saml2idp.base import get_time_string
from saml2idp.tests import google_apps, salesforce

from identityprovider.models import (
    Account,
    EmailAddress,
    OpenIDRPConfig,
    twofactor,
)
from identityprovider.models.const import EmailStatus
from identityprovider.tests import DEFAULT_USER_PASSWORD
from identityprovider.tests.utils import AuthenticatedTestCase
from ubuntu_sso_saml.models import SAMLConfig
from ubuntu_sso_saml.tests.helpers import cert_string_from_file, find_cert_path


SAML_TEAM = 'saml2team'  # Name of team enabled for SAML 2.0

CUSTOM_TEST_CERTIFICATE = "custom-test-certificate.pem"


def get_saml_response(response):
    tree = PyQuery(response.content)
    inputtag = tree.find('input[name="SAMLResponse"]')
    assert len(inputtag) == 1
    encoded_response = inputtag[0].get('value')
    samlresponse = base64.b64decode(encoded_response)
    return samlresponse


def load_saml_schema():
    """Loads an XSD and returns an XMLSchema for it."""
    if not hasattr(load_saml_schema, "parser"):
        with open("saml-schema-xsd/saml-schema-protocol-2.0.xsd") as schema_f:
            schema_doc = etree.parse(schema_f)
            load_saml_schema.parser = etree.XMLSchema(schema_doc)
    return load_saml_schema.parser


def validate_saml(samlresponse):
    """Validates the given string is a valid SAML XML response."""
    response_f = StringIO(samlresponse)
    response_doc = etree.parse(response_f)
    saml_schema = load_saml_schema()
    saml_schema.assertValid(response_doc)


class BaseProcessorTestCase(AuthenticatedTestCase):
    """
    Base class with low-level methods to set up saml dance prereqs.

    Not to be used by itself since its methods assume some setup that should be
    done by subclasses.

    Its main reason for existing is avoiding duplicating these methods in the
    InfoProcessorTestCase class which can't subclass
    GoogleAppsAssertionTestCase because some of its test methods don't apply
    to InfoProcessor.
    """

    def setUp(self, *args, **kwargs):
        super(BaseProcessorTestCase, self).setUp(*args, **kwargs)
        # Save settings.
        new_saml2idp_config = copy.deepcopy(saml2idp_metadata.SAML2IDP_CONFIG)
        self.patch(saml2idp_metadata, 'SAML2IDP_CONFIG',
                   new=new_saml2idp_config)
        cert_file = find_cert_path('certificate.pem')
        priv_file = find_cert_path('private-key.pem')
        test_config = {
            'certificate_file': cert_file,
            'private_key_file': priv_file,
            'signing': True,
        }
        saml2idp_metadata.SAML2IDP_CONFIG.update(test_config)

        saml2idp_metadata.SAML2IDP_REMOTES['foobar'] = {
            'acs_url': self.ACS_URL,
            'processor': self.PROCESSOR,
        }
        self.addCleanup(saml2idp_metadata.SAML2IDP_REMOTES.pop, 'foobar')

        # enable TWOFACTOR flag
        switch = Switch.objects.create(key='TWOFACTOR', status=GLOBAL)
        self.addCleanup(switch.delete)

        # mock twofactor auth
        self.patch(twofactor, 'authenticate')
        self.saml2team = self.factory.make_team(SAML_TEAM)

    def setup_saml_sp(self, **kwargs):
        # create an RP that prefers @canonical emails
        OpenIDRPConfig.objects.create(
            trust_root=kwargs.get('acs_url', self.ACS_URL),
            prefer_canonical_email=kwargs.get('prefer_canonical_email', False),
            twofactor_required=kwargs.get('twofactor_required', False))

    def setup_saml_emails(self, email='first.last@canonical.com',
                          status=EmailStatus.VALIDATED):
        self.factory.add_account_to_team(self.account, self.saml2team)
        # add @canonical.com email address to the account
        account = EmailAddress.objects.get(email=self.login_email).account
        EmailAddress.objects.create(
            account=account, email=email, status=status)
        return account.preferredemail.email

    def get_request_data(self):
        return self.REQUEST_DATA


class CanonicalProcessorBaseTestCase(BaseProcessorTestCase):

    PROCESSOR = 'ubuntu_sso_saml.processors.CanonicalProcessor'

    def setup_saml_sp(self, **kwargs):
        acs_url = kwargs.pop('acs_url', self.ACS_URL)
        issuer = kwargs.pop('issuer', getattr(self, 'REQUEST_ISSUER', None))
        prefer_canonical_email = kwargs.get('prefer_canonical_email', False)
        twofactor_required = kwargs.pop('twofactor_required', False)
        allowed_teams = kwargs.pop('allowed_teams', '')
        # create an rp config that prefers @canonical emails
        # (for compatibility)
        OpenIDRPConfig.objects.create(
            trust_root=acs_url, prefer_canonical_email=prefer_canonical_email,
            twofactor_required=twofactor_required)
        # create a saml config that prefers @canonical emails
        # set 'enabled' field if not defined
        kwargs.setdefault('enabled', True)
        return SAMLConfig.objects.create(
            acs_url=acs_url, issuer=issuer, allowed_teams=allowed_teams,
            **kwargs)

    def do_saml_request(self, data=None):
        # do twofactor dance
        self.client.post(reverse('twofactor'), {'oath_token': '123456'})
        # do saml dance
        if data is None:
            data = self.get_request_data()
        response = self.client.get('/+saml', data=data, follow=True)
        samlresponse = get_saml_response(response)
        validate_saml(samlresponse)
        return samlresponse


class NoSamlRequestTestCase(BaseProcessorTestCase):
    """
    Test accessing the /+saml endpoint with no SAML request.
    """
    # Ensure the SAML setup can be done properly - though it won't be used.
    ACS_URL = google_apps.GOOGLE_APPS_ACS
    REQUEST_DATA = google_apps.REQUEST_DATA
    PROCESSOR = 'ubuntu_sso_saml.processors.GoogleAppsProcessor'

    def test_direct_access(self):
        # Data is empty - no SAML request!
        response = self.client.get('/+saml', data={}, follow=True)
        self.assertTemplateUsed(response, 'ubuntu_sso_saml/no_saml.html')
        self.assertContains(response, 'not intended for direct access')

    def test_with_referrer(self):
        # Data is empty - no SAML request!
        response = self.client.get(
            '/+saml', data={}, follow=True,
            HTTP_REFERER="http://foo/chop_this_off")
        self.assertTemplateUsed(response, 'ubuntu_sso_saml/no_saml.html')
        self.assertContains(
            response,
            ('Referred from URL: <span class="samlerror__badurl">'
             'http://foo</span>'))
        self.assertNotContains(response, 'chop_this_off')


class GoogleAppsAssertionTestCase(BaseProcessorTestCase):
    """
    Test SAML Assertions for a Google Apps Service Point.

    Sub-classes with identical behavior, but different ACS_URL or REQUEST_DATA
    simply specify their own attributes on the class.
    """
    ACS_URL = google_apps.GOOGLE_APPS_ACS
    REQUEST_DATA = google_apps.REQUEST_DATA
    PROCESSOR = 'ubuntu_sso_saml.processors.GoogleAppsProcessor'

    def test_authnrequest_handled(self):
        """ AuthnRequest isn't handled by a Processor. """
        # Arrange/Act:
        data = self.get_request_data()
        response = self.client.get('/+saml', data=data, follow=False)

        # Assert:
        self.assertEqual(response.status_code, 302)

    def test_authnrequest_csrf_exempt(self):
        client = Client(enforce_csrf_checks=True)
        data = self.get_request_data()
        response = client.post('/+saml', data=data, follow=False)
        self.assertEqual(response.status_code, 302)

    def test_user_logged_in(self):
        """ Unable to produce a valid Assertion for SAML 2-enabled user. """
        # Act:
        self.factory.add_account_to_team(self.account, self.saml2team)
        self.setup_saml_sp()
        # upgrade to 2f session
        response = self.client.post(reverse('twofactor'),
                                    {'oath_token': '123456'})
        data = self.get_request_data()
        response = self.client.get('/+saml', data=data, follow=True)
        samlresponse = get_saml_response(response)

        # Assert:
        self.assertIn(self.login_email, samlresponse)

    def test_user_has_different_preferred_email(self):
        """ User's preferred email is not being passed as SAML identifier. """
        # make sure no @canonical.com email address is used so the preferred
        # email is returned
        self.login_email = 'mark@example.com'
        self.account = self.factory.make_account(
            email=self.login_email, password=self.login_password)
        assert self.client.login(
            username=self.login_email, password=self.login_password)
        # Act:
        self.factory.add_account_to_team(self.account, self.saml2team)
        # upgrade to 2f session
        response = self.client.post(reverse('twofactor'),
                                    {'oath_token': '123456'})
        # create alternate email
        # and make sure the email is linked to the account
        alternate_email = 'alternate@example.com'
        alternate = EmailAddress.objects.create(
            email=alternate_email, account=self.account,
            status=EmailStatus.VALIDATED)
        # and make the alternate email the preferred
        self.account.preferredemail = alternate
        self.account.save()
        self.setup_saml_sp()

        account = Account.objects.get(pk=self.account.pk)
        self.assertEqual(account.preferredemail, alternate)
        data = self.get_request_data()
        response = self.client.get('/+saml', data=data, follow=True)
        samlresponse = get_saml_response(response)

        # Assert:
        self.assertIn(alternate_email, samlresponse)

    def test_user_not_2f_but_no_rpconfig(self):
        data = self.get_request_data()
        response = self.client.get('/+saml', data=data, follow=True)

        self.assertRedirects(response, reverse('login_process'))

    def test_user_not_2f_but_rpconfig_without_flag(self):
        OpenIDRPConfig.objects.create(trust_root=self.ACS_URL)

        data = self.get_request_data()
        response = self.client.get('/+saml', data=data, follow=True)

        self.assertRedirects(response, reverse('login_process'))

    def test_user_not_2f_but_not_required(self):
        OpenIDRPConfig.objects.create(
            trust_root=self.ACS_URL, twofactor_required=True)

        with switches(TWOFACTOR=False):
            data = self.get_request_data()
            response = self.client.get('/+saml', data=data, follow=True)

        self.assertRedirects(response, reverse('login_process'))

    def test_user_not_2f_but_required(self):
        # make sure the rpconfig entry requires 2f
        self.setup_saml_sp(twofactor_required=True)

        with switches(TWOFACTOR=True):
            data = self.get_request_data()
            response = self.client.get('/+saml', data=data, follow=True)

        # make sure the last redirect was for twofactor
        self.assertIn(reverse('twofactor'), response.redirect_chain[-1][0])

    def do_saml_request(self):
        # do twofactor dance
        self.client.post(reverse('twofactor'), {'oath_token': '123456'})
        # do saml dance
        data = self.get_request_data()
        response = self.client.get('/+saml', data=data, follow=True)
        samlresponse = get_saml_response(response)
        validate_saml(samlresponse)
        return samlresponse

    def test_canonical_email_is_preferred(self):
        self.setup_saml_sp(prefer_canonical_email=True)
        preferred = self.setup_saml_emails()
        samlresponse = self.do_saml_request()
        # verify @canonical.com email address is returned despite it's not the
        # preferredemail
        self.assertIn('first.last@canonical.com', samlresponse)
        self.assertNotEqual('first.last@canonical.com', preferred)

    def test_canonical_email_is_preferred_despite_case(self):
        email = 'FIRST.LAST@CANONICAL.COM'
        self.setup_saml_sp(prefer_canonical_email=True)
        preferred = self.setup_saml_emails(email=email)
        samlresponse = self.do_saml_request()
        # verify @canonical.com email address is returned despite it's not the
        # preferredemail
        self.assertIn(email, samlresponse)
        self.assertNotEqual(email, preferred)

    def test_only_long_form_canonical_email_is_preferred(self):
        email = self.login_email
        # add non-preferred long form email address
        EmailAddress.objects.create(
            account=self.account, email='first.last+alias@canonical.com',
            status=EmailStatus.VALIDATED)
        self.setup_saml_sp(prefer_canonical_email=True)
        preferred = self.setup_saml_emails()
        self.assertEqual(preferred, email)
        self.assertNotEqual(preferred, 'first.last@canonical.com')

        samlresponse = self.do_saml_request()
        # verify @canonical.com email address is returned despite it's not the
        # preferredemail
        self.assertIn('first.last@canonical.com', samlresponse)
        self.assertNotIn(email, samlresponse)

    def test_canonical_email_is_not_preferred(self):
        self.setup_saml_sp(prefer_canonical_email=False)
        preferred = self.setup_saml_emails()
        samlresponse = self.do_saml_request()
        # verify @canonical.com email address is *not* returned despite because
        # the flag was not set
        self.assertIn(preferred, samlresponse)
        self.assertNotEqual('first.last@canonical.com', preferred)

    def test_canonical_email_without_rpconfig(self):
        self.setup_saml_sp()
        preferred = self.setup_saml_emails()
        # make sure there is no RP for this url
        OpenIDRPConfig.objects.filter(trust_root=self.ACS_URL).delete()

        samlresponse = self.do_saml_request()
        # verify @canonical.com email address is *not* returned despite because
        # the flag was not set
        self.assertIn(preferred, samlresponse)
        self.assertNotEqual('first.last@canonical.com', preferred)

    def test_short_form_canonical_email_ignored(self):
        # start off a clean slate
        short_email = EmailAddress.objects.get(
            account=self.account, email='test@canonical.com')
        short_email.status = EmailStatus.VALIDATED
        short_email.save()
        EmailAddress.objects.create(account=self.account, email='foo@foo.com',
                                    status=EmailStatus.PREFERRED)
        self.login_email = 'foo@foo.com'

        # make sure the short email it's not the preferred address
        self.setup_saml_sp(prefer_canonical_email=True)
        preferred = self.setup_saml_emails()
        assert preferred == 'foo@foo.com'
        # make sure there's no long form @canonical email
        EmailAddress.objects.filter(
            account=self.account, email='first.last@canonical.com').delete()

        samlresponse = self.do_saml_request()
        # verify @canonical.com email address is *not* returned
        self.assertIn(preferred, samlresponse)
        self.assertNotEqual('first.last@canonical.com', preferred)
        self.assertNotEqual('test@canonical.com', preferred)

    def test_non_canonical_email(self):
        # prepare account for SAML
        account = self.factory.make_account()
        preferred = account.preferredemail.email
        self.factory.add_account_to_team(account, self.saml2team)
        self.setup_saml_sp(prefer_canonical_email=True)

        # make sure account has only one email
        assert account.emailaddress_set.count() == 1
        # and it's not a @canonical.com one
        assert '@canonical.com' not in preferred

        assert self.client.login(
            username=preferred, password=DEFAULT_USER_PASSWORD)

        samlresponse = self.do_saml_request()
        self.assertIn(preferred, samlresponse)

    def test_multiple_long_form_canonical_emails(self):
        self.setup_saml_sp(prefer_canonical_email=True)
        self.setup_saml_emails()
        # change preferred email to non-standard long form email address
        preferred = EmailAddress.objects.create(
            account=self.account, email='other.name@canonical.com',
            status=EmailStatus.PREFERRED)
        email = self.account.preferredemail
        email.status = EmailStatus.VALIDATED
        email.save()
        # make sure account shows new email as preferred
        account = Account.objects.get(pk=self.account.pk)
        assert account.preferredemail == preferred

        samlresponse = self.do_saml_request()
        # verify @canonical.com email address is returned despite it's not the
        # preferredemail
        self.assertIn('other.name@canonical.com', samlresponse)
        self.assertNotIn('first.last@canonical.com', samlresponse)


class GoogleAppsAssertionCanonicalProcessorTestCase(
        CanonicalProcessorBaseTestCase, GoogleAppsAssertionTestCase):
    pass


class SalesForceAssertionTestCase(GoogleAppsAssertionTestCase):
    """
    Test SAML Assertions for a SalesForce Service Point.
    """
    ACS_URL = salesforce.SALESFORCE_ACS
    REQUEST_DATA = salesforce.REQUEST_DATA
    PROCESSOR = 'ubuntu_sso_saml.processors.SalesForceProcessor'


class SalesForceAssertionCanonicalProcessorTestCase(
        CanonicalProcessorBaseTestCase, SalesForceAssertionTestCase):
    pass


class SalesForcePortalAssertionTestCase(SalesForceAssertionTestCase):
    ACS_URL = salesforce.SALESFORCE_ACS
    REQUEST_DATA = salesforce.REQUEST_DATA
    PROCESSOR = 'ubuntu_sso_saml.processors.SalesForceAttributeProcessor'

    def setUp(self, *args, **kwargs):
        super(SalesForcePortalAssertionTestCase, self).setUp(*args, **kwargs)
        target = 'https://somesite.salesforce.com/%(target)s'
        saml2idp_metadata.SAML2IDP_REMOTES['foobar'].update({
            'links': {
                # NOTE: This doesn't match the SALESFORCE_ACS,
                # but it will in production.
                r'portal/(?P<target>\w+)': target,
            },
            'attributes': {
                'organization_id': 'SAMPLE_ORG_ID',
                'portal_id': 'SAMPLE_PORTAL_ID',
            },
            'email_pattern': '{username}+portal@{domain}',
        })

    def test_canonical_email_is_preferred(self):
        self.setup_saml_sp(prefer_canonical_email=True)
        preferred = self.setup_saml_emails()
        samlresponse = self.do_saml_request()
        self.assertIn('first.last+portal@canonical.com', samlresponse)
        self.assertNotEqual('first.last+portal@canonical.com', preferred)

    def test_canonical_email_is_preferred_despite_case(self):
        email = 'FIRST.LAST@CANONICAL.COM'
        portal_email = 'FIRST.LAST+portal@CANONICAL.COM'
        self.setup_saml_sp(prefer_canonical_email=True)
        preferred = self.setup_saml_emails(email=email)
        samlresponse = self.do_saml_request()
        # verify @canonical.com email address is returned despite it's not the
        # preferredemail
        self.assertIn(portal_email, samlresponse)
        self.assertNotEqual(portal_email, preferred)

    def test_only_long_form_canonical_email_is_preferred(self):
        email = self.login_email
        portal_email = 'first.last+portal@canonical.com'
        # add non-preferred long form email address
        EmailAddress.objects.create(
            account=self.account, email='first.last+alias@canonical.com',
            status=EmailStatus.VALIDATED)
        self.setup_saml_sp(prefer_canonical_email=True)
        preferred = self.setup_saml_emails()
        self.assertEqual(preferred, email)
        self.assertNotEqual(preferred, 'first.last@canonical.com')

        samlresponse = self.do_saml_request()
        # verify @canonical.com email address is returned despite it's not the
        # preferredemail
        self.assertIn(portal_email, samlresponse)
        self.assertNotEqual(portal_email, preferred)

    def test_non_canonical_email(self):
        # prepare account for SAML
        account = self.factory.make_account()
        preferred = account.preferredemail.email
        self.factory.add_account_to_team(account, self.saml2team)
        self.setup_saml_sp(prefer_canonical_email=True)

        # make sure account has only one email
        assert account.emailaddress_set.count() == 1
        # and it's not a @canonical.com one
        assert '@canonical.com' not in preferred

        assert self.client.login(
            username=preferred, password=DEFAULT_USER_PASSWORD)

        samlresponse = self.do_saml_request()
        self.assertIn(preferred, samlresponse)
        self.assertNotIn('+portal@', samlresponse)

    def test_multiple_long_form_canonical_emails(self):
        self.setup_saml_sp(prefer_canonical_email=True)
        self.setup_saml_emails()
        # change preferred email to non-standard long form email address
        preferred = EmailAddress.objects.create(
            account=self.account, email='other.name@canonical.com',
            status=EmailStatus.PREFERRED)
        email = self.account.preferredemail
        email.status = EmailStatus.VALIDATED
        email.save()
        # make sure account shows new email as preferred
        account = Account.objects.get(pk=self.account.pk)
        assert account.preferredemail == preferred

        samlresponse = self.do_saml_request()
        # verify @canonical.com email address is returned despite it's not the
        # preferredemail
        self.assertIn('other.name+portal@canonical.com', samlresponse)
        self.assertNotIn('first.last+portal@canonical.com', samlresponse)

    @override_settings(ROOT_URLCONF='ubuntu_sso_saml.tests.urls_deeplink')
    def test_portal_deeplink(self):
        """Unable to produce a valid Portal Assertion.

        (for SAML 2-enabled user)
        """
        # Act:
        self.factory.add_account_to_team(self.account, self.saml2team)
        self.setup_saml_sp()
        # upgrade to 2f session
        response = self.client.post(reverse('twofactor'),
                                    {'oath_token': '123456'})
        data = self.get_request_data()
        response = self.client.get(
            '/+saml/init/portal/SOMETARGETID', data=data, follow=True)
        samlresponse = get_saml_response(response)

        tree = PyQuery(response.content)
        relsttag = tree.find('input[name="RelayState"]')
        assert len(relsttag) == 1
        relaystate = relsttag[0].get('value')

        samlsoup = PyQuery(samlresponse.replace('xmlns:', 'xmlnamespace:'))

        org_attr = samlsoup.find('attribute[name="organization_id"]')
        assert len(org_attr) == 1
        org_val = org_attr[0].find('attributevalue')

        port_attr = samlsoup.find('attribute[name="portal_id"]')
        assert len(port_attr) == 1
        port_val = port_attr[0].find('attributevalue')

        # Assert:
        self.assertIn(self.login_email, samlresponse)
        self.assertEqual(relaystate,
                         'https://somesite.salesforce.com/SOMETARGETID')
        self.assertEqual(org_val.text, 'SAMPLE_ORG_ID')
        self.assertEqual(port_val.text, 'SAMPLE_PORTAL_ID')

    @override_settings(
        ROOT_URLCONF='ubuntu_sso_saml.tests.urls_deeplink_custom'
    )
    def test_portal_deeplink_with_custom_attributes(self):
        # patch remote config
        saml2idp_metadata.SAML2IDP_REMOTES['foobar'].update({
            'attributes': {
                'organization_id': 'SAMPLE_ORG_ID',
            },
        })

        # Act:
        self.factory.add_account_to_team(self.account, self.saml2team)
        self.setup_saml_sp(
            attributes=json.dumps({'organization_id': 'SAMPLE_ORG_ID'}))
        # upgrade to 2f session
        response = self.client.post(reverse('twofactor'),
                                    {'oath_token': '123456'})
        data = self.get_request_data()
        response = self.client.get(
            '/+saml/init/portal/SOMETARGETID', data=data, follow=True)
        samlresponse = get_saml_response(response)

        tree = PyQuery(response.content)
        relsttag = tree.find('input[name="RelayState"]')
        assert len(relsttag) == 1
        relaystate = relsttag[0].get('value')

        samlsoup = PyQuery(samlresponse.replace('xmlns:', 'xmlnamespace:'))

        org_attr = samlsoup.find('attribute[name="organization_id"]')
        assert len(org_attr) == 1
        org_val = org_attr[0].find('attributevalue')

        port_attr = samlsoup.find('attribute[name="portal_id"]')
        # this remote doesn't have a portal_id attribute
        assert len(port_attr) == 0

        # Assert:
        self.assertIn(self.login_email, samlresponse)
        self.assertEqual(relaystate,
                         'https://somesite.salesforce.com/SOMETARGETID')
        self.assertEqual(org_val.text, 'SAMPLE_ORG_ID')

    def test_custom_email_pattern(self):
        # patch remote config
        saml2idp_metadata.SAML2IDP_REMOTES['foobar'].update({
            'email_pattern': '{username}@hr.{domain}'
        })
        self.setup_saml_sp(prefer_canonical_email=True,
                           email_pattern='{username}@hr.{domain}')
        preferred = self.setup_saml_emails()
        # make sure transformed email is not a valid email for account
        account = Account.objects.get_by_email(self.login_email)
        account_emails = [email.email
                          for email in account.emailaddress_set.all()]
        self.assertNotIn('first.last@hr.canonical.com', account_emails)
        # and make sure the preferred canonical email address is in there
        self.assertIn('first.last@canonical.com', account_emails)
        # even if it's not the preferred email
        self.assertNotEqual(preferred, 'first.last@canonical.com')

        samlresponse = self.do_saml_request()
        # make sure the login email was transformed
        self.assertIn('first.last@hr.canonical.com', samlresponse)

    def test_custom_email_pattern_with_short_form_canonical_email(self):
        # patch remote config
        saml2idp_metadata.SAML2IDP_REMOTES['foobar'].update({
            'email_pattern': '{username}@hr.{domain}'
        })
        self.setup_saml_sp(prefer_canonical_email=True,
                           email_pattern='{username}@hr.{domain}')
        preferred = self.setup_saml_emails()
        account = Account.objects.get_by_email(self.login_email)
        # make sure long-form email is not registered with account
        EmailAddress.objects.filter(
            account=account, email='first.last@canonical.com').delete()
        # make sure transformed email is not a valid email for account
        account_emails = [email.email
                          for email in account.emailaddress_set.all()]
        self.assertNotIn('test@hr.canonical.com', account_emails)
        self.assertNotIn('first.last@hr.canonical.com', account_emails)
        # and make sure the preferred canonical email address is in there
        self.assertIn('test@canonical.com', account_emails)
        self.assertNotIn('first.last@canonical.com', account_emails)
        # even if it's not the preferred email
        self.assertNotEqual(preferred, 'first.last@canonical.com')

        samlresponse = self.do_saml_request()
        # make sure the login email was transformed
        self.assertIn('test@hr.canonical.com', samlresponse)

    def make_saml_request(self, acs_url=None, saml_url=None):
        saml_request = self.REQUEST_DATA['SAMLRequest']
        if acs_url or saml_url:
            saml_request = base64.b64decode(saml_request)
            if acs_url:
                saml_request = saml_request.replace(
                    'https://login.salesforce.com', acs_url)
            if saml_url:
                saml_request = saml_request.replace(
                    'https://saml.salesforce.com', saml_url)
            saml_request = base64.b64encode(saml_request)
        return saml_request

    def test_custom_acs_domain(self):
        # patch remote config
        saml2idp_metadata.SAML2IDP_REMOTES['foobar'].update({
            'acs_domain': 'cloudforce.com',
            'acs_url': 'https://login.cloudforce.com'
        })
        self.setup_saml_sp(acs_url='https://login.cloudforce.com',
                           acs_domain='cloudforce.com')
        saml_request = self.make_saml_request(
            acs_url='https://login.cloudforce.com')

        with patch.dict(self.REQUEST_DATA, SAMLRequest=saml_request):
            samlresponse = self.do_saml_request()
        self.assertIn('https://login.cloudforce.com', samlresponse)

    @patch('ubuntu_sso_saml.views.logger.exception')
    def test_custom_acs_name(self, mock_logger_exception):
        saml2idp_metadata.SAML2IDP_REMOTES['foobar'].update({
            'acs_domain': 'cloudforce.com',
            'acs_name': 'CloudForce',
        })
        self.setup_saml_sp(
            acs_domain='cloudforce.com', displayname='CloudForce')

        # Do the saml request manually since it won't exception out
        # do twofactor dance
        self.client.post(reverse('twofactor'), {'oath_token': '123456'})
        # do saml dance
        data = self.get_request_data()
        response = self.client.get('/+saml', data=data, follow=True)
        # Instead of an raised exception, we should get a response with
        # error message and a *logged* trace
        self.assertIn("https://login.salesforce.com", response.content)
        mock_logger_exception.assert_called_with(
            "Can't handle assertion from %s",
            "https://login.salesforce.com")

    def test_custom_saml_url(self):
        # patch remote config
        saml2idp_metadata.SAML2IDP_REMOTES['foobar'].update({
            'saml_url': 'https://saml.cloudforce.com',
        })
        self.setup_saml_sp(audience='https://saml.cloudforce.com')
        saml_request = self.make_saml_request(
            saml_url='https://saml.cloudforce.com')

        with patch.dict(self.REQUEST_DATA, SAMLRequest=saml_request):
            samlresponse = self.do_saml_request()
        self.assertIn('https://saml.cloudforce.com', samlresponse)


class SalesForcePortalAssertionCanonicalProcessorTestCase(
        CanonicalProcessorBaseTestCase, SalesForcePortalAssertionTestCase):

    def setup_saml_sp(self, **kwargs):
        # define a default value for the email pattern if none is given
        kwargs.setdefault('email_pattern', '{username}+portal@{domain}')
        # define a default value for the saml attributes
        kwargs.setdefault('attributes', json.dumps({
            'organization_id': 'SAMPLE_ORG_ID',
            'portal_id': 'SAMPLE_PORTAL_ID',
        }))
        # define a default value for the saml deeplinks
        target = 'https://somesite.salesforce.com/%(target)s'
        kwargs.setdefault('links', json.dumps({
            # NOTE: This doesn't match the SALESFORCE_ACS,
            # but it will in production.
            r'portal/(?P<target>\w+)': target,
        }))
        super(
            SalesForcePortalAssertionCanonicalProcessorTestCase,
            self).setup_saml_sp(**kwargs)


class ExpensifyAssertionTestCase(GoogleAppsAssertionTestCase):
    """Test SAML Assertions for an Expensify Service Point."""

    ACS_URL = (
        'https://www.expensify.com/authentication/saml/loginCallback?'
        'domain=example.com')
    SAML_REQUEST = codex.deflate_and_base64_encode("""
<?xml version="1.0" encoding="UTF-8"?>
<samlp:AuthnRequest xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol"
                    xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion"
                    ID="fsdajhfdskjhfdajfadfhksfhdklfhkfjshlfsahkdhsa"
                    Version="2.0"
                    IssueInstant="2015-09-24T15:40:06Z"
                    Destination="https://www.expensify.com/authentication/saml/loginCallback?domain=example.com"
                    ProtocolBinding="urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST"
                    AssertionConsumerServiceURL="https://www.expensify.com/authentication/saml/loginCallback?domain=example.com"
                    >
    <saml:Issuer>https://www.expensify.com</saml:Issuer>
    <samlp:NameIDPolicy Format="urn:oasis:names:tc:SAML:1.1:nameid-format:emailAddress"
                        AllowCreate="true"
                        />
    <samlp:RequestedAuthnContext Comparison="exact">
        <saml:AuthnContextClassRef>urn:oasis:names:tc:SAML:2.0:ac:classes:PasswordProtectedTransport</saml:AuthnContextClassRef>
    </samlp:RequestedAuthnContext>
</samlp:AuthnRequest>
""")  # noqa
    RELAY_STATE = (
        'https://www.expensify.com/authentication/saml/login.php')
    REQUEST_DATA = {
        'SAMLRequest': SAML_REQUEST,
        'RelayState': RELAY_STATE,
    }
    PROCESSOR = 'ubuntu_sso_saml.processors.ExpensifyProcessor'


class ExpensifyAssertionCanonicalProcessorTestCase(
        CanonicalProcessorBaseTestCase, ExpensifyAssertionTestCase):
    pass


class CornerstoneAssertionTestCase(GoogleAppsAssertionTestCase):
    """Test SAML Assertions for an Expensify Service Point."""

    ACS_URL = 'https://site.csod.com/samldefault.aspx'
    SAML_REQUEST = base64.b64encode("""
<?xml version="1.0" encoding="UTF-8"?>
<AuthnRequest xmlns:xsd="http://www.w3.org/2001/XMLSchema"
              xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
              ID="THE_ID"
              Version="2.0"
              IssueInstant="2016-01-25T15:18:25.9931866Z"
              Destination="https://login.staging.ubuntu.com/saml"
              ProtocolBinding="urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST"
              AssertionConsumerServiceURL="https://site.csod.com/samldefault.aspx"
              xmlns="urn:oasis:names:tc:SAML:2.0:protocol"
              >
    <Issuer xmlns="urn:oasis:names:tc:SAML:2.0:assertion">https://site.csod.com</Issuer>
    <Signature xmlns="http://www.w3.org/2000/09/xmldsig#">
        <SignedInfo>
            <CanonicalizationMethod Algorithm="http://www.w3.org/2001/10/xml-exc-c14n#" />
            <SignatureMethod Algorithm="http://www.w3.org/2001/04/xmldsig-more#rsa-sha256" />
            <Reference URI="#_THE_ID">
                <Transforms>
                    <Transform Algorithm="http://www.w3.org/2000/09/xmldsig#enveloped-signature" />
                    <Transform Algorithm="http://www.w3.org/2001/10/xml-exc-c14n#" />
                </Transforms>
                <DigestMethod Algorithm="http://www.w3.org/2001/04/xmlenc#sha256" />
                <DigestValue>DIGEST_VALUE</DigestValue>
            </Reference>
        </SignedInfo>
        <SignatureValue>SIGNATURE_VALUE</SignatureValue>
    </Signature>
    <NameIDPolicy Format="urn:oasis:names:tc:SAML:1.1:nameid-format:unspecified" />
</AuthnRequest>
""")  # noqa
    RELAY_STATE = 'https://site.csod.com/'
    REQUEST_DATA = {
        'SAMLRequest': SAML_REQUEST,
        'RelayState': RELAY_STATE,
    }
    PROCESSOR = 'ubuntu_sso_saml.processors.CornerstoneProcessor'


class CornerstoneAssertionCanonicalProcessorTestCase(
        CanonicalProcessorBaseTestCase, CornerstoneAssertionTestCase):
    pass


class BlueJeansAssertionTestCase(GoogleAppsAssertionTestCase):
    """Test SAML Assertions for an BlueJeans Service Point."""

    ACS_URL = 'https://bluejeans.com/sso/saml2/'
    SAML_REQUEST = codex.deflate_and_base64_encode("""
<?xml version="1.0" encoding="UTF-8"?>
<saml2p:AuthnRequest xmlns:saml2p="urn:oasis:names:tc:SAML:2.0:protocol"
                     AssertionConsumerServiceURL="https://bluejeans.com/sso/saml2/"
                     ID="THE_ID"
                     IssueInstant="2016-06-29T18:30:45.083Z"
                     ProtocolBinding="urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST"
                     ProviderName="Blue Jeans Network"
                     Version="2.0"
                     >
    <saml2:Issuer xmlns:saml2="urn:oasis:names:tc:SAML:2.0:assertion">http://samlsp.bluejeans.com</saml2:Issuer>
</saml2p:AuthnRequest>
""")  # noqa
    RELAY_STATE = 'http://samlsp.bluejeans.com/'
    REQUEST_DATA = {
        'SAMLRequest': SAML_REQUEST,
        'RelayState': RELAY_STATE,
    }
    PROCESSOR = 'ubuntu_sso_saml.processors.BlueJeansProcessor'

    def test_assertion_contains_email_attribute(self):
        self.setup_saml_sp()
        preferred = self.setup_saml_emails()

        samlresponse = self.do_saml_request()
        expected = ('<saml:Attribute Name="email"><saml:AttributeValue>'
                    '{}</saml:AttributeValue></saml:Attribute>'.format(
                        preferred))
        self.assertIn(expected, samlresponse)

    def test_assertion_contains_username_attribute(self):
        self.setup_saml_sp()
        self.setup_saml_emails()

        samlresponse = self.do_saml_request()
        expected = ('<saml:Attribute Name="username"><saml:AttributeValue>'
                    '{}</saml:AttributeValue></saml:Attribute>'.format(
                        self.account.person_name))
        self.assertIn(expected, samlresponse)


class BlueJeansAssertionCanonicalProcessorTestCase(
        CanonicalProcessorBaseTestCase, BlueJeansAssertionTestCase):

    def setup_saml_sp(self, **kwargs):
        # define a default value for the saml attributes
        kwargs.setdefault('attributes', json.dumps({
            'email': '{{subject}}',
            'username': '{{person_name}}',
        }))
        super(BlueJeansAssertionCanonicalProcessorTestCase,
              self).setup_saml_sp(**kwargs)


class RocketchatAssertionTestCase(GoogleAppsAssertionTestCase):
    """
    Test SAML Assertions for a RocketChat Service Point.

    This class tests a specific BASE_URL but can be used to test other
    RocketChat servers just by subclassing and changing the BASE_URL.
    """

    BASE_URL = 'https://rocket.ubuntu.com'
    PROCESSOR = 'ubuntu_sso_saml.processors.RocketchatProcessor'

    def setUp(self, *args, **kwargs):
        self.ACS_URL = '%s/_saml/validate/openidp' % self.BASE_URL
        self.SAML_REQUEST = codex.deflate_and_base64_encode("""
<?xml version="1.0" encoding="UTF-8"?>
<samlp:AuthnRequest xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol"
                    ID="THE_ID"
                    Version="2.0"
                    IssueInstant="2016-09-08T21:40:15.988Z"
                    ProtocolBinding="urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST"
                    AssertionConsumerServiceURL="%(base_url)s/_saml/validate/openidp"
                    Destination="http://login.ubuntu.com/+saml"
                    >
    <saml:Issuer xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion">%(base_url)s</saml:Issuer>
    <samlp:NameIDPolicy xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol"
                        Format="urn:oasis:names:tc:SAML:1.1:nameid-format:emailAddress"
                        AllowCreate="true"
                        />
    <samlp:RequestedAuthnContext xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol"
                                 Comparison="exact"
                                 >
        <saml:AuthnContextClassRef xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion">urn:oasis:names:tc:SAML:2.0:ac:classes:PasswordProtectedTransport</saml:AuthnContextClassRef>
    </samlp:RequestedAuthnContext>
</samlp:AuthnRequest>


""" % {'base_url': self.BASE_URL})      # noqa
        self.RELAY_STATE = '%s' % self.BASE_URL
        self.REQUEST_DATA = {
            'SAMLRequest': self.SAML_REQUEST,
            'RelayState': self.RELAY_STATE,
        }
        super(RocketchatAssertionTestCase, self).setUp(*args, **kwargs)

    def test_assertion_contains_email_attribute(self):
        self.setup_saml_sp()
        preferred = self.setup_saml_emails()

        samlresponse = self.do_saml_request()
        expected = ('<saml:Attribute Name="email"><saml:AttributeValue>'
                    '{}</saml:AttributeValue></saml:Attribute>'.format(
                        preferred))
        self.assertIn(expected, samlresponse)

    def test_assertion_contains_username_attribute(self):
        self.setup_saml_sp()
        self.setup_saml_emails()

        samlresponse = self.do_saml_request()
        expected = ('<saml:Attribute Name="username"><saml:AttributeValue>'
                    '{}</saml:AttributeValue></saml:Attribute>'.format(
                        self.account.person_name))
        self.assertIn(expected, samlresponse)


class RocketchatAssertionCanonicalProcessorTestCase(
        CanonicalProcessorBaseTestCase, RocketchatAssertionTestCase):

    def setup_saml_sp(self, **kwargs):
        # define a default value for the saml attributes
        kwargs.setdefault('attributes', json.dumps({
            'email': '{{subject}}',
            'username': '{{person_name}}',
        }))
        super(RocketchatAssertionCanonicalProcessorTestCase,
              self).setup_saml_sp(**kwargs)


class RocketchatCanonicalAssertionTestCase(RocketchatAssertionTestCase):
    """Tests a rocketchat server under the below BASE_URL."""

    BASE_URL = 'https://rocket.canonical.com'


class RocketchatCanonicalAssertionCanonicalProcessorTestCase(
        RocketchatAssertionCanonicalProcessorTestCase,
        RocketchatCanonicalAssertionTestCase):
    pass


class LoopioAssertionTestCase(GoogleAppsAssertionTestCase):
    """Test SAML Assertions for a Loopio Service Point."""

    ACS_URL = (
        'https://canonical.loopio.com/sso_saml/module.php/'
        'saml/sp/saml2-acs.php/loopio-sp')
    SAML_REQUEST = codex.deflate_and_base64_encode("""
<?xml version="1.0" encoding="UTF-8"?>
<samlp:AuthnRequest xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol"
                    xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion"
                    ID="THE_ID"
                    Version="2.0"
                    IssueInstant="2017-01-25T15:44:20Z"
                    Destination="https://login.staging.ubuntu.com/saml"
                    AssertionConsumerServiceURL="https://canonical.loopio.com/sso_saml/module.php/saml/sp/saml2-acs.php/loopio-sp"
                    ProtocolBinding="urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST"
                    >
    <saml:Issuer>loopio-sp</saml:Issuer>
    <samlp:NameIDPolicy Format="urn:oasis:names:tc:SAML:1.1:nameid-format:emailAddress"
                        AllowCreate="true"
                        />
</samlp:AuthnRequest>
""")  # noqa
    RELAY_STATE = 'https://canonical.loopio.com/'
    REQUEST_DATA = {
        'SAMLRequest': SAML_REQUEST,
        'RelayState': RELAY_STATE,
    }
    PROCESSOR = 'ubuntu_sso_saml.processors.LoopioProcessor'


class LoopioAssertionCanonicalProcessorTestCase(
        CanonicalProcessorBaseTestCase, LoopioAssertionTestCase):
    pass


class InfoProcessorTestCase(BaseProcessorTestCase):
    """Tests the InfoProcessor for diagnostic purposes.

    InfoProcessor just extracts data from the SAML request with no expectation
    to actually process it and generate a proper SAML response.
    """

    SAML_PAYLOAD = ("""
<?xml version="1.0" encoding="UTF-8"?>
<samlp:AuthnRequest xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol"
                    ID="THE_ID"
                    Version="2.0"
                    IssueInstant="2016-09-08T21:40:15.988Z"
                    ProtocolBinding="urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST"
                    AssertionConsumerServiceURL="https://saml.example.com/saml_acs_url"
                    Destination="http://login.ubuntu.com/+saml"
                    >
    <saml:Issuer xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion">https://example.com</saml:Issuer>
    <samlp:NameIDPolicy xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol"
                        Format="urn:oasis:names:tc:SAML:1.1:nameid-format:emailAddress"
                        AllowCreate="true"
                        />
    <samlp:RequestedAuthnContext xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol"
                                 Comparison="exact"
                                 >
        <saml:AuthnContextClassRef xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion">urn:oasis:names:tc:SAML:2.0:ac:classes:PasswordProtectedTransport</saml:AuthnContextClassRef>
    </samlp:RequestedAuthnContext>
</samlp:AuthnRequest>


""")  # noqa
    RELAY_STATE = 'https://example.com/'
    ACS_URL = "https://saml.example.com/saml_acs_url"
    # Note REQUEST_DATA is incomplete and needs a SAMLRequest added by
    # individual test methods.
    REQUEST_DATA = {
        'RelayState': RELAY_STATE,
    }
    # This processor is not really used in the tests, it's just here so
    # the setUp method can prepare at least one saml_remote.
    PROCESSOR = 'ubuntu_sso_saml.processors.GoogleAppsProcessor'

    def get_response_to_saml_request(self):
        # do twofactor dance
        self.client.post(reverse('twofactor'), {'oath_token': '123456'})
        # do saml dance. This should fail and get us a plain response.
        data = self.get_request_data()
        response = self.client.get('/+saml', data=data, follow=True)
        # Common checks
        self.assertTemplateUsed(
            response, 'ubuntu_sso_saml/saml_remote_not_found.html')
        self.assertIn("Unable to process SAML request", response.content)
        return response

    def test_response_chops_off_url_path(self):
        """Test the message contains only the URL scheme and host.
        The URL path components may expose secrets so we want to chop them off
        and leave only the scheme/host sections.
        """
        self.REQUEST_DATA['SAMLRequest'] = base64.b64encode(
            self.SAML_PAYLOAD)
        self.setup_saml_sp()
        self.setup_saml_emails()

        samlresponse = self.get_response_to_saml_request()
        self.assertNotIn("saml_acs_url", samlresponse.content)

    def test_plain_b64_saml_request(self):
        self.REQUEST_DATA['SAMLRequest'] = base64.b64encode(
            self.SAML_PAYLOAD)
        self.setup_saml_sp()
        self.setup_saml_emails()

        samlresponse = self.get_response_to_saml_request()
        self.assertIn("https://saml.example.com", samlresponse.content)

    def test_deflated_saml_request(self):
        self.REQUEST_DATA['SAMLRequest'] = codex.deflate_and_base64_encode(
            self.SAML_PAYLOAD)
        self.setup_saml_sp()
        self.setup_saml_emails()

        samlresponse = self.get_response_to_saml_request()
        self.assertIn("https://saml.example.com", samlresponse.content)

    def test_non_encoded_saml_request(self):
        # This is a form of bogus SAML request; if it's not b64-encoded it
        # counts as a bad request anyway.
        self.REQUEST_DATA['SAMLRequest'] = self.SAML_PAYLOAD
        self.setup_saml_sp()
        self.setup_saml_emails()

        samlresponse = self.get_response_to_saml_request()
        self.assertIn("a non-SAML request", samlresponse.content)

    def test_bogus_SAML_request(self):
        self.REQUEST_DATA['SAMLRequest'] = base64.b64encode(
            "Some totally bogus data")
        self.setup_saml_sp()
        self.setup_saml_emails()

        samlresponse = self.get_response_to_saml_request()
        self.assertIn("a non-SAML request", samlresponse.content)


class CanonicalProcessorTestCase(CanonicalProcessorBaseTestCase):
    SAML_PAYLOAD = ("""
<?xml version="1.0" encoding="UTF-8"?>
<samlp:AuthnRequest xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol"
                    ID="THE_ID"
                    Version="2.0"
                    IssueInstant="2016-09-08T21:40:15.988Z"
                    ProtocolBinding="urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST"
                    AssertionConsumerServiceURL="{acs_url}"
                    Destination="{audience}"
                    >
    <saml:Issuer xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion">{issuer}</saml:Issuer>
    <samlp:NameIDPolicy xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol"
                        Format="urn:oasis:names:tc:SAML:1.1:nameid-format:emailAddress"
                        AllowCreate="true"
                        />
    <samlp:RequestedAuthnContext xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol"
                                 Comparison="exact"
                                 >
        <saml:AuthnContextClassRef xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion">urn:oasis:names:tc:SAML:2.0:ac:classes:PasswordProtectedTransport</saml:AuthnContextClassRef>
    </samlp:RequestedAuthnContext>
</samlp:AuthnRequest>


""")  # noqa

    SAML_PAYLOAD_NO_ACS_URL = ("""
<?xml version="1.0" encoding="UTF-8"?>
<samlp:AuthnRequest xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol"
                    ID="THE_ID"
                    Version="2.0"
                    IssueInstant="2016-09-08T21:40:15.988Z"
                    ProtocolBinding="urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST"
                    Destination="{audience}"
                    >
    <saml:Issuer xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion">{issuer}</saml:Issuer>
    <samlp:NameIDPolicy xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol"
                        Format="urn:oasis:names:tc:SAML:1.1:nameid-format:emailAddress"
                        AllowCreate="true"
                        />
    <samlp:RequestedAuthnContext xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol"
                                 Comparison="exact"
                                 >
        <saml:AuthnContextClassRef xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion">urn:oasis:names:tc:SAML:2.0:ac:classes:PasswordProtectedTransport</saml:AuthnContextClassRef>
    </samlp:RequestedAuthnContext>
</samlp:AuthnRequest>

""")  # noqa

    ACS_URL = 'https://saml.example.com/saml_acs_url'
    REQUEST_ISSUER = 'https://saml.example.com/metadata'
    AUDIENCE = 'https://login.ubuntu.com/+saml'
    RELAY_STATE = 'https://example.com/'
    # Note REQUEST_DATA is incomplete and needs a SAMLRequest added by
    # individual test methods.
    REQUEST_DATA = {
        'RelayState': RELAY_STATE,
    }

    def assert_successful_saml_response(self, samlresponse):
        self.assertIn(
            '<samlp:Status>'
            '<samlp:StatusCode '
            'Value="urn:oasis:names:tc:SAML:2.0:status:Success">'
            '</samlp:StatusCode>'
            '</samlp:Status>',
            samlresponse)

    def assert_restricted_saml_response(self, samlresponse):
        self.assertIn(
            '<saml:AudienceRestriction><saml:Audience>{}'
            '</saml:Audience></saml:AudienceRestriction>'.format(
                self.AUDIENCE),
            samlresponse)

    def assert_unrestricted_saml_response(self, samlresponse):
        self.assertNotIn('<saml:AudienceRestriction>', samlresponse)

    def assert_saml_response_email(self, samlresponse, email):
        self.assertIn(email, samlresponse)

    def get_request_data(self, compressed=True, has_acs_url=True):
        data = dict(**self.REQUEST_DATA)
        params = {
            'audience': self.AUDIENCE,
            'issuer': self.REQUEST_ISSUER,
            'relay_state': self.RELAY_STATE,
        }
        if has_acs_url:
            params['acs_url'] = self.ACS_URL
            payload = self.SAML_PAYLOAD.format(**params)
        else:
            payload = self.SAML_PAYLOAD_NO_ACS_URL.format(**params)
        if compressed:
            data['SAMLRequest'] = codex.deflate_and_base64_encode(payload)
        else:
            data['SAMLRequest'] = base64.b64encode(payload)
        return data

    def test_decode_request_compressed(self):
        # make sure there is a saml config defined for the SP
        self.setup_saml_sp()

        samlresponse = self.do_saml_request()
        self.assert_successful_saml_response(samlresponse)

    def test_decode_request_uncompressed(self):
        # make sure there is a saml config defined for the SP
        self.setup_saml_sp()

        data = self.get_request_data(compressed=False)
        samlresponse = self.do_saml_request(data=data)
        self.assert_successful_saml_response(samlresponse)

    def test_format_assertion_restricted(self):
        # make sure there is a saml config with an audience defined for the SP
        self.setup_saml_sp(audience=self.AUDIENCE)

        samlresponse = self.do_saml_request()
        self.assert_successful_saml_response(samlresponse)
        self.assert_restricted_saml_response(samlresponse)

    def test_format_assertion_unrestricted(self):
        # make sure there is no audience defined for the SP
        self.setup_saml_sp(audience='')

        samlresponse = self.do_saml_request()
        self.assert_successful_saml_response(samlresponse)
        self.assert_unrestricted_saml_response(samlresponse)

    def test_authnrequest_unknown_sp(self):
        data = self.get_request_data()
        response = self.client.get('/+saml', data=data, follow=True)
        self.assertTemplateUsed(
            response, 'ubuntu_sso_saml/saml_remote_not_found.html')
        self.assertIn("Unable to process SAML request", response.content)

    def test_authnrequest_for_disabled_sp(self):
        self.setup_saml_sp(enabled=False)

        data = self.get_request_data()
        response = self.client.get('/+saml', data=data, follow=True)
        self.assertTemplateUsed(
            response, 'ubuntu_sso_saml/saml_remote_not_found.html')
        self.assertIn("Unable to process SAML request", response.content)

    def test_authnrequest_without_acs_url_unrestricted(self):
        self.setup_saml_sp()

        data = self.get_request_data(has_acs_url=False)
        samlresponse = self.do_saml_request(data=data)
        self.assert_successful_saml_response(samlresponse)
        self.assert_unrestricted_saml_response(samlresponse)

    def test_authnrequest_without_acs_url_restricted(self):
        self.setup_saml_sp(audience=self.AUDIENCE)

        data = self.get_request_data(has_acs_url=False)
        samlresponse = self.do_saml_request(data=data)
        self.assert_successful_saml_response(samlresponse)
        self.assert_restricted_saml_response(samlresponse)

    def test_authnrequest_without_acs_url_posts_to_spconfig_acs_url(self):
        self.setup_saml_sp()

        data = self.get_request_data(has_acs_url=False)
        # do twofactor dance
        self.client.post(reverse('twofactor'), {'oath_token': '123456'})
        # do saml dance
        if data is None:
            data = self.get_request_data()
        response = self.client.get('/+saml', data=data, follow=True)
        self.assertEqual(response.status_code, 200)
        tree = PyQuery(response.content)
        forms = tree.find('form')
        assert len(forms) == 1
        self.assertEqual(forms[0].get('action'), self.ACS_URL)

    def test_response_with_system_issuer_not_authnrequest_issuer(self):
        self.setup_saml_sp()

        data = self.get_request_data(has_acs_url=False)
        samlresponse = self.do_saml_request(data=data)
        self.assertNotIn(
            'https://saml.example.com/metadata</saml:Issuer>', samlresponse)
        self.assertIn(
            '{}</saml:Issuer>'.format(
                saml2idp_metadata.SAML2IDP_CONFIG['issuer']),
            samlresponse)

    def test_subject_confirmation_recipient_when_no_request_acs_url(self):
        sp_config = self.setup_saml_sp()

        data = self.get_request_data(has_acs_url=False)
        samlresponse = self.do_saml_request(data=data)
        self.assertNotIn('Recipient="None"', samlresponse)
        self.assertIn(
            'Recipient="{}"></saml:SubjectConfirmationData>'.format(
                sp_config.acs_url),
            samlresponse)

    def test_enabled_db_spconfig_overrides_settings(self):
        # make sure there is an "old/custom" processor handling this request
        self.ACS_URL = google_apps.GOOGLE_APPS_ACS
        self.REQUEST_DATA = google_apps.REQUEST_DATA
        remotes = {
            '00_foobar': {
                'acs_url': self.ACS_URL,
                'processor': self.PROCESSOR,
            },
            '01_foobar': {
                'acs_url': self.ACS_URL,
                'processor': 'ubuntu_sso_saml.processors.GoogleAppsProcessor',
            },
        }
        self.patch(saml2idp_metadata, 'SAML2IDP_REMOTES', new=remotes)

        self.setup_saml_sp(email_pattern='{username}+test@{domain}',
                           prefer_canonical_email=True, enabled=True)
        self.setup_saml_emails()

        samlresponse = self.do_saml_request()
        self.assert_successful_saml_response(samlresponse)
        self.assert_saml_response_email(
            samlresponse, 'first.last+test@canonical.com')

    def test_disabled_db_spconfig_falls_back_to_settings(self):
        # make sure there is an "old/custom" processor handling this request
        self.ACS_URL = google_apps.GOOGLE_APPS_ACS
        self.REQUEST_DATA = google_apps.REQUEST_DATA
        remotes = {
            '00_foobar': {
                'acs_url': self.ACS_URL,
                'processor': self.PROCESSOR,
            },
            '01_foobar': {
                'acs_url': self.ACS_URL,
                'processor': 'ubuntu_sso_saml.processors.GoogleAppsProcessor',
            },
        }
        self.patch(saml2idp_metadata, 'SAML2IDP_REMOTES', new=remotes)

        self.setup_saml_sp(email_pattern='{username}+test@{domain}',
                           prefer_canonical_email=True, enabled=False)
        self.setup_saml_emails()

        samlresponse = self.do_saml_request()
        self.assert_successful_saml_response(samlresponse)
        self.assert_saml_response_email(
            samlresponse, 'first.last@canonical.com')
        self.assertNotIn('first.last+test@canonical.com', samlresponse)

    def test_authnrequest_csrf_exempt(self):
        client = Client(enforce_csrf_checks=True)
        data = self.get_request_data()
        response = client.post('/+saml', data=data, follow=False)
        self.assertEqual(response.status_code, 302)

    def test_session_index_in_response(self):
        self.setup_saml_sp()

        data = self.get_request_data()
        samlresponse = self.do_saml_request(data=data)
        self.assertIn('SessionIndex=', samlresponse)

    def test_canonical_email_is_preferred(self):
        self.setup_saml_sp(prefer_canonical_email=True)
        preferred = self.setup_saml_emails()
        samlresponse = self.do_saml_request()
        # verify @canonical.com email address is returned despite it's not the
        # preferredemail
        self.assertIn('first.last@canonical.com', samlresponse)
        self.assertNotEqual('first.last@canonical.com', preferred)

    def test_canonical_email_is_preferred_despite_case(self):
        email = 'FIRST.LAST@CANONICAL.COM'
        self.setup_saml_sp(prefer_canonical_email=True)
        preferred = self.setup_saml_emails(email=email)
        samlresponse = self.do_saml_request()
        # verify @canonical.com email address is returned despite it's not the
        # preferredemail
        self.assertIn(email, samlresponse)
        self.assertNotEqual(email, preferred)

    def test_only_long_form_canonical_email_is_preferred(self):
        email = self.login_email
        # add non-preferred long form email address
        EmailAddress.objects.create(
            account=self.account, email='first.last+alias@canonical.com',
            status=EmailStatus.VALIDATED)
        self.setup_saml_sp(prefer_canonical_email=True)
        preferred = self.setup_saml_emails()
        self.assertEqual(preferred, email)
        self.assertNotEqual(preferred, 'first.last@canonical.com')

        samlresponse = self.do_saml_request()
        # verify @canonical.com email address is returned despite it's not the
        # preferredemail
        self.assertIn('first.last@canonical.com', samlresponse)
        self.assertNotIn(email, samlresponse)

    def test_canonical_email_is_not_preferred(self):
        self.setup_saml_sp(prefer_canonical_email=False)
        preferred = self.setup_saml_emails()
        samlresponse = self.do_saml_request()
        # verify @canonical.com email address is *not* returned despite because
        # the flag was not set
        self.assertIn(preferred, samlresponse)
        self.assertNotEqual('first.last@canonical.com', preferred)

    def test_non_preferred_short_form_canonical_email_ignored(self):
        # start off a clean slate
        short_email = EmailAddress.objects.get(
            account=self.account, email='test@canonical.com')
        short_email.status = EmailStatus.VALIDATED
        short_email.save()
        EmailAddress.objects.create(account=self.account, email='foo@foo.com',
                                    status=EmailStatus.PREFERRED)
        self.login_email = 'foo@foo.com'

        # make sure the short email it's not the preferred address
        self.setup_saml_sp(prefer_canonical_email=True)
        preferred = self.setup_saml_emails()
        assert preferred == 'foo@foo.com'
        # make sure there's no long form @canonical email
        EmailAddress.objects.filter(
            account=self.account, email='first.last@canonical.com').delete()

        samlresponse = self.do_saml_request()
        # verify @canonical.com email address is *not* returned
        self.assertIn(preferred, samlresponse)
        self.assertNotEqual('first.last@canonical.com', preferred)
        self.assertNotEqual('test@canonical.com', preferred)

    def test_preferred_short_form_canonical_email_ignored(self):
        # start off a clean slate
        short_email = EmailAddress.objects.get(
            account=self.account, email='test@canonical.com')
        short_email.status = EmailStatus.PREFERRED
        short_email.save()
        EmailAddress.objects.create(account=self.account, email='foo@foo.com',
                                    status=EmailStatus.VALIDATED)
        self.login_email = 'foo@foo.com'

        # make sure the short email is the preferred address
        self.setup_saml_sp(prefer_canonical_email=True)
        preferred = self.setup_saml_emails()
        assert preferred == 'test@canonical.com'
        # make sure there's no long form @canonical email
        EmailAddress.objects.filter(
            account=self.account, email='first.last@canonical.com').delete()

        samlresponse = self.do_saml_request()
        # verify @canonical.com email address is *not* returned
        self.assertIn(preferred, samlresponse)
        self.assertEqual('test@canonical.com', preferred)
        self.assertNotEqual('first.last@canonical.com', preferred)
        self.assertNotEqual('foo@foo.com', preferred)

    def test_non_canonical_email(self):
        # prepare account for SAML
        account = self.factory.make_account()
        preferred = account.preferredemail.email
        self.factory.add_account_to_team(account, self.saml2team)
        self.setup_saml_sp(prefer_canonical_email=True)

        # make sure account has only one email
        assert account.emailaddress_set.count() == 1
        # and it's not a @canonical.com one
        assert '@canonical.com' not in preferred

        assert self.client.login(
            username=preferred, password=DEFAULT_USER_PASSWORD)

        samlresponse = self.do_saml_request()
        self.assertIn(preferred, samlresponse)

    def test_multiple_long_form_canonical_emails(self):
        self.setup_saml_sp(prefer_canonical_email=True)
        self.setup_saml_emails()
        # change preferred email to non-standard long form email address
        preferred = EmailAddress.objects.create(
            account=self.account, email='other.name@canonical.com',
            status=EmailStatus.PREFERRED)
        email = self.account.preferredemail
        email.status = EmailStatus.VALIDATED
        email.save()
        # make sure account shows new email as preferred
        account = Account.objects.get(pk=self.account.pk)
        assert account.preferredemail == preferred

        samlresponse = self.do_saml_request()
        # verify @canonical.com email address is returned despite it's not the
        # preferredemail
        self.assertIn('other.name@canonical.com', samlresponse)
        self.assertNotIn('first.last@canonical.com', samlresponse)

    def test_validate_bogus_acs_domain(self):
        self.setup_saml_sp(acs_domain='bogus.domain')

        data = self.get_request_data()
        response = self.client.get('/+saml', data=data, follow=True)
        self.assertTemplateUsed(
            response, 'ubuntu_sso_saml/saml_remote_not_found.html')
        self.assertIn("Unable to process SAML request", response.content)

    def test_validate_partial_domain_match(self):
        acs_url = self.ACS_URL.replace('example.com', 'example.com.ar')
        self.setup_saml_sp(acs_url=acs_url, acs_domain='example.com')

        data = self.get_request_data()
        response = self.client.get('/+saml', data=data, follow=True)
        self.assertTemplateUsed(
            response, 'ubuntu_sso_saml/saml_remote_not_found.html')
        self.assertIn("Unable to process SAML request", response.content)

    def test_validate_valid_acs_domain(self):
        self.setup_saml_sp(acs_domain='example.com')

        samlresponse = self.do_saml_request()
        self.assert_successful_saml_response(samlresponse)

    def test_validates_domain_ok_when_acs_domain_empty(self):
        self.setup_saml_sp()

        samlresponse = self.do_saml_request()
        self.assert_successful_saml_response(samlresponse)

    def test_setup_links_from_spconfig(self):
        links = {
            'prefix/(?P<target>\w+)': 'http://example.com/%(target)s',
        }
        self.setup_saml_sp(links=json.dumps(links))

        data = self.get_request_data()
        response = self.client.get(
            '/+saml/init/prefix/foo', data=data, follow=True)

        tree = PyQuery(response.content)
        relsttag = tree.find('input[name="RelayState"]')
        assert len(relsttag) == 1
        relaystate = relsttag[0].get('value')
        self.assertEqual(relaystate, 'http://example.com/foo')

        samlresponse = get_saml_response(response)
        self.assert_successful_saml_response(samlresponse)

    @override_settings(
        ROOT_URLCONF='ubuntu_sso_saml.tests.urls_deeplinking_fallback'
    )
    def test_deeplink_fallback_to_settings(self):
        target = 'http://example.com/%(target)s'
        saml2idp_metadata.SAML2IDP_REMOTES['foobar'].update({
            'links': {
                r'prefix/(?P<target>\w+)': target,
            },
        })

        # create the entry in the db but doesn't let it match any links
        self.setup_saml_sp()

        data = self.get_request_data()
        response = self.client.get(
            '/+saml/init/prefix/foo', data=data, follow=True)

        tree = PyQuery(response.content)
        relsttag = tree.find('input[name="RelayState"]')
        assert len(relsttag) == 1
        relaystate = relsttag[0].get('value')
        self.assertEqual(relaystate, 'http://example.com/foo')

        samlresponse = get_saml_response(response)
        self.assert_successful_saml_response(samlresponse)

    @override_settings(
        SAML_NOT_BEFORE_MINUTES=2*60, SAML_NOT_AFTER_MINUTES=10)
    @patch('saml2idp.base.time.time')
    def test_saml_time_conditions(self, mock_time):
        now = time.time()
        mock_time.time.return_value = now

        self.setup_saml_sp()
        self.setup_saml_emails()
        samlresponse = self.do_saml_request()

        doc = minidom.parseString(samlresponse)
        conditions = doc.getElementsByTagName('saml:Conditions')
        assert len(conditions) == 1
        attrs = dict(conditions[0].attributes.items())
        not_before = get_time_string(-2*60*60)
        not_after = get_time_string(10*60)
        self.assertEqual(attrs['NotBefore'], not_before)
        self.assertEqual(attrs['NotOnOrAfter'], not_after)

    def test_team_restricted_access_ok(self):
        self.team = self.factory.make_team("kaboo")
        self.factory.add_account_to_team(self.account, self.team)
        self.setup_saml_sp(allowed_teams="kaboo")
        samlresponse = self.do_saml_request()
        self.assert_successful_saml_response(samlresponse)

    def test_team_restricted_access_ok_many_teams(self):
        self.team = self.factory.make_team("kaboo")
        self.factory.add_account_to_team(self.account, self.team)
        self.setup_saml_sp(allowed_teams="foo, kaboo , zoo")
        samlresponse = self.do_saml_request()
        self.assert_successful_saml_response(samlresponse)

    def test_no_team_restrictions_access_ok(self):
        """Explicitly check that SP with no allowed_teams allows all."""
        self.team = self.factory.make_team("kaboo")
        self.factory.add_account_to_team(self.account, self.team)
        self.setup_saml_sp(allowed_teams='')
        samlresponse = self.do_saml_request()
        self.assert_successful_saml_response(samlresponse)

    def test_team_restricted_access_denied(self):
        # Add a team restriction to the rp
        self.team = self.factory.make_team("kaboo")
        self.factory.add_account_to_team(self.account, self.team)
        self.setup_saml_sp(allowed_teams=["not_here"])

        data = self.get_request_data()
        response = self.client.get('/+saml', data=data, follow=True)
        self.assertTemplateUsed(
            response, 'ubuntu_sso_saml/saml_remote_not_found.html')
        self.assertIn("Unable to process SAML request", response.content)

    def test_team_restricted_access_denied_user_with_no_team(self):
        # Add a team restriction to the rp
        self.setup_saml_sp(allowed_teams=["not_here"])

        data = self.get_request_data()
        response = self.client.get('/+saml', data=data, follow=True)
        self.assertTemplateUsed(
            response, 'ubuntu_sso_saml/saml_remote_not_found.html')
        self.assertIn("Unable to process SAML request", response.content)

    def test_team_restricted_access_anonymous_user_no_oops(self):
        # Add a team restriction to the rp
        self.setup_saml_sp(allowed_teams=["not_here"])

        # Log out so we're effectively AnonymousUser
        self.client.logout()
        data = self.get_request_data()
        response = self.client.get('/+saml', data=data, follow=True)
        # I should be sent to the login page
        query_params = urllib.urlencode({'next': "/saml/process"})
        self.assertRedirects(response, reverse('login') + "?" + query_params)

    def test_default_certificate_in_assertion(self):
        self.setup_saml_sp()

        data = self.get_request_data()
        response = self.client.get('/+saml', data=data, follow=True)
        samlresponse = get_saml_response(response)

        cert_filename = saml2idp_metadata.SAML2IDP_CONFIG['certificate_file']

        self.assertIn(cert_string_from_file(cert_filename), samlresponse)

    def test_custom_certificate_in_assertion(self):
        cert_filename = saml2idp_metadata.SAML2IDP_CONFIG['certificate_file']
        custom_cert_filename = find_cert_path(CUSTOM_TEST_CERTIFICATE)
        with open(custom_cert_filename) as custom_cert_file:
            custom_cert_string = custom_cert_file.read()
        # Setup the SP with the delimited certificate
        self.setup_saml_sp(certificate=custom_cert_string)
        # Keep the non-delimited certificate for asserting later.
        custom_cert_string = cert_string_from_file(custom_cert_filename)

        data = self.get_request_data()
        response = self.client.get('/+saml', data=data, follow=True)
        samlresponse = get_saml_response(response)
        # Ensure the assertion contains the custom cert, not the default one.
        self.assertIn(custom_cert_string, samlresponse)
        self.assertNotIn(cert_string_from_file(cert_filename), samlresponse)

    def test_openid_as_attribute(self):
        self.setup_saml_sp(attributes=json.dumps({
            'openid': '{{openid}}',
        }))
        self.setup_saml_emails()
        openid_url = 'http://testserver' + self.account.get_absolute_url()

        samlresponse = self.do_saml_request()
        expected = ('<saml:Attribute Name="openid"><saml:AttributeValue>'
                    '{}</saml:AttributeValue></saml:Attribute>'.format(
                        openid_url))
        self.assertIn(expected, samlresponse)
