from django.conf import settings
from django.test import TestCase


class SettingsTestCase(TestCase):

    def setUp(self):
        # Generate a bunch of items to add to the dict. These are just the
        # candidates to be injected in SAML2IDP_REMOTES.
        self.items = [
            ("item-%s " % num, "Item %s" % num) for num in range(200)
        ]

    def tearDown(self):
        # Once a test finishes running, clean up the setting so it doesn't
        # affect the next test. Just remove the keys we've possibly inserted.
        # Note the test does insert them all, so we can just try removing each
        # one.
        for item in self.items:
            settings.SAML2IDP_REMOTES.pop(item[0])

    def test_samlremotes_ordering(self):
        """
        Ensure SAML2IDP_REMOTES maintains order of appended items.

        Note the use of setUp and tearDown to create a list of items that'll be
        added to the setting to test its order-awareness. This is because we
        want to ensure proper clean-up after finishing (either pass or fail) to
        avoid polluting the settings for other tests.

        Also, note we can't simply replace the setting with e.g. a clean dict
        or an OrderedDict because we want to test the actual behavior of
        whatever SAML2IDP_REMOTES is implemented as. Replacing it with another
        object would render the test invalid.

        Finally, existing items in the setting should be accounted for when
        obtaining "order1" and so shouldn't affect the outcome of the test, and
        unless a remote is named "item-*", the final clean-up should also not
        render the existing settings invalid.

        """
        # Add a few items
        for item in self.items[:len(self.items)/2]:
            settings.SAML2IDP_REMOTES[item[0]] = item[1]

        # Substring match is a hack but easiest way to check
        # list of keys is in a larger set of keys in the same order
        order1 = "".join(settings.SAML2IDP_REMOTES.keys())

        # Now add more items and ensure the first string is still a subset of
        # the second, order included.
        for item in self.items[len(self.items)/2:]:
            settings.SAML2IDP_REMOTES[item[0]] = item[1]
        order2 = "".join(settings.SAML2IDP_REMOTES.keys())

        self.assertIn(order1, order2)
        self.assertTrue(order2.startswith(order1))
