import base64
import json

from django.test import TestCase
from django.test.client import RequestFactory

from ubuntu_sso_saml.models import SAMLConfig
from ubuntu_sso_saml.processors import CanonicalProcessor
from ubuntu_sso_saml.tests.helpers import find_cert_path
from ubuntu_sso_saml.utils import (
    certificate_from_sp_config,
    get_config_and_link_for_resource,
    get_config_for_acs,
    get_config_from_processor,
    get_deeplink_url,
)


class GetConfigTestCase(TestCase):

    def test_get_config_for_acs(self):
        conf = SAMLConfig.objects.create(
            acs_url='http://localhost', enabled=True)
        self.assertEqual(get_config_for_acs('http://localhost'), conf)

    def test_get_config_for_acs_when_disabled(self):
        SAMLConfig.objects.create(
            acs_url='http://localhost', enabled=False)
        self.assertIsNone(get_config_for_acs('http://localhost'))

    def test_get_config_for_acs_not_found(self):
        self.assertIsNone(get_config_for_acs('http://localhost'))

    def test_get_config_for_acs_by_issuer(self):
        conf = SAMLConfig.objects.create(
            acs_url='http://localhost/acs',
            issuer='http://localhost', enabled=True)
        self.assertEqual(
            get_config_for_acs(None, issuer='http://localhost'), conf)

    def test_get_config_from_processor(self):
        conf = SAMLConfig.objects.create(
            acs_url='https://example.com/acs', enabled=True)
        _payload = ("""
<?xml version="1.0" encoding="UTF-8"?>
<samlp:AuthnRequest xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol"
                    ID="THE_ID"
                    Version="2.0"
                    IssueInstant="2016-09-08T21:40:15.988Z"
                    ProtocolBinding="urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST"
                    AssertionConsumerServiceURL="{acs_url}"
                    >
    <saml:Issuer xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion">{issuer}</saml:Issuer>
    <samlp:NameIDPolicy xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol"
                        Format="urn:oasis:names:tc:SAML:1.1:nameid-format:emailAddress"
                        AllowCreate="true"
                        />
    <samlp:RequestedAuthnContext xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol"
                                 Comparison="exact"
                                 >
        <saml:AuthnContextClassRef xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion">urn:oasis:names:tc:SAML:2.0:ac:classes:PasswordProtectedTransport</saml:AuthnContextClassRef>
    </samlp:RequestedAuthnContext>
</samlp:AuthnRequest>
""")  # noqa
        payload = _payload.format(
            acs_url='https://example.com/acs',
            issuer='https://example.com/issuer')
        request = RequestFactory().get('')
        request.session = {
            'SAMLRequest': base64.b64encode(payload),
            'RelayState': None,
        }
        proc = CanonicalProcessor()
        proc.can_handle(request)
        self.assertEqual(get_config_from_processor(proc), conf)

    def test_get_config_and_link_for_resource(self):
        links = {
            'foo': 'http://example.com/foo',
        }
        conf = SAMLConfig.objects.create(
            acs_url='http://localhost', links=json.dumps(links),
            enabled=True)
        sp_config, link = get_config_and_link_for_resource('foo')
        self.assertEqual(sp_config, conf)
        self.assertEqual(link, ('foo', 'http://example.com/foo'))

    def test_get_config_and_link_for_resource_when_disabled(self):
        links = {
            'foo': 'http://example.com/foo',
        }
        SAMLConfig.objects.create(
            acs_url='http://localhost', links=json.dumps(links),
            enabled=False)
        sp_config, link = get_config_and_link_for_resource('foo')
        self.assertIsNone(sp_config)
        self.assertIsNone(link)

    def test_get_config_and_link_for_resource_not_found(self):
        links = {
            'foo': 'http://example.com/foo',
        }
        SAMLConfig.objects.create(
            acs_url='http://localhost', links=json.dumps(links))
        sp_config, link = get_config_and_link_for_resource('foo')
        self.assertIsNone(sp_config)
        self.assertIsNone(link)


class GetDeeplinkUrlTestCase(TestCase):

    def test_get_deeplink_url_simple_pattern(self):
        resource_name = 'prefix/foo'
        resource_pattern = 'prefix/(?P<target>.+)$'
        url_pattern = 'http://example.com/%(target)s'

        url = get_deeplink_url(resource_name, (resource_pattern, url_pattern))
        self.assertEqual(url, 'http://example.com/foo')

    def test_get_deeplink_url_complex_pattern(self):
        resource_name = 'prefix/foo/bar/baz'
        resource_pattern = 'prefix/(?P<one>.+)/(?P<two>.+)/(?P<three>.+)$'
        url_pattern = 'http://example.com/%(one)s/%(two)s-%(three)s'

        url = get_deeplink_url(resource_name, (resource_pattern, url_pattern))
        self.assertEqual(url, 'http://example.com/foo/bar-baz')


class CertificateFromSpConfigTestCase(TestCase):

    def test_certificate_from_no_sp_config(self):
        self.assertEqual(None, certificate_from_sp_config(None))

    def test_certificate_from_sp_config_no_cert(self):
        sp_config = SAMLConfig.objects.create(
            acs_url='http://localhost')
        self.assertEqual(None, certificate_from_sp_config(sp_config))

    def test_certificate_from_sp_config_with_cert(self):
        with open(find_cert_path('certificate.pem')) as cert_file:
            cert_string = cert_file.read()
        sp_config = SAMLConfig.objects.create(
            acs_url='http://localhost', certificate=cert_string)
        # Ensure it's identical to a what we read from the file, and not
        # a djangoized unicode.
        self.assertEqual(cert_string, certificate_from_sp_config(sp_config))
