import copy

from django.urls import reverse
from saml2idp import saml2idp_metadata

from identityprovider.tests.utils import SSOBaseTestCase
from ubuntu_sso_saml.models import SAMLConfig
from ubuntu_sso_saml.tests.helpers import cert_string_from_file, find_cert_path


CUSTOM_TEST_CERTIFICATE = "custom-test-certificate.pem"


class SAMLMetadataViewTestCase(SSOBaseTestCase):

    def setUp(self, *args, **kwargs):
        super(SAMLMetadataViewTestCase, self).setUp(*args, **kwargs)
        # Save settings.
        new_saml2idp_config = copy.deepcopy(saml2idp_metadata.SAML2IDP_CONFIG)
        self.patch(saml2idp_metadata, 'SAML2IDP_CONFIG',
                   new=new_saml2idp_config)
        cert_file = find_cert_path('certificate.pem')
        priv_file = find_cert_path('private-key.pem')
        test_config = {
            'certificate_file': cert_file,
            'private_key_file': priv_file,
            'signing': True,
        }
        saml2idp_metadata.SAML2IDP_CONFIG.update(test_config)

        custom_cert_filename = find_cert_path(CUSTOM_TEST_CERTIFICATE)
        with open(custom_cert_filename) as custom_cert_file:
            custom_cert_string = custom_cert_file.read()
        self.custom_sp_config = SAMLConfig.objects.create(
            acs_url="https://custom.example.com",
            certificate=custom_cert_string)
        self.default_sp_config = SAMLConfig.objects.create(
            acs_url="https://default.example.com")
        self.custom_cert_string = cert_string_from_file(custom_cert_filename)

    def test_metadata_view(self):
        response = self.client.get(reverse('saml-metadata'))
        self.assertEqual(response['content-type'], 'application/xml')
        self.assertContains(response, reverse('logout'))
        self.assertContains(response, reverse('login_begin'))

    def test_sp_specific_metadata_view_nonexistent_sp(self):
        # Can never have a record with sp_id of 0
        # (relying on the fact that the sp_id is actually the database ID for
        # the record).
        response = self.client.get(
            reverse('saml-metadata', args=(0,)))
        self.assertEqual(response.status_code, 404)

    def test_sp_specific_metadata_view_default_cert(self):
        cert_filename = saml2idp_metadata.SAML2IDP_CONFIG['certificate_file']

        response = self.client.get(
            reverse('saml-metadata', args=(self.default_sp_config.pk,)))
        self.assertEqual(response['content-type'], 'application/xml')
        self.assertContains(response, reverse('logout'))
        self.assertContains(response, reverse('login_begin'))

        self.assertContains(
            response, cert_string_from_file(cert_filename))

    def test_sp_specific_metadata_view_custom_cert(self):
        response = self.client.get(
            reverse('saml-metadata', args=(self.custom_sp_config.pk,)))
        self.assertEqual(response['content-type'], 'application/xml')
        self.assertContains(response, reverse('logout'))
        self.assertContains(response, reverse('login_begin'))

        self.assertContains(response, self.custom_cert_string)


class SAMLBeginViewTestCase(SSOBaseTestCase):

    def test_saml_begin_with_relay_state(self):
        url = reverse('login_begin')
        response = self.client.get(
            url, data={'SAMLRequest': 'saml-request-data',
                       'RelayState': 'relay-state'})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response['location'], reverse('login_process'))
        self.assertEqual(
            self.client.session['SAMLRequest'], 'saml-request-data')
        self.assertEqual(self.client.session['RelayState'], 'relay-state')

    def test_saml_begin_without_relay_state(self):
        url = reverse('login_begin')
        response = self.client.get(
            url, data={'SAMLRequest': 'saml-request-data'})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response['location'], reverse('login_process'))
        self.assertEqual(
            self.client.session['SAMLRequest'], 'saml-request-data')
        self.assertIsNone(self.client.session['RelayState'])
