import json
import re

from identityprovider.models import OpenIDRPConfig
from identityprovider.utils import get_object_or_none
from ubuntu_sso_saml.models import SAMLConfig


def get_rpconfig_from_request(request):
    from saml2idp.registry import find_processor
    from saml2idp.exceptions import CannotHandleAssertion

    rpconfig = None

    try:
        proc = find_processor(request)
    except CannotHandleAssertion:
        return

    trust_root = proc._request_params['ACS_URL']
    if not trust_root:
        spconfig = get_config_from_processor(proc)
        if spconfig is not None:
            trust_root = spconfig.acs_url
    rpconfig = OpenIDRPConfig.objects.for_url(trust_root)

    return rpconfig


def get_config_from_processor(proc):
    acs_url = proc._request_params.get('ACS_URL')
    issuer = proc._request_params.get('REQUEST_ISSUER')
    return get_config_for_acs(acs_url, issuer=issuer)


def get_config_for_acs(acs_url, issuer=None):
    """
    Return SP configuration instance that handles acs_url.
    """
    config = get_object_or_none(SAMLConfig, acs_url=acs_url, enabled=True)
    if config is None:
        config = get_object_or_none(SAMLConfig, issuer=issuer, enabled=True)
    return config


def get_config_and_link_for_resource(resource_name):
    """
    Return the SP configuration that handles a deep-link resource_name.

    Also return the matching link patterns.
    """
    for sp_config in SAMLConfig.objects.filter(enabled=True).exclude(links=''):
        links = json.loads(sp_config.links)
        for resource_pattern, url_pattern in links.items():
            exp = re.compile(resource_pattern)
            if exp.match(resource_name):
                return sp_config, (resource_pattern, url_pattern)
    return None, None


def get_deeplink_url(resource_name, (resource_pattern, url_pattern)):
    exp = re.compile(resource_pattern)
    match = exp.match(resource_name)
    if match is not None:
        url = url_pattern % match.groupdict()
        return url


def certificate_from_sp_config(sp_config):
    """
    Return the SP's certificate (bytes) or None if there's no certificate.

    Also returns None if the passed sp_config is None.

    """
    if sp_config and sp_config.certificate:
        # Django stores string/unicode but M2Crypto needs bytes for
        # certificate values.
        certificate = sp_config.certificate.encode('ascii')
    else:
        certificate = None
    return certificate
