# Copyright 2016 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import json
import logging
import re
import ssl
import urllib
import urllib2

from django.conf import settings
from oauthlib import oauth1


from webservices.utils import http_request_with_timeout


logger = logging.getLogger(__name__)

NAME_TAKEN_RE = re.compile(
    r'^name: .* is already in use by another person or team.$')
BLACKLISTED_RE = re.compile(
    r'^name: .* has been blocked by the Launchpad administrators.')
NOT_PLACEHOLDER_RE = re.compile(
    r'^An account for that OpenID identifier already exists.')


class LaunchpadAPIError(Exception):

    def __init__(self, code, message):
        super(LaunchpadAPIError, self).__init__(message)
        self.code = code


class NameAlreadyTaken(LaunchpadAPIError):
    """A username can't be set because it's in use by another account."""
    pass


class NotPlaceholderAccount(LaunchpadAPIError):
    """That account's username can't be changed except by LP."""
    pass


class NoSuchAccount(LaunchpadAPIError):
    """The specified openid_identifier doesn't exist."""

    MATCH = re.compile(
        r"^No account found for openid identifier '.*?'$")


class SSHKeyAdditionTypeError(LaunchpadAPIError):
    """Something was wrong with the specified SSH key text."""

    MATCH = re.compile(
        r"^Invalid SSH key type:.*$")


class SSHKeyAdditionDataError(LaunchpadAPIError):

    MATCH = re.compile(
        r"^Invalid SSH key data:.*$")


class SSHKeyCompromisedError(LaunchpadAPIError):
    """The SSH key is known to be easily compromised."""

    MATCH = re.compile(
        r"^This key cannot be added as it is known to be compromised\.$"
    )


class UnknownLaunchpadError(Exception):
    """Something went wrong while talking to Launchpad.

    This exception must not derive from LaunchpadAPIError, since we don't want
    it getting caught and shown to the user. It's a separate exception type
    because we want to make sure that the response body shows up in the
    exception's __str__ method.
    """

    def __init__(self, code, message):
        self.code = code
        self.message = message
        self.args = (code, message)

    def __str__(self):
        return (
            "<UnknownLaunchpadError code: %d message: '%s'>" % (
                self.code,
                self.message))


def _lp_api_url():
    if not settings.LP_API_URL:
        raise RuntimeError(
            "Error: LP_API_URL not defined.\n"
            "Ensure LP_API_URL is defined in ../local_config/settings.py\n"
            "See 'Setup Launchpad Integration' in README for more details.")
    else:
        return settings.LP_API_URL


def _lp_api_opener():
    return urllib2.build_opener()


def _username_placeholder_exception(code, message):
    if 400 <= code <= 499:
        if NAME_TAKEN_RE.match(message) or BLACKLISTED_RE.match(message):
            return NameAlreadyTaken(code, message)
        elif NOT_PLACEHOLDER_RE.match(message):
            return NotPlaceholderAccount(code, message)


def _lp_oauth_headers():
    oauth_client = oauth1.Client(
        settings.LP_API_CONSUMER_KEY,
        resource_owner_key=settings.LP_API_TOKEN,
        resource_owner_secret=settings.LP_API_TOKEN_SECRET,
        signature_method=oauth1.SIGNATURE_PLAINTEXT)
    return oauth_client.sign(settings.LP_API_URL)[1]


def set_lp_username(openid_identifier, username, dry_run=False):
    headers = _lp_oauth_headers()
    try:
        response = http_request_with_timeout(
            '%s/people' % _lp_api_url(),
            data={
                'ws.op': 'setUsernameFromSSO',
                'openid_identifier': openid_identifier,
                'name': username, 'dry_run': json.dumps(dry_run)},
            headers=headers, opener=_lp_api_opener())
    except urllib2.HTTPError as e:
        new_e = _username_placeholder_exception(e.code, e.read())
        if new_e is not None:
            raise new_e
        else:
            raise
    if response.code != 200:
        raise LaunchpadAPIError(response.code, response.read())


def get_lp_username(openid_identifier):
    headers = _lp_oauth_headers()
    response = http_request_with_timeout(
        '%s/people?%s' % (
            _lp_api_url(),
            urllib.urlencode({
                'ws.op': 'getUsernameForSSO',
                'openid_identifier': openid_identifier})),
        headers=headers, opener=_lp_api_opener())
    if response.code != 200:
        raise LaunchpadAPIError(response.code, response.read())
    return json.loads(response.read())


def _sshkey_placeholder_exception(code, message):
    possible_exceptions = (
        NoSuchAccount,
        SSHKeyAdditionTypeError,
        SSHKeyAdditionDataError,
        SSHKeyCompromisedError,
    )
    if 400 <= code <= 499:
        for error_class in possible_exceptions:
            if error_class.MATCH.match(message):
                return error_class(code, message)
    return UnknownLaunchpadError(code, message)


def get_lp_ssh_keys(openid_identifier):
    headers = _lp_oauth_headers()
    response = http_request_with_retry(
        '%s/people?%s' % (
            _lp_api_url(),
            urllib.urlencode({
                'ws.op': 'getSSHKeysForSSO',
                'openid_identifier': openid_identifier})),
        headers=headers, opener=_lp_api_opener())
    if response.code != 200:
        raise LaunchpadAPIError(response.code, response.read())
    return json.loads(response.read())


def add_lp_ssh_key(openid_identifier, key_text, dry_run):
    headers = _lp_oauth_headers()
    response = http_request_with_retry(
        '%s/people' % _lp_api_url(),
        data={
            'ws.op': 'addSSHKeyFromSSO',
            'openid_identifier': openid_identifier,
            'key_text': key_text, 'dry_run': json.dumps(dry_run)},
        headers=headers, opener=_lp_api_opener())
    if response.code != 200:
        raise LaunchpadAPIError(response.code, response.read())


def delete_lp_ssh_key(openid_identifier, key_text, dry_run):
    headers = _lp_oauth_headers()
    response = http_request_with_retry(
        '%s/people' % _lp_api_url(),
        data={
            'ws.op': 'deleteSSHKeyFromSSO',
            'openid_identifier': openid_identifier,
            'key_text': key_text, 'dry_run': json.dumps(dry_run)},
        headers=headers, opener=_lp_api_opener())
    if response.code != 200:
        raise LaunchpadAPIError(response.code, response.read())


def http_request_with_retry(*args, **kwargs):
    retry = False
    try:
        return http_request_with_timeout(*args, **kwargs)
    except ssl.SSLError as e:
        # Always retry SSLErrors - this is almost certainly a network blip...
        logger.warning(
            "Got SSLError (%s) while processing request (%r, %r), retrying...",
            e, args, _filter_request_kwargs(kwargs))
        retry = True
    except urllib2.HTTPError as e:
        # Retry only if the exception isn't a known failure mode:
        new_e = _sshkey_placeholder_exception(e.code, e.read())
        if isinstance(new_e, UnknownLaunchpadError):
            logger.warning(
                "Got HTTPError (%s, %s) while processing request (%r, %r), "
                "retrying...", e.code, e.read(), args,
                _filter_request_kwargs(kwargs))
            retry = True
        else:
            raise new_e
    if retry:
        # Run the request again, always raising the converted exception
        try:
            return http_request_with_timeout(*args, **kwargs)
        except urllib2.HTTPError as e:
            raise _sshkey_placeholder_exception(e.code, e.read())


def _filter_request_kwargs(request_kwargs):
    kwargs = request_kwargs.copy()
    if 'headers' in kwargs:
        del kwargs['headers']
    return kwargs
