# Copyright 2016-2018 Canonical Ltd.  This software is licensed under
# the GNU Affero General Public License version 3 (see the file LICENSE).

import re
import unicodedata

from django import forms
from django.utils.html import escape, format_html
from django.utils.translation import ugettext_lazy as _

from webservices.launchpad import (
    LaunchpadAPIError,
    add_lp_ssh_key,
    delete_lp_ssh_key,
    get_lp_ssh_keys,
)


class SSHKeyField(forms.CharField):

    def __init__(self, *args, **kwargs):
        placeholder = kwargs.pop('placeholder')
        attrs = {
            'placeholder': placeholder,
            'cols': '80',
            'rows': '4',
        }
        if kwargs.pop('disabled', False):
            attrs['disabled'] = 'disabled'

        kwargs['widget'] = forms.Textarea(attrs=attrs)
        super(SSHKeyField, self).__init__(*args, **kwargs)


class ClaimSSHKeyForm(forms.Form):

    def __init__(self, account, *args, **kwargs):
        super(ClaimSSHKeyForm, self).__init__(*args, **kwargs)
        self.account = account
        if not self.account.is_verified:
            disabled = True
            placeholder = _(
                "You must verify your account and set a username first.")
        elif self.account.person_name is None:
            disabled = True
            placeholder = _("You must set a username before importing a key.")
        else:
            disabled = False
            placeholder = _("Begins with 'ssh-rsa', or 'ssh-dss'.")
        self.fields['ssh_key'] = SSHKeyField(
            required=True,
            disabled=disabled,
            placeholder=placeholder,
            label=_("Public SSH Key")
        )

    def clean_ssh_key(self):
        keytext = self.cleaned_data['ssh_key']
        # Clean unicode control characters and other non-printable characters.
        keytext = u''.join(
            c if not unicodedata.category(c).startswith('C') else ' '
            for c in keytext)
        keytext = re.sub(r'\s+', ' ', keytext)
        keytext = keytext.strip()
        keytext = keytext.encode('utf8')
        if 'PRIVATE KEY' in keytext:
            raise forms.ValidationError(
                "This is a private key. "
                "You must upload your public key instead.")
        try:
            # Ask launchpad for a dry_run so nothing is actually changed:
            add_lp_ssh_key(self.account.openid_identifier, keytext,
                           dry_run=True)
        except LaunchpadAPIError as e:
            raise forms.ValidationError(str(e))
        return keytext

    def add_key_to_launchpad(self):
        keytext = self.cleaned_data['ssh_key']
        key_comment = keytext.split(' ', 2)[-1]
        add_lp_ssh_key(self.account.openid_identifier, keytext, dry_run=False)
        return key_comment


def format_ssh_key(ssh_key):
    """Decide how to display 'ssh_key'.

    SSH keys are freeform text, so users absolutely *will* try and add keys
    that contain javascript, so we need to make sure they're escaped when
    being displayed to the user.
    """
    try:
        key_type, key_text, comment = ssh_key.split(' ', 2)
    except ValueError:
        # This should never happen, as launchpad is supposed to guarantee a
        # valid ssh key format. However, we need to exercise this code path
        # in tests to ensure that users malicously POST'ing bad key formats
        # will get a ValidationError.
        # Just return the escaped data - this will never be run in production:
        return format_html(
            '<label>Bad Key Data: <span>{key}</span></label>',
            key=escape(ssh_key))
    else:
        return format_html(
            u'<div class="key-display">'
            u'  <h4>{comment}</h4> '
            u'  <ul>'
            u'    <li>'
            u'      <dl>'
            u'        <dt>{key_type_label}</dt>'
            u'        <dd>{key_type}</dt>'
            u'      </dl>'
            u'    </li>'
            u'    <li>'
            u'      <dl>'
            u'        <dt>{key_text_label}</dt>'
            u'        <dd class="key-text">'
            u'          <a href="#">{key_text}</a>'
            u'        </dt>'
            u'      </dl>'
            u'    </li>'
            u'  </ul>'
            u'</div>',
            comment=comment,
            key_type_label=_("Type: "),
            key_type=key_type,
            key_text_label=_("Text: "),
            key_text=key_text,
        )


class DeleteSSHKeyForm(forms.Form):

    ssh_keys = forms.MultipleChoiceField(
        required=True,
        widget=forms.CheckboxSelectMultiple(attrs={'class': 'ssh_input'}),
        label='',
    )

    def __init__(self, account, *args, **kwargs):
        super(DeleteSSHKeyForm, self).__init__(*args, **kwargs)
        self.account = account
        keys = get_lp_ssh_keys(self.account.openid_identifier) or []
        choices = [(key, format_ssh_key(key)) for key in keys]
        self.fields['ssh_keys'].choices = choices

    def delete_keys_from_launchpad(self):
        """Delete selected keys from launchpad.

        :returns: A list of key comments for keys that were deleted.
        """
        deleted_keys = []
        for key in self.cleaned_data['ssh_keys']:
            delete_lp_ssh_key(
                self.account.openid_identifier,
                key.encode('utf8'),
                dry_run=False)
            deleted_keys.append(key.split(' ', 2)[-1])
        return deleted_keys
