# -*- coding: utf-8 -*-
# Copyright 2016-2018 Canonical Ltd.  This software is licensed under
# the GNU Affero General Public License version 3 (see the file LICENSE).

from __future__ import unicode_literals

from identityprovider.tests.utils import SSOBaseTestCase
from webservices.launchpad import (
    NoSuchAccount,
    SSHKeyAdditionDataError,
    SSHKeyAdditionTypeError,
    SSHKeyCompromisedError,
)
from webui.forms import (
    ClaimSSHKeyForm,
    DeleteSSHKeyForm,
    format_ssh_key,
)


class ClaimSSHKeyFormTestCase(SSOBaseTestCase):

    def assert_field_is_disabled(self, field):
        self.assertIn('disabled', field.widget.attrs)

    def assert_field_placeholder_equals(self, field, expected):
        self.assertEqual(expected, str(field.widget.attrs['placeholder']))

    def make_full_account(self):
        """A convenience function to make a 'full' account.

        A 'full' account in this context means that the email has been verified
        and a launchpad username has been set.
        """
        account = self.factory.make_account(email_validated=True)
        self.factory.make_person(account=account)
        return account

    def patch_lp_add_key(self, raise_exception=None):
        """Patch the 'add_lp_ssh_key', creating realistic side effects.

        :param raise_exception: If specified, must be one of NoSuchAccount,
            SSHKeyAdditionTypeError, SSHKeyAdditionDataError or
            SSHKeyCompromisedError. Exceptions will be raised with the same
            message as if they'd been returned from launchpad.
        """
        exception_messages = {
            NoSuchAccount: "No account found for openid identifier '{openid}'",
            SSHKeyAdditionTypeError: "Invalid SSH key type: '{sshkeytype}'",
            SSHKeyAdditionDataError: "Invalid SSH key data: '{sshkey}'",
            SSHKeyCompromisedError: "This key cannot be added as it is "
                                    "known to be compromised.",
        }

        def _side_effect(openid_id, key_text, dry_run):
            key_type = key_text.split(' ')[0]
            message = exception_messages[raise_exception].format(
                openid=openid_id,
                sshkeytype=key_type,
                sshkey=key_text,
            )
            raise raise_exception(400, message)

        mock = self.patch('webui.forms.add_lp_ssh_key')
        if raise_exception is not None:
            mock.side_effect = _side_effect
        return mock

    def assert_form_field_has_error(self, form, field, expected_error):
        self.assertIn(field, form.errors)
        self.assertEqual([expected_error], form.errors[field])

    def test_is_disabled_with_unverified_account(self):
        account = self.factory.make_account(email_validated=False)
        form = ClaimSSHKeyForm(account)

        self.assert_field_is_disabled(form.fields['ssh_key'])
        self.assert_field_placeholder_equals(
            form.fields['ssh_key'],
            "You must verify your account and set a username first.",
        )

    def test_is_disabled_with_no_username_set(self):
        account = self.factory.make_account(email_validated=True)
        form = ClaimSSHKeyForm(account)

        self.assert_field_is_disabled(form.fields['ssh_key'])
        self.assert_field_placeholder_equals(
            form.fields['ssh_key'],
            "You must set a username before importing a key.",
        )

    def test_form_not_disabled(self):
        account = self.make_full_account()
        form = ClaimSSHKeyForm(account)

        self.assertNotIn('disabled', form.fields['ssh_key'].widget.attrs)
        self.assert_field_placeholder_equals(
            form.fields['ssh_key'],
            "Begins with 'ssh-rsa', or 'ssh-dss'.",
        )

    def test_form_errors_with_missing_Data(self):
        account = self.make_full_account()
        form = ClaimSSHKeyForm(account, {'ssh_key': ''})

        self.assertFalse(form.is_valid())
        self.assert_form_field_has_error(
            form, 'ssh_key', 'This field is required.')

    def test_form_uses_dry_run_when_cleaning_data(self):
        account = self.make_full_account()
        form = ClaimSSHKeyForm(
            account, {'ssh_key': 'ssh-rsa keytext comment'})

        mock_lp_add = self.patch_lp_add_key()

        self.assertTrue(form.is_valid())
        mock_lp_add.assert_called_once_with(
            account.openid_identifier, 'ssh-rsa keytext comment', dry_run=True)

    def test_form_converts_lp_NoSuchAccount_to_form_error(self):
        account = self.make_full_account()
        form = ClaimSSHKeyForm(
            account, {'ssh_key': 'ssh-rsa keytext comment'})

        self.patch_lp_add_key(NoSuchAccount)
        self.assertFalse(form.is_valid())
        self.assert_form_field_has_error(
            form,
            'ssh_key',
            "No account found for openid identifier '%s'"
            % account.openid_identifier,
        )

    def test_form_converts_lp_SSHKeyAdditionTypeError_to_form_error(self):
        account = self.make_full_account()
        form = ClaimSSHKeyForm(
            account, {'ssh_key': 'ssh-flub keytext comment'})

        self.patch_lp_add_key(SSHKeyAdditionTypeError)
        self.assertFalse(form.is_valid())
        self.assert_form_field_has_error(
            form,
            'ssh_key',
            "Invalid SSH key type: 'ssh-flub'",
        )

    def test_form_converts_lp_SSHKeyAdditionDataError_to_form_error(self):
        account = self.make_full_account()
        form = ClaimSSHKeyForm(
            account, {'ssh_key': 'badkey'})

        self.patch_lp_add_key(SSHKeyAdditionDataError)
        self.assertFalse(form.is_valid())
        self.assert_form_field_has_error(
            form,
            'ssh_key',
            "Invalid SSH key data: 'badkey'",
        )

    def test_form_converts_lp_SSHKeyCompromisedError_to_form_error(self):
        account = self.make_full_account()
        form = ClaimSSHKeyForm(
            account, {'ssh_key': 'ssh-dss foo bar'})

        self.patch_lp_add_key(SSHKeyCompromisedError)
        self.assertFalse(form.is_valid())
        self.assert_form_field_has_error(
            form,
            'ssh_key',
            "This key cannot be added as it is known to be compromised.",
        )

    def test_form_adds_ssh_key(self):
        account = self.make_full_account()
        form = ClaimSSHKeyForm(
            account, {'ssh_key': 'ssh-rsa foo bar'})

        mock_lp_add = self.patch_lp_add_key()
        self.assertTrue(form.is_valid())
        mock_lp_add.reset_mock()
        form.add_key_to_launchpad()

        mock_lp_add.assert_called_once_with(
            account.openid_identifier,
            'ssh-rsa foo bar',
            dry_run=False)

    def test_form_normalises_whitespace(self):
        account = self.make_full_account()
        form = ClaimSSHKeyForm(
            account, {'ssh_key': 'ssh-rsa\t\n\tfoo\r\nbar\r\n\0'})

        mock_lp_add = self.patch_lp_add_key()
        self.assertTrue(form.is_valid())
        mock_lp_add.reset_mock()
        form.add_key_to_launchpad()

        mock_lp_add.assert_called_once_with(
            account.openid_identifier,
            'ssh-rsa foo bar',
            dry_run=False)

    def test_rejects_private_key(self):
        private_key = (
            '-----BEGIN RSA PRIVATE KEY-----\n'
            'keytext goes here\n'
            '-----END RSA PRIVATE KEY-----'
        )
        account = self.make_full_account()
        form = ClaimSSHKeyForm(
            account, {'ssh_key': private_key})
        self.assertFalse(form.is_valid())
        self.assert_form_field_has_error(
            form,
            'ssh_key',
            "This is a private key. You must upload your public key instead.",
        )

    def text_accepts_unicode(self):
        keytext = 'ssh-dss keytext ¿ʇuǝɯɯoɔ'
        account = self.make_full_account()
        form = ClaimSSHKeyForm(
            account, {'ssh_key': keytext})
        mock_lp_add = self.patch_lp_add_key()

        self.assertTrue(form.is_valid())
        mock_lp_add.assert_called_once_with(
            account.openid_identifier,
            keytext.encode('utf8'),
            dry_run=True)

    def test_adding_key_returns_key_comment(self):
        account = self.make_full_account()
        form = ClaimSSHKeyForm(
            account, {'ssh_key': 'ssh-rsa foo bar'})

        self.patch_lp_add_key()
        self.assertTrue(form.is_valid())
        self.assertEqual('bar', form.add_key_to_launchpad())


class FormatSSHKeyTestCase(SSOBaseTestCase):

    def test_escapes_xss(self):
        bad_content = '<script>window.alert("boo!");</script>'
        bad_key = ' '.join((bad_content,) * 3)
        formatted_key = format_ssh_key(bad_key)
        self.assertNotIn(bad_content, formatted_key)
        self.assertIn(
            '&lt;script&gt;window.alert(&quot;boo!&quot;);&lt;/script&gt;',
            formatted_key
        )

    def test_unicode(self):
        keytext = 'ssh-dss keytext ¿ʇuǝɯɯoɔ'
        # previously this would raise UnicodeDecodeError. Now it works as
        # expected.
        formatted_key = format_ssh_key(keytext)
        self.assertIn('¿ʇuǝɯɯoɔ', formatted_key)


class DeleteSSHKeyFormTestCase(SSOBaseTestCase):

    SAMPLE_RSA_KEY = 'ssh-rsa keytext_goes_here comment goes here'
    SAMPLE_DSA_KEY = 'ssh-dss more_keytext_here this is a comment'

    def setUp(self):
        super(DeleteSSHKeyFormTestCase, self).setUp()
        self.account = self.factory.make_account()

    def patch_lp_delete_key(self, raise_exception=None):
        """Patch the 'delete_lp_ssh_key', creating realistic side effects.

        :param raise_exception: If specified, must be one of NoSuchAccount,
            SSHKeyAdditionTypeError or SSHKeyAdditionDataError Exceptions will
            be raised with the same message as if they'd been returned from
            launchpad.
        """
        exception_messages = {
            NoSuchAccount: "No account found for openid identifier '{openid}'",
            SSHKeyAdditionTypeError: "Invalid SSH key type: '{sshkeytype}'",
            SSHKeyAdditionDataError: "Invalid SSH key data: '{sshkey}'",
        }

        def _side_effect(openid_id, key_text, dry_run):
            key_type = key_text.split(' ')[0]
            message = exception_messages[raise_exception].format(
                openid=openid_id,
                sshkeytype=key_type,
                sshkey=key_text,
            )
            raise raise_exception(400, message)

        mock = self.patch('webui.forms.delete_lp_ssh_key')
        if raise_exception is not None:
            mock.side_effect = _side_effect
        return mock

    def patch_get_keys(self, returned_keys=[]):
        mock = self.patch('webui.forms.get_lp_ssh_keys')
        mock.return_value = returned_keys
        return mock

    def assert_form_field_has_error(self, form, field, expected_error):
        self.assertIn(field, form.errors)
        self.assertEqual([expected_error], form.errors[field])

    def test_form_passes_account_id_to_get_lp_keys(self):
        mock_get_keys = self.patch_get_keys()
        DeleteSSHKeyForm(self.account)

        mock_get_keys.assert_called_once_with(self.account.openid_identifier)

    def test_form_with_no_keys(self):
        self.patch_get_keys()
        form = DeleteSSHKeyForm(self.account)

        self.assertEqual(0, len(form.fields['ssh_keys'].choices))

    def test_form_multiple_keys(self):
        self.patch_get_keys([self.SAMPLE_RSA_KEY, self.SAMPLE_DSA_KEY])
        form = DeleteSSHKeyForm(self.account)

        self.assertEqual(2, len(form.fields['ssh_keys'].choices))
        rsa_choice, dsa_choice = form.fields['ssh_keys'].choices
        self.assertEqual(
            (self.SAMPLE_RSA_KEY, format_ssh_key(self.SAMPLE_RSA_KEY)),
            rsa_choice,
        )
        self.assertEqual(
            (self.SAMPLE_DSA_KEY, format_ssh_key(self.SAMPLE_DSA_KEY)),
            dsa_choice,
        )

    def test_form_deletes_ssh_key(self):
        self.patch_get_keys([self.SAMPLE_RSA_KEY])
        form = DeleteSSHKeyForm(
            self.account, {'ssh_keys': [self.SAMPLE_RSA_KEY]})

        mock_lp_delete = self.patch_lp_delete_key()
        self.assertTrue(form.is_valid())
        mock_lp_delete.reset_mock()
        form.delete_keys_from_launchpad()

        mock_lp_delete.assert_called_once_with(
            self.account.openid_identifier,
            self.SAMPLE_RSA_KEY,
            dry_run=False)

    def test_form_returns_key_comments_from_delete(self):
        self.patch_get_keys([self.SAMPLE_RSA_KEY])
        form = DeleteSSHKeyForm(
            self.account, {'ssh_keys': [self.SAMPLE_RSA_KEY]})

        self.patch_lp_delete_key()
        self.assertTrue(form.is_valid())
        self.assertEqual(
            [self.SAMPLE_RSA_KEY.split(' ', 2)[-1]],
            form.delete_keys_from_launchpad()
        )

    def test_accepts_unicode(self):
        key_with_unicode = "ssh-rsa keytest 契金喇奈懶"
        self.patch_get_keys([key_with_unicode])
        form = DeleteSSHKeyForm(
            self.account, {'ssh_keys': [key_with_unicode]})

        mock_lp_delete = self.patch_lp_delete_key()
        self.assertTrue(form.is_valid())
        form.delete_keys_from_launchpad()
        mock_lp_delete.assert_called_once_with(
            self.account.openid_identifier,
            key_with_unicode.encode('utf8'),
            dry_run=False)
