from __future__ import unicode_literals

from django.template.loader import render_to_string
from django.urls import reverse
from mock import patch
from pyquery import PyQuery

from identityprovider.models.openidmodels import OpenIDRPConfig
from identityprovider.readonly import ReadOnlyManager
from identityprovider.tests.utils import (
    AuthenticatedTestCase,
    SSOBaseTestCase,
)


class LoginTemplateTestCase(SSOBaseTestCase):

    def setUp(self):
        super(LoginTemplateTestCase, self).setUp()
        self.preseed_cookie_check()

    @patch('webui.views.ui.get_rpconfig_from_request')
    def test_rpconfig_with_logo_url(self, mock_get_rpconfig):
        rpconfig = OpenIDRPConfig(
            trust_root='http://localhost/',
            logo='http://localhost/img.png')
        mock_get_rpconfig.return_value = rpconfig

        response = self.client.get('/+login')

        self.assertTemplateUsed(response, 'registration/login.html')

    @patch('webui.views.ui.get_rpconfig_from_request')
    def test_rpconfig_without_logo_url(self, mock_get_rpconfig):
        rpconfig = OpenIDRPConfig(
            trust_root='http://localhost/',
            logo='')
        mock_get_rpconfig.return_value = rpconfig

        response = self.client.get('/+login')

        self.assertTemplateUsed(response, 'registration/login.html')
        self.assertNotContains(response, 'id="rpconfig_logo"')

    def render_login_with_rpconfig(self, rpconfig):
        return render_to_string(
            'registration/login.html', dict(rpconfig=rpconfig))

    def get_title_style_and_text(self, dom):
        titles = dom.find('p[class~=login-title]')
        self.assertEqual(1, titles.length)
        text = " ".join(titles[0].text_content().split())
        style = dom.find('style[data-qa-id="test_login_rp"]')
        if len(style) == 1:
            style = " ".join(style.text().split())
        else:
            style = None

        return style, text

    def test_login_rp_details(self):
        rpconfig = OpenIDRPConfig(
            trust_root='http://localhost/',
            displayname='Landscape',
            logo='http://localhost/img.png')

        html = self.render_login_with_rpconfig(rpconfig)

        style, text = self.get_title_style_and_text(PyQuery(html))
        self.assertIn("url('http://localhost/img.png')", style)
        self.assertIn(u"Landscape log in with Ubuntu One", text)

    def test_login_rp_no_logo(self):
        """The rp displayname is still included."""
        rpconfig = OpenIDRPConfig(
            trust_root='http://localhost/',
            displayname='Landscape')

        html = self.render_login_with_rpconfig(rpconfig)

        style, text = self.get_title_style_and_text(PyQuery(html))
        self.assertIsNone(style)
        self.assertIn(u"Landscape log in with Ubuntu One", text)

    def test_login_rp_no_displayname(self):
        rpconfig = OpenIDRPConfig(
            trust_root='http://localhost/',
            displayname='Landscape',
            logo='http://localhost/img.png')

        html = self.render_login_with_rpconfig(rpconfig)

        style, text = self.get_title_style_and_text(PyQuery(html))
        self.assertIn("url('http://localhost/img.png')", style)
        self.assertIn(u"log in with Ubuntu One", text)

    def test_login_has_create_account_form(self):
        response = self.client.get('/+login')
        self.assertContains(response, "data-qa-id=\"create_account_form\"")

    def test_login_without_create_account_form(self):
        rm = ReadOnlyManager()
        rm.set_readonly()
        self.addCleanup(rm.clear_readonly)

        response = self.client.get('/+login')
        self.assertNotContains(response, "data-qa-id=\"create_account_form\"")


class NewAccountTemplateTestCase(SSOBaseTestCase):

    def test_with_logo_url(self):
        rpconfig = OpenIDRPConfig(
            trust_root='http://localhost/',
            logo='http://localhost/img.png'
        )
        html = render_to_string(
            'registration/new_account.html',
            {'rpconfig': rpconfig}
        )
        self.assertIn('id="rpconfig_logo"', html)
        self.assertIn('src="http://localhost/img.png"', html)

    def test_without_logo_url(self):
        rpconfig = OpenIDRPConfig(
            trust_root='http://localhost/',
            logo=''
        )
        html = render_to_string(
            'registration/new_account.html',
            {'rpconfig': rpconfig}
        )
        self.assertNotIn('id="rpconfig_logo"', html)

    def test_action_without_token(self):
        html = render_to_string('registration/new_account.html', {})
        dom = PyQuery(html)
        form = dom.find('form[name=newaccountform]')
        self.assertEqual(form.attr['action'], reverse('new_account'))

    def test_action_with_token(self):
        ctx = {'token': 'a' * 16}
        html = render_to_string('registration/new_account.html', ctx.copy())
        dom = PyQuery(html)
        form = dom.find('form[name=newaccountform]')
        self.assertEqual(
            form.attr['action'],
            reverse('new_account', kwargs=ctx)
        )


class UbuntuOneBaseTemplateTestCase(SSOBaseTestCase):
    def test_base_template_includes_analytics(self):
        # Google tag manager code is included if the id is set.
        with self.settings(GOOGLE_TAG_MANAGER_ID='foobar'):
            response = self.client.get('/')

        self.assertTemplateUsed(response, 'base.html')
        self.assertContains(response, 'gtm.start')

    def test_base_template_not_includes_analytics(self):
        # If the google tag manager id is not set the analytics
        # code is not included.
        with self.settings(GOOGLE_TAG_MANAGER_ID=None):
            response = self.client.get('/')

        self.assertTemplateUsed(response, 'base.html')
        self.assertNotContains(response, 'gtm.start')

    def test_login_header_without_token(self):
        html = render_to_string('base.html', {})
        dom = PyQuery(html)
        link = dom.find('a#login-link')
        self.assertEqual(link.attr['href'], reverse('login'))

    def test_login_header_with_token(self):
        ctx = {'token': 'a' * 16}
        html = render_to_string('base.html', ctx.copy())
        dom = PyQuery(html)
        link = dom.find('a#login-link')
        self.assertEqual(link.attr['href'], reverse('login', kwargs=ctx))


class UbuntuOneBaseTemplateAuthenticatedTestCase(AuthenticatedTestCase):
    def test_user_header_without_token(self):
        r = self.client.get(reverse('index'))
        dom = PyQuery(r.content)
        link = dom.find('a#account-link')
        self.assertEqual(link.attr['href'], reverse('account-index'))
        link = dom.find('a#logout-link')
        self.assertEqual(link.attr['href'], reverse('logout'))

    def test_user_header_with_token(self):
        ctx = {'token': 'a' * 16}
        r = self.client.get(reverse('index', kwargs=ctx))
        dom = PyQuery(r.content)
        link = dom.find('a#account-link')
        self.assertEqual(
            link.attr['href'], reverse('account-index', kwargs=ctx))
        link = dom.find('a#logout-link')
        self.assertEqual(link.attr['href'], reverse('logout', kwargs=ctx))
