# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from __future__ import unicode_literals

from django.urls import reverse
from gargoyle.testutils import switches
from mock import call

from identityprovider.tests import DEFAULT_USER_PASSWORD
from identityprovider.tests.utils import SSOBaseTestCase
from webui.views.account import (
    _make_added_key_message,
    _make_deleted_keys_message,
    ssh_keys,
)


class SSHViewFeatureFlagTestCase(SSOBaseTestCase):

    def create_user_and_login(self, name='person-2'):
        self.account = self.factory.make_account(
            password=DEFAULT_USER_PASSWORD,
            email='example_user@example.com')
        assert self.client.login(
            username=self.account.preferredemail.email,
            password=DEFAULT_USER_PASSWORD)

    def mock_get_ssh_keys(self):
        mock = self.patch('webui.forms.get_lp_ssh_keys')
        mock.return_value = []

    @switches(SSH_KEY_INTEGRATION=False)
    def test_ssh_keys_menu_item_not_rendered_without_feature_flag(self):
        self.create_user_and_login()
        resp = self.client.get(reverse('account-index'))

        self.assertNotContains(resp, '<span>SSH keys</span>', html=True)

    @switches(SSH_KEY_INTEGRATION=True)
    def test_ssh_keys_menu_item_is_rendered_with_feature_flag(self):
        self.create_user_and_login()
        resp = self.client.get(reverse('account-index'))

        self.assertContains(resp, 'SSH keys')

    @switches(SSH_KEY_INTEGRATION=False)
    def test_ssh_keys_view_returns_404_without_feature_flag(self):
        self.create_user_and_login()
        resp = self.client.get(reverse('ssh_keys'))

        self.assertEqual(404, resp.status_code)

    @switches(SSH_KEY_INTEGRATION=True)
    def test_ssh_keys_view_returns_200_with_feature_flag(self):
        self.create_user_and_login()
        self.mock_get_ssh_keys()
        resp = self.client.get(reverse('ssh_keys'))

        self.assertEqual(200, resp.status_code)
        self.assertTemplateUsed(resp, 'account/ssh_keys.html')

    def test_ssh_keys_view_returns_200_with_feature_flag_conditional(self):
        self.create_user_and_login()
        self.mock_get_ssh_keys()
        condition_set = ('identityprovider.gargoyle_conditions.'
                         'AccountConditionSet(identityprovider.account)')
        self.conditionally_enable_flag(
            'SSH_KEY_INTEGRATION',
            'email',
            self.account.preferredemail.email,
            condition_set
        )
        resp = self.client.get(reverse('ssh_keys'))

        self.assertEqual(200, resp.status_code)
        self.assertTemplateUsed(resp, 'account/ssh_keys.html')

    def test_ssh_keys_view_returns_404_without_feature_flag_conditional(self):
        self.create_user_and_login()
        # Set a conditional flag for a second account to verify that it really
        # is conditional:
        special_account = self.factory.make_account()
        condition_set = ('identityprovider.gargoyle_conditions.'
                         'AccountConditionSet(identityprovider.account)')
        self.conditionally_enable_flag(
            'SSH_KEY_INTEGRATION',
            'email',
            special_account.preferredemail.email,
            condition_set
        )
        resp = self.client.get(reverse('ssh_keys'))

        self.assertEqual(404, resp.status_code)

    @switches(SSH_KEY_INTEGRATION=True)
    def test_ssh_keys_view_requires_authentication(self):
        self.preseed_cookie_check()
        resp = self.client.get(reverse('ssh_keys'))

        self.assertRedirects(
            resp,
            reverse('login') + '?next=' + reverse('ssh_keys'),
        )

    @switches(SSH_KEY_INTEGRATION=True)
    def test_ssh_keys_view_disabled_in_readonly_mode(self):
        account = self.factory.make_account()
        with self.settings(READ_ONLY_MODE=True):
            request = self.factory.make_request(user=account)
            resp = ssh_keys(request)

        self.assertContains(
            resp,
            "<h1 class='main'>We're currently in read-only mode.</h1>",
            status_code=403,
            html=True)


class SSHKeyTestCaseBase(SSOBaseTestCase):

    def assert_response_includes_message(self, response, expected_message):
        self.assertIn('messages', response.context)
        messages = [m.message for m in response.context['messages']]
        self.assertIn(expected_message, messages)


class SSHViewDeleteKeysTestCase(SSHKeyTestCaseBase):

    def setUp(self):
        super(SSHViewDeleteKeysTestCase, self).setUp()
        switch_patch = switches(
            SSH_KEY_INTEGRATION=True,
            USERNAME_UI=True,
        )
        switch_patch.patch()
        self.addCleanup(switch_patch.unpatch)
        self.account = self.factory.make_account(
            password=DEFAULT_USER_PASSWORD,
            email='example_user@example.com')
        assert self.client.login(
            username=self.account.preferredemail.email,
            password=DEFAULT_USER_PASSWORD)

    def set_user_has_keys(self, keys):
        mock = self.patch('webui.forms.get_lp_ssh_keys')
        mock.return_value = keys

    def test_form_errors_are_rendered(self):
        self.set_user_has_keys(['ssh-rsa foo bar'])
        resp = self.client.post(
            reverse('ssh_keys'),
            data={
                'delete_keys': '',
                'ssh_keys': [],
            }
        )

        self.assertFormError(
            resp,
            'delete_key_form',
            'ssh_keys',
            'This field is required.',
        )

    def test_can_delete_ssh_key(self):
        key = 'ssh-dss keytext_goes_here comment goes here'
        self.set_user_has_keys([key])
        mock_delete = self.patch('webui.forms.delete_lp_ssh_key')

        resp = self.client.post(
            reverse('ssh_keys'),
            data={
                'delete_keys': '',
                'ssh_keys': [key],
            },
            follow=True
        )

        self.assertRedirects(resp, reverse('ssh_keys'))
        mock_delete.assert_called_once_with(
            self.account.openid_identifier,
            key,
            dry_run=False
        )

    def test_deleting_keys_shows_user_message(self):
        key = 'ssh-dss keytext_goes_here comment goes here'
        self.set_user_has_keys([key])
        self.patch('webui.forms.delete_lp_ssh_key')

        resp = self.client.post(
            reverse('ssh_keys'),
            data={
                'delete_keys': '',
                'ssh_keys': [key],
            },
            follow=True
        )
        self.assert_response_includes_message(
            resp, _make_deleted_keys_message(['comment goes here']))


class SSHViewAddKeyTestCase(SSHKeyTestCaseBase):

    def setUp(self):
        super(SSHViewAddKeyTestCase, self).setUp()
        switch_patch = switches(
            SSH_KEY_INTEGRATION=True,
            USERNAME_UI=True,
        )
        switch_patch.patch()
        self.addCleanup(switch_patch.unpatch)
        self.account = self.factory.make_account(
            password=DEFAULT_USER_PASSWORD,
            email='example_user@example.com')
        assert self.client.login(
            username=self.account.preferredemail.email,
            password=DEFAULT_USER_PASSWORD)
        self.patch('webui.forms.get_lp_ssh_keys', return_value=[])
        self.mock_add_keys = self.patch('webui.forms.add_lp_ssh_key')

    def test_form_errors_are_rendered(self):
        resp = self.client.post(
            reverse('ssh_keys'),
            data={
                'import_key': '',
                'ssh_key': '',
            },
        )
        self.assertFormError(
            resp, 'import_key_form', 'ssh_key', 'This field is required.')

    def test_can_add_ssh_key(self):
        key = 'ssh-rss foo bar'
        resp = self.client.post(
            reverse('ssh_keys'),
            data={
                'import_key': '',
                'ssh_key': key,
            },
            follow=True
        )
        self.assertRedirects(resp, reverse('ssh_keys'))
        self.assertEqual(
            call(self.account.openid_identifier, key, dry_run=False),
            self.mock_add_keys.call_args
        )

    def test_adding_ssh_key_notifies_user(self):
        key = 'ssh-rss foo bar'
        resp = self.client.post(
            reverse('ssh_keys'),
            data={
                'import_key': '',
                'ssh_key': key,
            },
            follow=True
        )
        self.assert_response_includes_message(
            resp, _make_added_key_message('bar'))
