# Copyright 2010, 2012 Canonical Ltd.  This software is licensed under
# the GNU Affero General Public License version 3 (see the file
# LICENSE).

from django.conf import settings
from django.conf.urls import url
from django.views.generic.base import RedirectView

from identityprovider.models import AUTHTOKEN_PATTERN

from webui.views.account import (
    account_deactivate,
    account_edit,
    account_emails,
    applications,
    auth_log,
    cookie,
    delete_account,
    delete_email,
    delete_email_by_id,
    index,
    invalidate_email,
    new_email,
    ssh_keys,
    verify_email,
    verify_email_by_id,
)
from webui.views.consumer import (
    finish_open_id,
    rpXRDS,
    start_open_id,
)
from webui.views.devices import (
    device_addition,
    device_generate,
    device_help,
    device_list,
    device_print,
    device_removal,
    device_rename,
)
from webui.views.i18n import set_language
from webui.views.registration import (
    forgot_password,
    sync_2fa,
)
from webui.views.ui import (
    bad_token,
    claim_token,
    combo_view,
    confirm_email,
    deactivated,
    login_view,
    logout,
    logout_to_confirm,
    new_account,
    permanent_redirect_home,
    reset_password,
    static_page,
    suspended,
    twofactor_view,
)


repls = {
    'token': '(?P<token>[A-Za-z0-9]{16})',
    'authtoken': '(?P<authtoken>%s)' % AUTHTOKEN_PATTERN,
    'email_address': '(?P<email_address>.+)',
    'fingerprint': '(?P<fingerprint>.+)',
    'device_id': '(?P<device_id>.+)',
}

urlpatterns = [
    # Django's URL-reversing doesn't like to play with regex
    # alternates '|' or optionals '?'.  Named parameters in such
    # clauses get clobbered to such names as '_0'.  Breaking out the
    # combinations ourselves ensures the names are maintained.
    url(r'^\+login$', login_view, name='login'),
    url(r'^%(token)s/\+login$' % repls, login_view, name='login'),

    url(r'^\+logout$' % repls, logout, name='logout'),
    url(r'^%(token)s/\+logout$' % repls, logout, name='logout'),

    url(r'^two_factor_auth$', twofactor_view, name='twofactor'),
    url(r'^%(token)s/two_factor_auth$' % repls, twofactor_view,
        name='twofactor'),

    url(r'^\+new_account$' % repls, new_account, name='new_account'),
    url(r'^%(token)s/\+new_account$' % repls, new_account,
        name='new_account'),

    url(r'^token/%(authtoken)s/$' % repls, claim_token, name='claim_token'),

    url(r'^token/%(authtoken)s/\+resetpassword/%(email_address)s$' % repls,
        reset_password, name='reset_password'),
    url(r'^%(token)s/token/%(authtoken)s/\+resetpassword/%(email_address)s$'
        % repls, reset_password, name='reset_password'),

    url(r'^token/%(authtoken)s/\+newaccount/%(email_address)s$' % repls,
        permanent_redirect_home),
    url(r'^confirm-account/%(authtoken)s/%(email_address)s$' % repls,
        permanent_redirect_home),
    url(r'^%(token)s/confirm-account/%(authtoken)s/%(email_address)s$' % repls,
        permanent_redirect_home),

    url(r'^token/%(authtoken)s/\+newemail/%(email_address)s$' % repls,
        confirm_email, name='confirm_email'),
    url(r'^%(token)s/token/%(authtoken)s/\+newemail/%(email_address)s$' %
        repls, confirm_email, name='confirm_email'),
    url(r'^\+bad-token', bad_token, name='bad_token'),
    url(r'^\+logout-to-confirm', logout_to_confirm,
        name='logout_to_confirm'),
    url(r'^\+deactivated$', deactivated, name='deactivated'),
    url(r'^\+suspended$', suspended, name='suspended'),
    url(r'^\+description$', static_page, {'page_name': 'description.html'},
        name='description'),
    url(r'^\+faq$', static_page, {'page_name': 'faq.html'}, name='faq'),
    url(r'^\+ubuntuone-account$', static_page,
        {'page_name': 'ubuntuone-account.html', 'context': {
            'use_app_links': True
        }}, name='ubuntuone-account'),
    url(r'^combo/$', combo_view, name='combo-view'),

    url(r'^robots.txt$', static_page,
        {'page_name': 'robots.txt', 'content_type': 'text/plain'},
        name='robots'),

    # the urls formerly know as Ubuntu One
    url(r'^one-redirect$', static_page, {'page_name': 'one-redirect.html'},
        name='one-redirect'),
    url(r'^terms/$', RedirectView.as_view(url=settings.TERMS_OF_SERVICE_URL),
        name='terms_and_conditions'),
    # XXX cjwatson 2018-05-23: Is this still used?  If so, we need an
    # embedded-themed version of this page on the legal site that we can
    # redirect to.
    url(r'^terms/embedded/$', static_page,
        {'page_name': 'terms_embedded.html'},
        name='terms_and_conditions_embedded'),
    url(r'^privacy/$', RedirectView.as_view(url=settings.PRIVACY_POLICY_URL),
        name='privacy_policy'),
    url(r'^privacy-notice/$',
        RedirectView.as_view(url=settings.PRIVACY_NOTICE_URL),
        name='privacy_notice'),
]

urlpatterns += [
    url(r'^\+forgot_password$' % repls, forgot_password,
        name='forgot_password'),
    url(r'^%(token)s/\+forgot_password$' % repls, forgot_password,
        name='forgot_password'),
    url(r'^sync_2fa$' % repls, sync_2fa,
        name='sync_2fa'),
]

urlpatterns += [
    url(r'^$', index, name='account-index'),
    url(r'^%(token)s$' % repls, index, name='account-index'),
    url(r'^\+emails$', account_emails, name='account-emails'),
    url(r'^%(token)s/\+emails$' % repls, account_emails,
        name='account-emails'),
    url(r'^\+edit$', account_edit, name='account-edit'),
    url(r'^%(token)s/\+edit$' % repls, account_edit, name='account-edit'),
    url(r'^\+cookie$', cookie),
    url(r'^\+index$' % repls, index, name='index'),
    url(r'^%(token)s/\+index$' % repls, index, name='index'),
    url(r'^\+new-email$' % repls, new_email, name='new_email'),
    url(r'^%(token)s/\+new-email$' % repls, new_email, name='new_email'),
    # XXX: deprecated
    url(r'^\+verify-email$' % repls, verify_email_by_id,
        name='verify_email_by_id'),
    url(r'^%(token)s/\+verify-email$' % repls, verify_email_by_id,
        name='verify_email_by_id'),
    # enddeprecated
    url(r'^verify/%(email_address)s$' % repls, verify_email,
        name='verify_email'),
    url(r'^%(token)s/verify/%(email_address)s$' % repls, verify_email,
        name='verify_email'),
    url(r'^invalidate-email/%(authtoken)s/%(email_address)s$' % repls,
        invalidate_email, name='invalidate_email'),
    # XXX: deprecated
    url(r'^\+remove-email$' % repls, delete_email_by_id,
        name='delete_email_by_id'),
    url(r'^%(token)s/\+remove-email$' % repls, delete_email_by_id,
        name='delete_email_by_id'),
    # enddeprecated
    url(r'^delete/%(email_address)s$' % repls, delete_email,
        name='delete_email'),
    url(r'^%(token)s/delete/%(email_address)s$' % repls, delete_email,
        name='delete_email'),
    url(r'^\+applications$', applications, name='applications'),
    url(r'^\+deactivate$', account_deactivate, name='account_deactivate'),
    url(r'^%(token)s/\+deactivate$' % repls, account_deactivate,
        name='account_deactivate'),
    url(r'^\+delete$', delete_account, name='delete_account'),
    url(r'^activity$', auth_log, name='auth_log'),
    url(r'^ssh-keys$', ssh_keys, name='ssh_keys'),
]
# TODO: Support the login-tokens, if it makes sense.
urlpatterns += [
    url(r'^device-list$', device_list, name='device-list'),
    url(r'^%(token)s/device-list$' % repls, device_list, name='device-list'),
    url(r'^device-addition$', device_addition, name='device-addition'),
    url(r'^device-removal/%(device_id)s$' % repls, device_removal,
        name='device-removal'),
    url(r'^device-rename/%(device_id)s$' % repls, device_rename,
        name='device-rename'),
    url(r'^\+device-help$', device_help, name='device-help'),
    url(r'^device-print/%(device_id)s$' % repls, device_print,
        name='device-print'),
    url(r'^%(token)s/device-print/%(device_id)s$' % repls, device_print,
        name='device-print'),
    url(r'^device-generate/%(device_id)s$' % repls, device_generate,
        name='device-generate'),
    url(r'^%(token)s/device-generate/%(device_id)s$' % repls,
        device_generate, name='device-generate'),
]

urlpatterns += [
    url(r'^set_language$', set_language, name='set_language'),
]

urlpatterns += [
    url(r'^consumer/$', start_open_id, name='start_open_id'),
    url(r'^consumer/finish/$', finish_open_id, name='finish_open_id'),
    url(r'^consumer/xrds/$', rpXRDS, name='rp_xrds'),
]
